/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   ccoverage.cpp
 *		- methods for verilog coverage tool 
 *
 ******************************************************************************
 */
						

#include <vector>
#include <fstream>
#include <config.h>
#include "main.h"
#include "ccoverage.h"
#include "cdumpverilog.h"
#include "cmodule.h"
#include "cport.h"
#include "cfref.h"
#include "cvar.h"
#include "cnet.h"
#include "cparam.h"
#include "cnode.h"
#include "cinstance.h"
#include "cfunction.h"
#include "cportdir.h"


static int		lineId;
static CNode*		moduleHeader;
static CVar*		moduleBBReg;
static FILE*		outputFile;
static const char*	moduleName;

enum    TranslateState { eTRANSLATE_DEFAULT, eTRANSLATE_OFF, eTRANSLATE_ON } ;
static  TranslateState	translateState;

extern  char*		outputPath;
extern  int		errorCount;

static CNode* InsertStatement( CNode* n ); 

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
#ifndef CYGWIN
extern "C" {
CBackend* CreateToolInstance()
{
    return new CCoverage();
}
}
#endif

/************************************************
 	Constructor
 ************************************************/
CCoverage::CCoverage( ) 
{
	translateState = eTRANSLATE_DEFAULT;
	switches.push_back( "+coverage_output_file=<filename>" );
	switchDescription["+coverage_output_file=<filename>"] = 
		"Filename for coverage line mapping info";
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CCoverage::GetToolName( void )
{
	static char	toolName[] = "coverage";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CCoverage::GetToolDescription( void )
{
	static char	toolDescription[] = "Add line coverage instrumentation";

	return toolDescription;
}


/*******************************************************
	IsTranslateOn
	- return TRUE if PRAGMA node is translate on 
********************************************************/

static int IsTranslateOn( CNode* n )
{
	char* comment = n->Arg<char*>(0);
	char* ptr = strstr( comment, "synopsys" );
	if( ptr == NULL ) {
		return FALSE;
	}
	ptr = strstr( ptr, "translate_on" );
	return (ptr != NULL );
}

/*******************************************************
	IsTranslateOff
	- return TRUE if PRAGMA node is translate off 
********************************************************/

static int IsTranslateOff( CNode* n )
{
	char* comment = n->Arg<char*>(0);
	char* ptr = strstr( comment, "synopsys" );
	if( ptr == NULL ) {
		return FALSE;
	}
	ptr = strstr( ptr, "translate_off" );
	return (ptr != NULL );
}
/*******************************************************
	FindCoord
	- find next specified coordinate
	  returns NULL if cannot find one
 *******************************************************/

static Coord_t* FindCoord( CNode* n ) 
{

	if( n == NULL ) {
		return NULL;
	}
	
	if( n->GetCoord()->filename != NULL ) {
		return n->GetCoord();
	}
	
	Coord_t* coord = NULL;
	if( n->GetOp() == eLIST ) {
		coord = FindCoord( n->Arg<CNode*>(0) );
		if( coord != NULL ) {
			return coord;
		}
		coord = FindCoord( n->Arg<CNode*>(1) );
		if( coord != NULL ) {
			return coord;
		}
	}
	return NULL;	
}

/*******************************************************
	AddMarker
	- insert coverage line marker 
********************************************************/

static CNode* AddMarker( CNode* n )
{
	if( n == NULL ) {
		return n;
	}
	if( translateState == eTRANSLATE_OFF ) {
		return n;
	}
	char buffer[32];
	
	/*
	 * for now force nodes without line info to work
	 */
	char* comment;
	Coord_t* coord = FindCoord( n );
	if( coord ) {
		sprintf( buffer, "// line id %d -> ", lineId ); 
		CNode::CurrentHeap()->Grow(buffer,strlen(buffer));
		const char* name = strrchr( coord->filename, '/' );
		if( !name ) {
		    name = coord->filename;
		} else {
		    name++;
		}
		CNode::CurrentHeap()->Grow(name,strlen(name));
		sprintf( buffer, "[%d]\n", coord->lineno ); 
		CNode::CurrentHeap()->Grow(buffer,strlen(buffer)+1);
		comment = (char*)CNode::CurrentHeap()->Finish();
		if( outputFile ) {
			const char* name = strrchr( coord->filename, '/' );
			if( name ) {
			    name++;
			} else {
			    name = coord->filename;
			}
			fprintf( outputFile, "%s[%d] %s %d\n",
				moduleName, lineId, 
				name,
				coord->lineno );
		}
	} else {
		sprintf( buffer, "// line id %d -> ???", lineId );
		CNode::CurrentHeap()->Grow(buffer,strlen(buffer)+1);
		comment = (char*)CNode::CurrentHeap()->Finish();
		if( outputFile ) {
			fprintf( outputFile, "%s[%d] ??? 0\n",
				moduleName, lineId  );
		}
	}
	if( moduleHeader ) {
		moduleHeader = cLIST(moduleHeader,cCOMMENT(comment));
	} else {
		moduleHeader = cCOMMENT(comment);
	}

	int value = lineId; 
	lineId++;
	return cLIST( 
		cASSIGN( NULL, 
			cARRAY(cVAR_REF(moduleBBReg),cINT32(value)),
			cINT32(1) 
		), 
	n );
}

/*******************************************************
	InsertCaseItem
	- insert coverage instrumentation 
********************************************************/

static CNode* InsertCaseItem( CNode* n, int addMarker ) 
{
        if( n->GetOp() == eCOMMENT ||
	    n->GetOp() == eVRQ ||
	    n->GetOp() == ePRAGMA ) {
                return n;
        }
        if( n->GetOp() == eLIST ) {
                n->Arg<CNode*>(0) = InsertCaseItem( n->Arg<CNode*>(0), addMarker );
                n->Arg<CNode*>(1) = InsertCaseItem( n->Arg<CNode*>(1), addMarker );
                return n;
        }

	if( addMarker ) {
            n->Arg<CNode*>(1) = AddMarker(InsertStatement(n->Arg<CNode*>(1)));
	} else {
            n->Arg<CNode*>(1) = InsertStatement(n->Arg<CNode*>(1));
	}
	return n;

}

/*******************************************************
	InsertStatement
	- insert coverage instrumentation 
********************************************************/

static CNode* InsertStatement( CNode* n ) 
{
	if( n == NULL ) {
		return n;
	}


	switch( n->GetOp() ) {
	case ePASSIGN:
	case eDEASSIGN:
	case eDISABLE:
	case eTRIGGER:
	case eNET_DECL:
	case eVAR_DECL:
	case ePARAM_DECL:
	case ePORT_DECL:
	case eGENVAR_DECL:
	case eTYPEDEF_DECL:
		break;
	case eFUNCTION_CALL:
	case eTASK_ENABLE:
	case eSYSTASK_CALL:
	case eINSTANCE_REF:
	case eGATE_REF:
		break;
	case ePRAGMA:
		if( IsTranslateOn(n) ) {
			translateState = eTRANSLATE_ON;
		} else if( IsTranslateOff(n) ) {
			if( translateState == eTRANSLATE_OFF ) {
				error( n->GetCoord(), "nested synopsys translate_off statements" );
			}
			translateState = eTRANSLATE_OFF;
		}
		break;
	case eCOMMENT:
	case eVRQ:
	case eATTRIBUTE:
		break;
	case eLIST:
		{
		  CNode* l = InsertStatement( n->Arg<CNode*>(0) );
		  CNode* r = InsertStatement( n->Arg<CNode*>(1) );
		  if( l == NULL ) {
			return r;
		  } else if( r == NULL ) {
			return l;
		  } else {
			n->Arg<CNode*>(0) = l;
			n->Arg<CNode*>(1) = r;
		  }
	   	}
		break;	
	case eALWAYS:
		n->Arg<CNode*>(0) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(0)));
		break;
	case eINIT:
		break;
	case eEVENT:
		n->Arg<CNode*>(1) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(1)));
		break;
	case eWHILE:
		n->Arg<CNode*>(1) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(1)));
		break;	
	case eREPEAT:
		n->Arg<CNode*>(1) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(1)));
		break;	
	case eFOREVER:
		n->Arg<CNode*>(1) = AddMarker(InsertStatement(n->Arg<CNode*>(1)));
		break;	
	case eFOR:
		n->Arg<CNode*>(3) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(3)));
		break;	
	case eBLOCK_REF:
		n->Arg<CNode*>(1) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(1)));
		break;
	case eDELAY:
		n->Arg<CNode*>(1) = 
			AddMarker(InsertStatement(n->Arg<CNode*>(1)));
		break;	
	case eFUNCTION_DEF:
		{
			CBlock* block = n->Arg<CBlock*>(0);
			CNode*  n = block->GetCodeList();
			n = InsertStatement( n );
			block->SetCodeList( AddMarker(n) );	
		}
		break;
	case eMODULE_DEF:
		{
			CBlock* block = n->Arg<CBlock*>(0);
			lineId = 0;
			moduleHeader = NULL;
			CModule* m = n->Arg<CModule*>(0);
			moduleName = m->GetName();
			CSimpleType* stype = new(CNode::CurrentHeap()) CSimpleType(eREG);
			moduleBBReg = new(CNode::CurrentHeap()) CVar( 
				CSymbol::Lookup("___vrq_bbvector" ), 
				n->GetCoord(), stype, TRUE );
			CNode* code = block->GetCodeList();
			code = InsertStatement(code);
			if( moduleHeader ) {
				code = cLIST(moduleHeader,code);
			}
			block->SetCodeList( code );	
		}
		break;
	case eSPECIFY_REF:
	case eWAIT:
	case eCASSIGN:
	case eFORCE:
	case eDEFPARAM:
	case eRELEASE:
	case ePORTLIST_END:
		break;
	case eCASEX:
	case eCASEZ:
	case eCASE:
		n->Arg<CNode*>(1) = InsertCaseItem( n->Arg<CNode*>(1), TRUE );
		break;
	case eASSIGN:
	case eNBASSIGN:
		break;
	case eGIF:
    	    n->Arg<CNode*>(1) = InsertStatement( n->Arg<CNode*>(1));
	    n->Arg<CNode*>(2) = InsertStatement( n->Arg<CNode*>(2));
	    break;
	case eGFOR:
	    n->Arg<CNode*>(3) = InsertStatement(n->Arg<CNode*>(3));
	    break;	
	case eGCASE:
	    n->Arg<CNode*>(1) = InsertCaseItem( n->Arg<CNode*>(1), FALSE );
	    break;
	case eIF:
    	    n->Arg<CNode*>(1) = AddMarker(InsertStatement( n->Arg<CNode*>(1)) );
	    n->Arg<CNode*>(2) = AddMarker(InsertStatement( n->Arg<CNode*>(2)) );
		break;
	default:
//cout << nodeOpName[n->GetOp()] << endl;
		MASSERT( FALSE );
	}

	return n;
	
}


/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CCoverage::Process( list<CElement>& inputList,
			 list<CElement>& outputList )
{
	/*
	 * determine output directory
	 */
	char* fname = GetPlusArg( "coverage_output_file=" );
	if( fname != NULL ) {
		outputFile = fopen( fname, "w" );
		if( outputFile == NULL ) {
			error( NULL, "Could not create file '%s'\n", fname );
		}
	} else {
		outputFile = NULL;
	}
	
	/*
	 * scan though code list adding
	 * code to instrument line coverage
	 */
        list<CElement>::iterator ptr;
        for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	    CNode* code = InsertStatement( ptr->Code() );
	    outputList.push_back( CElement( ptr->Filename(), 
					!ptr->Filename(), code ) );
    	}

}
