/******************************************************************************
 * 	Copyright (c) 1997-2003 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   cflatten.cpp
 *		- methods for tool to flatten designs
 *
 ******************************************************************************
 */
						

#include <vector>
#include <set>
#include <fstream>
#include <config.h>
#include "compile.h"
#include "cflatten.h"
#include "cdumpverilog.h"
#include "cmodule.h"
#include "cport.h"
#include "cfref.h"
#include "cvar.h"
#include "cnet.h"
#include "cparam.h"
#include "cnode.h"
#include "cinstance.h"
#include "cfunction.h"
#include "cportdir.h"

/*
 * memory management objects 
 */
static	CObstack 	theBackStack( "flattenStack" );
static	CObstack*	backStack = &theBackStack;

extern  char*		outputPath;

static map<CModule*,CNode*> module2Comments;

extern CDumpVerilog*	dumpVerilog;
extern int		errorCount;

static void Remap( CNode* n, void* arg );
static CNode* Substitute( CNode* n, void* arg );
static Message* mTLVMD;
static Message* mMISUD;
static Message* mPOIUC;
static Message* mFIANS;
static Message* mISEPC;

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
#ifndef CYGWIN
extern "C" {
CBackend* CreateToolInstance()
{
    return new CFlatten();
}
}
#endif

/************************************************
 	Constructor
 ************************************************/
CFlatten::CFlatten() 
{
	mTLVMD = Message::RegisterWarning( this, Message::eINFO,
		"TLVMD", "top level module '%s'\n", "<module>" );
	mMISUD = Message::RegisterWarning( this, Message::eWARNING,
		"MISUD", "module '%s' is undefined\n", "<module>" );
	mPOIUC = Message::RegisterWarning( this, Message::eWARNING,
		"POIUC", "port %s of %s is unconnected", 
				"<port>", "<module>" );
	mFIANS = Message::RegisterWarning( this, Message::eWARNING,
		"FIANS", "flattening of instance arrays is not supported" );
	mISEPC = Message::RegisterWarning( this, Message::eWARNING, 
		"ISEPC", "instance %s specifies empty port connection", 
		"<instance>" );
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CFlatten::GetToolName( void )
{
	static char	toolName[] = "flatten";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CFlatten::GetToolDescription( void )
{
	static char	toolDescription[] = "Reduce to single level hiearchy";

	return toolDescription;
}

/************************************************
	ResolveInstance
	- return TRUE for any instance that need
	  to be resolved
 ************************************************/
int CFlatten::ResolveInstance( CModule* m, CInstance* inst )
{
    return !inst->IsGate();
}


/*******************************************************
	ValidLval
	- returns TRUE if tree is a valid lval.
********************************************************/

static int ValidLval( 
	CNode* n
)
{
    if( n == NULL ) {
	    return FALSE;
    }

    switch( n->GetOp() ) {
    case eARRAY:
    case eNET_REF:
    case eVAR_REF:
    case eCAT:
    case eUCAT:
    case eEXTERNAL_REF:
	return TRUE;
    case ePORT_REF:
	{
	    CPortDir* port = (CPortDir*)n->Arg<CPortDir*>(0);
	    switch( port->GetType() ) {
	    case eINPUT:
		return FALSE;
	    case eOUTPUT:
	    case eINOUT:
		return TRUE;
	    default:
		MASSERT( FALSE );
	    }
	}
    default:;
    }

    return FALSE;
}

/*******************************************************
	MatchPorts
	- match up instance connections with port 
	  definitions. Returns TRUE if match succeeds
********************************************************/

static int MatchPorts( 
	CInstance*          inst,
	CModule*            def,
	map<CPort*,CNode*>& portMap,
	list<CPort*>&       emptyPorts )
{
    int byName = TRUE;
    int argNum = 0;
    vector<CPort*>* portList = def->GetPortList();
    vector<CNode*>* argList  = inst->GetArgumentVector();
    
    /*
     * build name to port mapping
     */
    vector<CPort*>::iterator portPtr;
    map<CSymbol*,CPort*> sym2Port;
    for( portPtr = portList->begin(); portPtr != portList->end(); ++portPtr ) {
	sym2Port[(*portPtr)->GetSymbol()] = *portPtr; 
    }
    copy( portList->begin(), portList->end(), emptyPorts.begin() );
    vector<CNode*>::iterator argPtr;
    for( argPtr = argList->begin(); argPtr != argList->end(); ++argPtr ) {
	CPort* port = NULL;
	int empty  = *argPtr == NULL || (*argPtr)->Arg<CNode*>(1) == NULL;
	int dotted = *argPtr != NULL && (*argPtr)->Arg<CNode*>(0) != NULL;
	CSymbol* sym = NULL;
	if( dotted ) {
	     sym = (*argPtr)->Arg<CSymbol*>(0);
	}
	if( argNum == 0 ) {
	    if( !dotted ) {
		    byName = FALSE;
	    }
	} else {
	    if( dotted != byName ) {
		error( inst->GetCoord(), 
		       "instance %s has mix of port mapping methods", inst->GetName() );
	    }
	}
	if( dotted ) {
	    if( sym2Port.find(sym) == sym2Port.end() ) {
		error( inst->GetCoord(), 
		       "instance %s specifies unknown port name %s", 
		       inst->GetName(), sym->GetName() );
		continue;
	    }
	    port = sym2Port[sym];
	} else {
	    if( portList->size() <= (unsigned)argNum ) {
		error( inst->GetCoord(), 
		       "instance %s specifies too many port connections", 
		       inst->GetName() );
		continue;
	    }
	    port = (*portList)[argNum];
	}
	CNode* exp = (*argPtr)->Arg<CNode*>(1);
	if( exp == NULL ) {
	    message( inst->GetCoord(), mISEPC, inst->GetName() );
	} else {
	    if( port->Direction() != eINPUT &&
		!ValidLval( exp ) ) {
		error( inst->GetCoord(), 
		       "instance %s cannot be connected to port %d of module %s", 
		       inst->GetName(), argNum+1, def->GetName() );
		continue;
	    }
	    CNode* conn = port->GetConnection();
	    MASSERT( conn );
	    if( exp->GetWidth() != conn->GetWidth() ) {
		error( inst->GetCoord(), 
		       "instance %s width mismatch to port %d of module %s", 
		       inst->GetName(), argNum+1, def->GetName() );
	    }	    
	}
	portMap[port] = exp;
	argNum++;
    }

    
    return TRUE;
}

/*******************************************************
	RemapDecl
	- add entry to decl remapping table
********************************************************/
static void RemapDecl( map<CDecl*,CDecl*>& remapDecl, CDecl* from, CDecl* to )
{
	remapDecl[from] = to;
}

/*******************************************************
	ElaborateModules
	- inline all modules
********************************************************/

static CNode* ElaborateModules( 
	set<CModule*>& topLevelModules,
	CModule* top,
	const char* path,
	CModule* m,
	CNode* n
)
{
	if( n == NULL ) {
		return n;
	}


	switch( n->GetOp() ) {
	case eDISABLE:
	case ePASSIGN:
	case eDEASSIGN:
	case eTRIGGER:
	case eFUNCTION_CALL:
	case eTASK_ENABLE:
	case eSYSTASK_CALL:
	case eGATE_REF:
	case ePRAGMA:
	case eDEFPARAM:
	case ePORTLIST_END:
	    break;
	case eTYPEDEF_DECL:
	case ePARAM_DECL:
	case eVAR_DECL:
	case eGENVAR_DECL:
	    /*
	     * remove local declarations from leaf modules
	     */
	    if( m != top ) {
		return NULL;
	    }
	    break; 
	case eNET_DECL:
	    /*
	     * remove local declarations from leaf modules
	     * move assigments if needed
	     */
	    if( m != top && n->Arg<CNode*>(1) ) {
		return n->Arg<CNode*>(1);
	    }
	    if( m != top ) {
		return NULL;
	    }
	    break; 
	case eLIST:
		{
		  CNode* l = ElaborateModules( topLevelModules, 
			                       top, path, 
					       m, n->Arg<CNode*>(0) );
		  CNode* r = ElaborateModules( topLevelModules, 
			    		       top, path, 
					       m, n->Arg<CNode*>(1) );
		  return cLINK( l, r );
	   	}
		break;	
	case eALWAYS:
	case eINIT:
	case eEVENT:
	case eWHILE:
	case eREPEAT:
	case eFOREVER:
	case eFOR:
	case eBLOCK_REF:
	case eSPECIFY_REF:
	case eDELAY:
	case eFUNCTION_DEF:
	case eWAIT:
	case eCASSIGN:
	case eFORCE:
	case eRELEASE:
	case eCASEX:
	case eCASEZ:
	case eASSIGN:
	case eNBASSIGN:
	case eCASE:
	case eGCASE:
	case eGFOR:
	case eGIF:
	case eIF:
		break;
	case eMODULE_DEF:
		{
		    CModule* m = n->Arg<CModule*>(0);
		    if( !topLevelModules.count(m) ) {
			return NULL;
		    }
		    CNode* newCode;
		    newCode = ElaborateModules( topLevelModules, 
						m, "", 
						m, m->GetCodeList() );
		    m->SetCodeList( newCode );
		    CNode* comments = NULL;
		    if( module2Comments.find(m) != module2Comments.end() ) {
			comments = module2Comments[m];
		    }
		    return cLINK( comments, n );
		}
	case eINSTANCE_REF:
		{
		    CNode* newDeclarations = NULL;
		    CNode* newAssigns = NULL;
		    CNode* code = NULL;
		    CInstance* inst = n->Arg<CInstance*>(0);
		    CModule* def = inst->GetDefinition();
		    if( !def || def->IsPrimitive() || !def->IsDefined() ) {
			break;
		    }
		    if( inst->GetNumberOfUnpackedDimensions() ) {
			message( n->GetCoord(), mFIANS );
			break;
		    }
		    if( m != NULL && def->Unit() != m->Unit() ) {
			error( n->GetCoord(), "incompatable timescales" );
			break;
		    }
		    map<CPort*,CNode*> portMap;
		    list<CPort*>       emptyPorts;
		    map<CDecl*,CNode*> substituteDecl;
		    map<CDecl*,CDecl*> remapDecl;
		    int success = MatchPorts( inst,
					       def,
					       portMap,
					       emptyPorts );
		    /*
		     * HACK: For some reason gcc4.0 (64 bit mode)
		     *       is modifying 'def'. For now just
		     * 	     reassign it to avoid the problem
		     */
		    def = inst->GetDefinition();

		    list<CPort*>::iterator ptr;
		    for( ptr = emptyPorts.begin(); ptr != emptyPorts.end(); ++ptr ) {
			message( n->GetCoord(), mPOIUC,
				(*ptr)->GetName(),
				def->GetName() );
		    }
		    if( !success ) {
			break;
		    }
		    /*
		     * supported connection cases:
		     * 1) wire connected to reg
		     *	  - remove reg declaration and 
		     *      change VAR_REF's to NET_REF's
		     *	    pointing to connections
		     * 2) expression(lval) connected to reg
		     *	  - create continious assign
		     * 3) wire connected to wire
		     *	  - delete wire declarations,
		     *	    promote wire type, change
		     *	    NET_REF's to point to connections
		     *	    decl
		     * 4) expression connected to wire
		     *	  - substitute expression for NET_REF's
		     *	    delete wire decls
		     */
		    map<CDecl*,CPort*> decl2Port;
		    map<CPort*,CNode*>::iterator mPtr;
		    for( mPtr = portMap.begin(); 
				    mPtr != portMap.end(); ++mPtr ) {
			CPort* p = mPtr->first;
			CNode* ref = p->GetConnection();
			int portIsReg;
			switch( ref->GetOp() ) {
			case eVAR_REF:
			    portIsReg = TRUE;
			    break;
			case eNET_REF:
			    portIsReg = FALSE;
			    decl2Port[ref->Arg<CNet*>(0)] = p;
			    break;
			default:
			    MASSERT( FALSE );
			}
			CNode* conn = portMap[p];
			if( conn == NULL ) {
			    trace( NULL, "EMPTY: port %s of %s", def->GetName(), p->GetName() );
			    // -- do nothing
			    continue;
			}
			int connIsNet = (conn->GetOp() == eNET_REF);
			MASSERT(conn->GetOp() != ePORT_REF);
			if( connIsNet && portIsReg ) {
			    trace( NULL, "NET TO REG: port %s of %s", def->GetName(), ref->Arg<CVar*>(0)->GetName() );
			    StrengthPair_t* pair = (StrengthPair_t*) 
				    backStack->Alloc( sizeof(StrengthPair_t) );
			    pair->s0 = eUNDEFINED;
			    pair->s1 = eUNDEFINED;
			    CNode* assign = cCASSIGN( pair, NULL, 
				    conn->Clone(), ref->Clone() );
			    newAssigns = cLINK( newAssigns, assign );	    
			    // -- in m: re-type net to reg
			    // -- in code substitute reg to net(reg)
			}
			if( !connIsNet && portIsReg ) {
			    trace( NULL, "EXP TO REG: port %s of %s", def->GetName(), p->GetName() );
			    StrengthPair_t* pair = (StrengthPair_t*) 
				    backStack->Alloc( sizeof(StrengthPair_t) );
			    pair->s0 = eUNDEFINED;
			    pair->s1 = eUNDEFINED;
			    CNode* assign = cCASSIGN( pair, NULL, 
				    conn->Clone(), ref->Clone() );
			    newAssigns = cLINK( newAssigns, assign );	    
			    // -- create continious assign
			}
			if( connIsNet && !portIsReg ) {
			    // -- in code substitute net for net
			    RemapDecl( remapDecl, ref->Arg<CNet*>(0),
						  conn->Arg<CNet*>(0) );
			    trace( NULL, "WIRE TO WIRE: port %s of %s", def->GetName(), p->GetName() );
			}
			if( !connIsNet && !portIsReg ) {
			    substituteDecl[ref->Arg<CNet*>(0)] = conn;
			    trace( NULL, "EXP TO WIRE: port %s of %s %d", def->GetName(), p->GetName(), conn->GetOp() );
			    // -- in code subsitute expression for net
			}
		    }

		    /*
		     * add instance name to path
		     */
		    if( strcmp( path, "" ) ) {
			backStack->Grow(path, strlen(path));
			backStack->Grow((void*)"_", 1);
		    }
		    backStack->Grow(inst->GetName(), strlen(inst->GetName())+1);
		    char* newPath = (char*)backStack->Finish();

		    /*
		     * Rename local declarations to
		     * make them unique
		     */
		    vector<CParam*>* paramList = def->GetParamList();
		    vector<CParam*>::iterator paramPtr;
		    for( paramPtr = paramList->begin(); 
			 paramPtr != paramList->end(); ++paramPtr ) {
			if( decl2Port.find(*paramPtr) != decl2Port.end() ) {
			    continue;
			}
			CParam* newParam = dynamic_cast<CParam*>((*paramPtr)->Clone(backStack));
			backStack->Grow(newPath, strlen(newPath));
			backStack->Grow("_", 1);
			backStack->Grow((*paramPtr)->GetName(), strlen((*paramPtr)->GetName())+1);
			newParam->SetSymbol( CSymbol::Lookup( (char*)backStack->Finish() ) );
			newParam->SetDeclStatementCreated();
			newDeclarations = cLINK( newDeclarations, 
				 cPARAM_DECL( newParam, newParam->GetCoord() ) );
			top->Add(newParam);
			RemapDecl( remapDecl, *paramPtr, newParam );
			trace( NULL, "RENAME: param %s", newParam->GetName() );
		    }

		    vector<CNet*>* netList = def->GetNetList();
		    vector<CNet*>::iterator netPtr;
		    for( netPtr = netList->begin(); 
			 netPtr != netList->end(); ++netPtr ) {
			if( decl2Port.find(*netPtr) != decl2Port.end() &&
			    portMap[decl2Port[*netPtr]] != NULL ) {
			    continue;
			}
			CNet* newNet = dynamic_cast<CNet*>((*netPtr)->Clone(backStack));
			backStack->Grow(newPath, strlen(newPath));
			backStack->Grow("_", 1);
			backStack->Grow((*netPtr)->GetName(), strlen((*netPtr)->GetName())+1);
			newNet->SetSymbol( CSymbol::Lookup( (char*)backStack->Finish() ) );
			newNet->SetDeclStatementCreated();
			newDeclarations = cLINK( newDeclarations, 
				 cNET_DECL( newNet, NULL, newNet->GetCoord() ) );
			top->Add(newNet);
			RemapDecl( remapDecl, *netPtr, newNet );
			trace( NULL, "RENAME: net %s", newNet->GetName() );
		    }

		    vector<CVar*>* regList = def->GetVarList();
		    vector<CVar*>::iterator regPtr;
		    for( regPtr = regList->begin(); 
			 regPtr != regList->end(); ++regPtr ) {
			if( decl2Port.find(*regPtr) != decl2Port.end() ) {
			    continue;
			}
			if( (*regPtr)->GetNodeType() != eS && 
			    (*regPtr)->GetNodeType() != eB ) {
			    continue;
			}
			CVar* newReg = dynamic_cast<CVar*>((*regPtr)->Clone(backStack));
			backStack->Grow(newPath, strlen(newPath));
			backStack->Grow("_", 1);
			backStack->Grow((*regPtr)->GetName(), strlen((*regPtr)->GetName())+1);
			newReg->SetSymbol( CSymbol::Lookup( (char*)backStack->Finish() ) );
			newReg->SetDeclStatementCreated();
			newDeclarations = cLINK( newDeclarations, 
				 cVAR_DECL( newReg, NULL, newReg->GetCoord() ) );
			top->Add(newReg);
			RemapDecl( remapDecl, *regPtr, newReg );
			trace( NULL, "RENAME: reg %s", newReg->GetName() );
		    }

		    vector<CVar*>* realList = def->GetVarList();
		    vector<CVar*>::iterator realPtr;
		    for( realPtr = realList->begin(); 
			 realPtr != realList->end(); ++realPtr ) {
			if( decl2Port.find(*realPtr) != decl2Port.end() ) {
			    continue;
			}
			if( (*realPtr)->GetNodeType() != eR ) {
			    continue;
			}
			CVar* newReal = dynamic_cast<CVar*>((*realPtr)->Clone(backStack));
			backStack->Grow(newPath, strlen(newPath));
			backStack->Grow("_", 1);
			backStack->Grow((*realPtr)->GetName(), strlen((*realPtr)->GetName())+1);
			newReal->SetSymbol( CSymbol::Lookup( (char*)backStack->Finish() ) );
			newReal->SetDeclStatementCreated();
			newDeclarations = cLINK( newDeclarations, 
				 cVAR_DECL( newReal, NULL, newReal->GetCoord() ) );
			top->Add(newReal);
			RemapDecl( remapDecl, *realPtr, newReal );
			trace( NULL, "RENAME: real %s", newReal->GetName() );
		    }


		    CNode* ccode = def->GetCodeList();
		    if( ccode ) {
			ccode = ccode->Clone();
		    }
		    code = cLINK( code, 
			    ElaborateModules( topLevelModules,
					      top, newPath, 
					      def, ccode ) );
		    if( module2Comments.find(def) != module2Comments.end() ) {
			code = cLINK( module2Comments[def], code );
		    }
		    code = cLINK( newDeclarations, code );

		    const char* header = "// BEGIN INSTANCE ";
		    backStack->Grow((void*)header, strlen(header) );
		    backStack->Grow((void*)newPath, strlen(newPath));
		    backStack->Grow((void*)"\n", 2);
		    code = cLINK( cCOMMENT( (char*)backStack->Finish() ), code );
		    code = cLINK( code, newAssigns );

		    const char* trailer = "// END INSTANCE ";
		    backStack->Grow((void*)trailer, strlen(trailer) );
		    backStack->Grow((void*)newPath, strlen(newPath));
		    backStack->Grow((void*)"\n\n", 3);
		    code = cLINK( code, cCOMMENT( (char*)backStack->Finish() ) );
		    code->PostVisit1( Remap, &remapDecl );
		    code = code->PostSubVisit1( Substitute, &substituteDecl );
		    return code;
		}
	case ePORT_DECL:
		if( !topLevelModules.count(m) ) {
		    return NULL;
		}
		break;
	case eVRQ:
	case eATTRIBUTE:
		break;
	case eCOMMENT:
		{
		    if( !m ) {
			return NULL;
		    }
		    /*
		     * remove compiler directives
		     */
		    char* s = n->Arg<char*>(0);
		    s += strspn( s, " \t\f" );
		    if( s[0] == '`' ) {
			return NULL;
		    }
		}
		break;
	default:
		MASSERT( FALSE );
	}

	return n;
	
}

/************************************************
	Remap
	- remap modified decls
 ************************************************/
static void Remap( CNode* n, void* arg )
{
    map<CDecl*,CDecl*>* remapDecl = (map<CDecl*,CDecl*>*)arg;
    CDecl* decl;
    switch( n->GetOp() ) {
    case eNET_REF:
	decl = n->Arg<CDecl*>(0);
	if( remapDecl->find(decl) != remapDecl->end() ) {
	    CDecl* newDecl = (*remapDecl)[decl];
	    MASSERT( newDecl->GetClass() == eNET );
	    n->Arg<CNet*>(0) = (CNet*)newDecl; 
	}
	break;
    case eVAR_REF:
	decl = n->Arg<CVar*>(0);
	if( remapDecl->find(decl) != remapDecl->end() ) {
	    CDecl* newDecl = (*remapDecl)[decl];
	    MASSERT( newDecl->GetClass() == eVAR );
	    n->Arg<CVar*>(0) = (CVar*)newDecl; 
	}
	break;
    case ePARAM_REF:
	decl = n->Arg<CParam*>(0);
	if( remapDecl->find(decl) != remapDecl->end() ) {
	    CDecl* newDecl = (*remapDecl)[decl];
	    MASSERT( newDecl->GetType() == ePARAM );
	    n->Arg<CParam*>(0) = (CParam*)newDecl; 
	}
	break;
    default:
	return;
    }
}
/************************************************
	Substitute
	- substitute expressions for decls
 ************************************************/
static CNode* Substitute( CNode* n, void* arg )
{
    map<CDecl*,CNode*>* substitute = (map<CDecl*,CNode*>*)arg;
    CNode* result = n;
    CDecl* decl;
    switch( n->GetOp() ) {
    case eNET_REF:
    case eVAR_REF:
    case ePARAM_REF:
	decl = n->Arg<CDecl*>(0);
	if( substitute->find(decl) != substitute->end() ) {
	    result = (*substitute)[decl]->Clone();
	}
	break;
    default:;
    }
    return result;
}
/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CFlatten::Process( list<CElement>& inputList,
			list<CElement>& outputList ) 
{
    CNode* code = NULL;

    /*
     * merge all input streams together
     */
    list<CElement>::iterator ptr;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
        code = cLINK( code, ptr->Code() );
    }

    /*
     * find top level modules
     */
    set<CModule*> leafModules;
    set<CModule*> allModules;
    set<CModule*> topLevelModules;
    set<CModule*> undefinedModules;

    AnalyseModules( code, 
		    allModules, 
		    topLevelModules,
		    leafModules,
		    undefinedModules,
		    module2Comments );

    set<CModule*>::iterator p;
    for( p = undefinedModules.begin(); p != undefinedModules.end(); ++p ) {
	message( NULL, mMISUD, (*p)->GetName() );
    }	
    for( p = topLevelModules.begin(); p != topLevelModules.end(); ++p ) {
	message( NULL, mTLVMD, (*p)->GetName() );
    }	
    code = ElaborateModules( topLevelModules,
				    NULL, "", NULL, code );
    outputList.push_back( CElement( "", 0, code ) );

}


