/******************************************************************************
 * 	Copyright (c) 2010 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   crectify.cc
 *		- methods for tool to rectify x's out of a design
 *
 ******************************************************************************
 */
						

#include <set>
#include <map>
#include <config.h>
#include "cnode.h"
#include "crectify.h"
#include "main.h"

/**
 * \page rectify_plugin Rectify Plugin
 * The rectify tool edits the parse trees replacing 'x' bits in constants 
 * with a default value of 0 or 1. The default replacement value is supplied by 
 * a command line switch. Selected constructs may be flagged to utilize 
 * non-default value by tagging them with a attribute.
 * - <b>+rectify-default-value={0|1}</b> Default value to replace x
 * - <b>+rectify-attribute-name=<attributeName></b> Attribute name for overriding default value.
 *
 * Attribute override syntax:
 * (* attributeName={0|1} *) statement 
 */

/**
 * \cond PLUGIN_INTERNALS
 */

struct SubXArgs {
    const char* defaultAttribute;
    int  defaultValue;
    int  skip;
}; 
struct CheckConstructsArgs {
    int deleteDeclPragmas;
};
static int SubX( CNode* n, void* arg );
static  Message* mCCBID;
static  Message* mPATOD;

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
#ifndef CYGWIN
extern "C" {
CBackend* CreateToolInstance()
{
    return new CRectify();
}
}
#endif

/************************************************
 	Constructor
 ************************************************/
CRectify::CRectify() 
{
     switches.push_back( "+rectify-default-value={0|1}" );
     switchDescription["+rectify-default-value={0|1}"] = "change default replacement value";
     switches.push_back( "+rectify-attribute-name=<attrName>" );
     switchDescription["+rectify-attribute-name=<attrName>"] = "attribute name used to override default replacement value";
     switches.push_back( "+rectify-delete-decl-pragmas" );
     switchDescription["+rectify-delete-decl-pragmas"] = "delete pragmas attached to declarations";
     mCCBID = Message::RegisterWarning( this, Message::eWARNING,
		"CCBID", "casex cannot be safely instrumented, please use casez" );
     mPATOD = Message::RegisterWarning( this, Message::eWARNING,
		"PATOD", "pragmas attached to declarations cannot be safely handled by vrq" );

}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CRectify::GetToolName( void )
{
	static char	toolName[] = "rectify";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CRectify::GetToolDescription( void )
{
	static char	toolDescription[] = "Replace all X constants with 0 or 1";

	return toolDescription;
}

/************************************************
	SubXCaseItem
	- Find X's and replace them in casex items
 ************************************************/
static int SubXCaseItem( CNode* n, void* arg )
{
    if( !n ) {
        return 0;
    }
    switch( n->GetOp() ) {
    case eCASEITEM:
        n->Arg<CNode*>(1)->PreVisit1( SubX, arg );
        return 0;
    default:
        return 1;
    }

}
/************************************************
	SubX
	- Find X's and replace them
 ************************************************/
static int SubX( CNode* n, void* arg )
{
    SubXArgs *args = reinterpret_cast<SubXArgs*>(arg);
    if( !n ) {
	return 0;
    }
    if( !args->skip && n->HasAttribute( args->defaultAttribute ) ) {
	SubXArgs newArgs = *args;
	newArgs.skip = 1;

        CAttr* attr = n->GetAttribute( args->defaultAttribute );
        MASSERT( attr );
	CNode* expr = attr->GetAttributeExpression( args->defaultAttribute );
	if( expr ) {
	    newArgs.defaultValue = expr->EvalINT32();
	}
	n->PreVisit1( SubX, &newArgs );
	return 0;
    }
    args->skip = 0;

    if( n->GetOp() == eCASEX ) { 
	message( n->GetCoord(), mCCBID );
	CNode* r = n->Arg<CNode*>(1);
	r->PreVisit1( SubXCaseItem, arg );
        return 0;
    }

    if( n->GetOp() == eVCONSTANT && 
	n->Arg<CVector*>(0)->HasX() ) {
	CVector* v = n->Arg<CVector*>(0);
	CUInt aval = v->Aval();
	CUInt bval = v->Bval();
	CUInt mask = aval & bval;
	if( args->defaultValue ) {
	    v->Aval( mask | aval );
	    v->Bval( ~mask & bval );
	} else {
	    v->Aval( ~mask & aval );
	    v->Bval( ~mask & bval );
	}
    }
    return 1;
}
/************************************************
	CheckConstructs
	- find constructs rectify cannot handle
          safely and report/alter them
 ************************************************/
static int CheckConstructs( CNode* n, void* arg )
{
    CheckConstructsArgs *args = reinterpret_cast<CheckConstructsArgs*>(arg);
    if( !n ) {
	return 0;
    }
    switch( n->GetOp() ) {
    case eVAR_DECL:
    case ePARAM_DECL:
    case eSPECPARAM_DECL:
        if( n->Arg<CDecl*>(0)->GetPragmas() ) {
            if( args->deleteDeclPragmas ) {
                n->Arg<CDecl*>(0)->SetPragmas(NULL);
            } else {
	        message( n->GetCoord(), mPATOD );
            }
        }
        break;
    }
    return 1;
}
/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CRectify::Process( list<CElement>& inputList,
   		        list<CElement>& outputList ) 
{
    char* arg = GetPlusArg( "rectify-default-value" );
    if( arg ) {
	// skip over "="
	defaultValue = atoi( arg+1 );
    } else {
	defaultValue = 0;
    }

    arg = GetPlusArg( "rectify-attribute-name" );
    if( arg ) {
	// skip over "="
	defaultAttribute = arg+1;;
    } else {
	defaultAttribute = "RV";
    }
    arg = GetPlusArg( "rectify-delete-decl-pragmas" );
    deleteDeclPragmas = (arg != NULL);


    list<CElement>::iterator ptr;
    set<CNode*> cutList;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	/*
	 * filter each section of code
	 */
	CNode* code = ptr->Code();
        /*
         * Check for constructs that rectify cannot safely handle
         */
        CheckConstructsArgs ccArgs;
        ccArgs.deleteDeclPragmas = deleteDeclPragmas;
        code->PreVisit1( CheckConstructs, &ccArgs );
	/*
 	 * find and substitute for X's
 	 */
	SubXArgs args;
	args.defaultValue = DefaultValue();
	args.defaultAttribute = DefaultAttribute();
	args.skip = 0;
	code->PreVisit1( SubX, &args );
	outputList.push_back( CElement( ptr->Filename(), 
					ptr->Filename()!=NULL, code ) );
    }

}

/**
 * \endcond 
 */

