/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

/******************************************************************************
 *
 *
 *	  analyse.cc
 *	  - class to analyse verilog hierarchy 
 *
 *
 ******************************************************************************
 */

#include "analyse.h"
#include "io.h"

/**************************************************
     Constructor
 *************************************************/
Analyse::Analyse( tree top_level )
  : top_level( top_level )
{
    /*
     * initialize scope to null
     */
    scope = NULL;
    /* 
     * scan each top level module
     */
    for(tree t = top_level; t; t = TREE_CHAIN(t)) {
	AnalyseModule( t );
    }
}
/**************************************************
    PrintDeclName 
    - print decl name with flags
 *************************************************/
void Analyse::PrintDeclName( tree t )
{
    printf_V( "%s{%s%s%s%s%s%s%s%s}",
		IDENT(DECL_NAME(t)), 
		PORT_INPUT_ATTR(t)?"i":"",
		PORT_OUTPUT_ATTR(t)?"o":"",
		PORT_REDEFINED_ATTR(t)?"r":"",
		PORT_MIRRORED_ATTR(t)?"m":"",
                NET_ASSIGN_ATTR(t)?"n":"",
                STMT_SURROGATE_ATTR(t)?"s":"",
		PORT_IMMEDIATE_ATTR(t)?"e":"",
		PORT_COLLAPSED_ATTR(t)?"c":"" );
}

/**************************************************
    PrintAssign
    - print cont assign info
 *************************************************/
void Analyse::PrintAssign( tree assign )
{
    int ltype = *tree_code_type[(int) TREE_CODE(STMT_ASSIGN_LVAL(assign))];
    int rtype = *tree_code_type[(int) TREE_CODE(STMT_ASSIGN_RVAL(assign))];
    printf_V( "AnalyseModule: ASSIGN " ); 
    if( ltype == 'd' ) {
        PrintDeclName(STMT_ASSIGN_LVAL(assign));
    } else {
        printf_V( "???" );
    }
    printf_V( " = " );
    if( rtype == 'd' ) {
        PrintDeclName(STMT_ASSIGN_RVAL(assign));
    } else {
        printf_V( "???" );
    }
    printf_V( "\n" );
}
/**************************************************
    AnalyseModule
    - analyse module and it's children
 *************************************************/
void Analyse::AnalyseModule( tree module )
{
    tree scopeSave = scope;

    scope = module;
    printf_V( "AnalyseModule: module name = %s\n", MODULE_NAME(module) );
    /*
     * analyse declarations
     */
    for (tree t = BLOCK_DECL(module); t; t = TREE_CHAIN(t)) {
	printf_V( "AnalyseModule: DECL %s ", tree_code_name[TREE_CODE(t)] );
	PrintDeclName( t );
	if( TREE_CODE(t) == NET_SCALAR_DECL ||
	    TREE_CODE(t) == NET_VECTOR_DECL ) {
	    printf_V( " <= [ " );
	    for( tree d = NET_SOURCE(t); d; d = NET_SOURCE(d) ) {
	        PrintDeclName(d);
	        printf_V( " " );
	    }
	    printf_V( "]" );
	}
	printf_V( "\n" );
    }
    /*
     * analyse ports
     */
    for (tree t = BLOCK_PORTS(module); t; t = TREE_CHAIN(t)) {
    	printf_V( "AnalyseModule: PORT %s ", tree_code_name[TREE_CODE(t)] );
	PrintDeclName( t );
        if (PORT_REDEFINED_ATTR(t)) {
	    tree port = DECL_THREAD(t);
	    printf_V( " <= [ %s ", tree_code_name[TREE_CODE(port)] );
	    PrintDeclName(port);
	    printf_V( "]" );
	} else if( TREE_CODE(t) == NET_SCALAR_DECL ||
	    TREE_CODE(t) == NET_VECTOR_DECL ) {
	    printf_V( " <= [ " );
	    for( tree d = NET_SOURCE(t); d; d = NET_SOURCE(d) ) {
	        PrintDeclName(d);
	        printf_V( " " );
	    }
	    printf_V( "]" );
	}
	printf_V( "\n" );
	if( PORT_MIRRORED_ATTR(t) ) {
	    tree mirror = DECL_THREAD(t);
    	    printf_V( "AnalyseModule: PORT %s ", tree_code_name[TREE_CODE(mirror)] );
	    PrintDeclName( mirror );
	    printf_V( " <- %s ", tree_code_name[TREE_CODE(t)] );
	    PrintDeclName(t);
	    printf_V( "\n" );
	}
	if( PORT_COLLAPSED_ATTR(t) ) {
	    tree collapsedNet = DECL_COLLAPSED_NET(t);
    	    printf_V( "AnalyseModule: PORT %s ", tree_code_name[TREE_CODE(collapsedNet)] );
	    PrintDeclName( collapsedNet );
	    printf_V( " <* %s ", tree_code_name[TREE_CODE(t)] );
	    PrintDeclName(t);
	    printf_V( "\n" );
	}
   }
   /*
    * analyse continous assignments
    */
    for (tree t = MODULE_ASSIGNMENTS(module); t; t = TREE_CHAIN(t)) {
	tree assign = TREE_PURPOSE(t);
	PrintAssign( assign );
    }
    /*
     * recursively scan each instance
     */
    for(tree t = MODULE_INSTANCES(module); t; t = TREE_CHAIN(t)) {
	if(TREE_CODE(t) != INSTANCE_NODE || UDP_ATTR(t)) {
	    continue;
	}
	printf_V( "AnalyseModule: instance name = %s\n", 
					IDENT(INSTANCE_NAME(t) ));
        for( tree arg = INSTANCE_PORTS(t); arg; arg = TREE_CHAIN(arg)) {
	    if( !TREE_3RD(arg) ) {
	 	continue;
	    }
	    tree n = TREE_3RD(arg);
    	    if( TREE_PURPOSE(n) ) {
		PrintAssign( TREE_PURPOSE(n) );
	    }
    	    if( TREE_VALUE(n) ) {
		PrintAssign( TREE_VALUE(n) );
	    }
	}
	AnalyseModule(INSTANCE_BLOCK(t));
    }
    scope = scopeSave;
}
