/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* IO.C - Compiler-dependent input/output routines */

/* fbufopen "opens" a buffer so that getc will read from it as if it
   were a file */

#define IO_C

#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include "io.h"
#include "veriwellinterface.h"

namespace veriwell {
File *fin;			/* current input file */
const char *input_filename;
const char *filename;

int global_print_override;
char *global_print_buffer;
char *global_print_p;

FILE *log_file;
FILE *key_file;
char *log_file_name;
char *key_file_name;
int log_enable = 1;
int log_available = 0;
int key_enable = 1;
int key_available = 0;

LibPath_t *ypathList;
char *ylibext;

FILE *file_handles[NHANDLES];
int file_used[NHANDLES];
char print_buffer[2000];

/* Print a fatal error message.  NAME is the text.
 *    Also include a system error message based on `errno'.  */


/* Same as `malloc' but report error if no memory available.  */

char *xmalloc(unsigned size)
{
    char *value;
    value = (char *) malloc(size);
    if (value == 0) {
        fatal("Memory exhausted.", NULL);
    }
    return value;
}

/* Same as `malloc' but aligned to cache line and
   report error if no memory available.  
   NOTE: must not use normal free on this
*/


char *xmalloc_aligned(unsigned size)
{
    char *value;
    value = (char *) malloc(size+cacheLineSize);
    if (value == 0) {
        fatal("Memory exhausted.", NULL);
    }
    int offset = (size_t)value % cacheLineSize;
    if( offset ) {
	value += cacheLineSize-offset;
    }
    return value;
}

/* Same as `realloc' but report error if no memory available.  */

char *xrealloc(char *ptr, int size)
{
    char *result = (char *) realloc(ptr, size);
    if (!result) {
        fatal("Memory exhausted.",  NULL);
    }
    return result;
}
void fatal(const char *s, const char *v)
{
    error(s, v, (char *) NULL);
    simulator.Exit(1);
}

void runtime_error(tree node)
{
    input_filename = STMT_SOURCE_FILE(node);
    lineno = STMT_SOURCE_LINE(node);
}


void fclose_V(handle_t handle)
{
    int i;
    handle_t tmp_handle;

    if (handle & 1)
	warning("Cannot close standard output", NULL_CHAR, NULL_CHAR);

    if (handle - 1) {
	tmp_handle = handle >> 1;
	for (i = 0; i < NHANDLES; i++)
	    if (tmp_handle >> i & 1) {
		fclose(file_handles[i]);
		file_used[i] = 0;
	    }
    }
}

/* Open a file, assign one of 31 bits to the handle of the file */

handle_t fopen_V(const char *file)
{
    int i;
    FILE *f_handle;

    /* find an available handle */
    for (i = 0; i < NHANDLES; i++)
	if (!file_used[i])
	    break;
    if (i >= NHANDLES) {
	warning("Exeeded limit on number of open files", NULL_CHAR,
		NULL_CHAR);
	return 0;
    }
    f_handle = fopen(file, "w+t");
    if (!f_handle) {
	warning("Cannot open file '%s'", file, NULL_CHAR);
	return 0;
    }
    file_handles[i] = f_handle;
    file_used[i] = 1;
    return 1 << (i + 1);
}

/* Given a handle, test to make sure each handle is valid; filter
   out invalid handles and give warnings if necessary */

handle_t check_handle(handle_t handle)
{
    int i;
    handle_t tmp_handle;

    /* Only check for file handles */
    if (handle - 1) {
	tmp_handle = handle >> 1;
	for (i = 0; i < NHANDLES; i++)
	    if (tmp_handle >> i & 1)
		if (!file_used[i]) {
		    warning("File not open", NULL_CHAR, NULL_CHAR);
		    handle &= ~((handle_t) 1 << (i + 1));
		}
    }
    return handle;
}

void fflush_V(handle_t handle)
{
    int i;
    handle_t tmp_handle;

    /* First, flush stdout (and log file) */
    if (handle & 1) {
	fflush(stdout);
	if (log_enable)
	    fflush(log_file);
    }
    /* If there is some file output, flush it */
    if (handle - 1) {
	tmp_handle = handle >> 1;
	for (i = 0; i < NHANDLES; i++)
	    if (tmp_handle >> i & 1)
		if (file_used[i])
		    fflush(file_handles[i]);
		else
		    warning("File not open", NULL_CHAR, NULL_CHAR);
    }
}

void vfprintf_V(handle_t handle, const char *fmt, va_list args)
{
    int i;
    handle_t tmp_handle;

    if (global_print_override) {
	/* Watcom chokes on this... */
	global_print_p += vsprintf(global_print_p, fmt, args);
	return;
    }

    /* Do this because some compilers (and ANSI?) don't like to use "args"
       more than once */

    vsprintf(print_buffer, fmt, args);

    /* First, do output to stdout (and to log file) */
    if (handle & 1) {
	printf("%s", print_buffer);
	if (log_enable)
	    fprintf(log_file, "%s", print_buffer);
    }
    /* If there is some file output to do, do it */
    if (handle - 1) {
	tmp_handle = handle >> 1;
	for (i = 0; i < NHANDLES; i++)
	    if (tmp_handle >> i & 1)
		if (file_used[i])
		    fprintf(file_handles[i], "%s", print_buffer);
		else
		    warning("File not open", NULL_CHAR, NULL_CHAR);
    }
}

void fprintf_V(handle_t handle, const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vfprintf_V(handle, fmt, args);
    va_end(args);
}

void vprintf_V(const char *fmt, va_list args)
{
    if (global_print_override) {
	global_print_p += vsprintf(global_print_p, fmt, args);
	return;
    } else
	/* Do this because some compilers (and ANSI?) don't like to use "args"
	   more than once */

	vsprintf(print_buffer, fmt, args);

    printf("%s", print_buffer);

    if (log_enable)
	fprintf(log_file, "%s", print_buffer);
}

void printf_V(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vprintf_V(fmt, args);
    va_end(args);
}

void vprintf_V_log(const char *fmt, va_list args)
{

    vsprintf(print_buffer, fmt, args);

    if (log_enable)
	fprintf(log_file, "%s", print_buffer);
}

void printf_V_log(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vprintf_V_log(fmt, args);
    va_end(args);
}

void printf_error_V(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);
    if (log_enable) {
	va_start(args, fmt);
	vfprintf(log_file, fmt, args);
	va_end(args);
    }
}

void printf_V_key(const char *fmt, ...)
{
    va_list args;

    if (!fin->isatty()) {
	va_start(args, fmt);
	vprintf(fmt, args);
	va_end(args);
    }
    if (log_enable) {
	va_start(args, fmt);
	vfprintf(log_file, fmt, args);
	va_end(args);
    }
    if (key_enable) {
	va_start(args, fmt);
	vfprintf(key_file, fmt, args);
	va_end(args);
    }
}

/* Report an error at line LINE of file FILE.
   S and V are a string and an arg for `printf'.  */

void
error_with_file_and_line(const char *file,
                         lineno_t line, const char *s, const char *v, const char *v2)
{
    simulator.CountError(0);
    if (is_interactive) {
        printf_error_V("C%d: ", cmd_level);
        new_cmd_level = cmd_level;
    } else if (file) {
        printf_error_V("%s: L%ld: ", file, line);
    } else {
        printf_error_V("VeriWell: ");
    }

    printf_error_V("error: ");
    printf_error_V(s, v, v2);
    printf_error_V("\n");
}

/* Report an error at the current line number.
   S and V are a string and an arg for `printf'.  */

void error(const char *s, const char *v, const char *v2)
{
    error_with_file_and_line(input_filename, lineno, s, v, v2);
}
/* Report a warning at line LINE.
   S and V are a string and an arg for `printf'.  */

void
warning_with_file_and_line(const char *file, lineno_t line, const char *s, const char *v,
                           const char *v2)
{
    if (simulator.CountError(1) == 0) {
        return;
    }


    if (file) {
        printf_error_V("%s: L%ld: ", file, line);
    } else {
        printf_error_V("VeriWell: ");
    }
    printf_error_V("warning: ");
    printf_error_V(s, v, v2);
    printf_error_V("\n");
}

/* Report a warning at the current line number.
   S and V are a string and an arg for `printf'.  */

void warning(const char *s, const char *v, const char *v2)
{
    warning_with_file_and_line(input_filename, lineno, s, v, v2);
}
/* Report a warning at line LINE.
   S and V are a string and an arg for `printf'.  */

void
sorry_with_file_and_line(const char *file, lineno_t line,
                         const char *s, const char *v, const char *v2)
{
    if (simulator.CountError(0) == 0) {
        return;
    }


    if (file) {
        printf_error_V("%s: L%ld: ", file, line);
    } else {
        printf_error_V("VeriWell: ");
    }
    printf_error_V("sorry: ");
    printf_error_V(s, v, v2);
    printf_error_V("\n");
}

/* Report a warning at the current line number.
   S and V are a string and an arg for `printf'.  */

void sorry(const char *s)
{
    sorry_with_file_and_line(input_filename, lineno, s, NULL_CHAR,
                             NULL_CHAR);
}

void sorry1(const char *s, const char *v)
{
    sorry_with_file_and_line(input_filename, lineno, s, v, NULL_CHAR);
}


void init_io()
{
    log_enable = 0;
    log_available = 0;
    key_enable = 0;
    key_available = 0;
    global_print_override = 0;
    global_print_buffer = 0;
    global_print_p = 0;
    ypathList = (LibPath_t *) xmalloc(sizeof(LibPath_t));
    ypathList->path = ".";
    ypathList->next = NULL;

    memset(&fin, 0, sizeof(fin));
    memset(&input_filename, 0, sizeof(input_filename));
    memset(&filename, 0, sizeof(filename));
    memset(&log_file, 0, sizeof(log_file));
    memset(&key_file, 0, sizeof(key_file));
    memset(&log_file_name, 0, sizeof(log_file_name));
    memset(&file_handles, 0, sizeof(file_handles));
    memset(&file_used, 0, sizeof(file_used));
    memset(&ylibext, 0, sizeof(ylibext));


}
 
}

