/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* SCHED.C - Scheduler and queues for simulation engine */

#define SCHED_C

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stack>
#include <map>
#include "vtypes.h"
#include "tree.h"
#include "schedule.h"
#include "flags.h"
#include "runtime.h"
#include "dumpvar.h"
#include "strobe.h"
#include "acc_user.h"
#include "gates.h"
#include "eval.h"
#include "exec.h"
#include "nsched.h"
#include "pli.h"
#include "systask.h"
#include "stats.h"
#include "veriwellinterface.h"

using std::stack;
using std::map;

namespace veriwell {

SCB *netlist = NULL;		/* List of SCBs containing pending net changes */
SCB *netlist_last = NULL;	/* Points to last SCB on net list */
SCB* SCB::readylist = NULL;		/* List of SCBs previously waiting for an event */
SCB* SCB::readylist_last = NULL;	/* Points to last SCB on ready list */
SCB *freelist = NULL;		/* List of available SCBs (eg. fork/join) */
SCB *retrigger = NULL;          /* set if scb was retrigger while active */

tree deferred_markers = NULL_TREE;

time64 CurrentTime;

static int BLK_SIZE = 1024;
extern unsigned long long int tscDelta;

SCB* TimeList::timelist;
unsigned long long TimeList::removeCycles;
unsigned long long TimeList::removeCount;
unsigned long long TimeList::removeGroupCycles;
unsigned long long TimeList::removeGroupCount;
unsigned long long TimeList::insertCycles;
unsigned long long TimeList::insertCount;

/************************************************************
	SchedulerHistogram
	- histogram class for tracking scheduler
 ***********************************************************/
class SchedulerHistogram : public Stats {
    static const int MAX_LABEL=100;
    long long int notifyCount;
    long long int markerCount;
    long long int fixedMarkerCount;
    long long int dispatcherCount[MAX_LABEL+1];
    long long int dispatcherCycles[MAX_LABEL+1];
    static const int MAX_FLAGS = 0x4000;
    struct HistEntry {
	unsigned flags;
	unsigned long long total;
	unsigned long long cycles;
    } entryHistogram[MAX_FLAGS];
public:
    SchedulerHistogram();
    virtual const char* Name() { return "scheduler"; }
    virtual unsigned long long TotalCycles();
    virtual void Dump( unsigned long long totalTicks );
    void Analyse( Marker* marker );
    void Cycles( marker_flags flags, unsigned long long time );
    void Dispatcher( int label, unsigned long long time );
    static int HistCompare( const void* p1, const void* p2 );
};
static SchedulerHistogram* histogram;
TimeList timelist;


SchedulerHistogram::SchedulerHistogram()
{
    memset( entryHistogram, 0, sizeof(entryHistogram) );
    for( int i = 0; i <= MAX_FLAGS; ++i ) {
	entryHistogram[i].flags = i;
    }
    for( int i = 0; i <= MAX_LABEL; ++i ) {
	dispatcherCount[i] = 0;
	dispatcherCycles[i] = 0;
    }
    notifyCount = 0;
    markerCount = 0;
    fixedMarkerCount = 0;
    masterStats.Add(this);
}

void SchedulerHistogram::Analyse( Marker* marker )
{
    notifyCount++;
    while( marker ) {
	markerCount++;
	if( marker->flags & M_FIXED ) {
	    fixedMarkerCount++;
	}
	marker = marker->next;
    }
}
void SchedulerHistogram::Cycles( marker_flags flags, unsigned long long cycles ){
    ASSERT( flags <= MAX_FLAGS );
    entryHistogram[flags].total++;
    entryHistogram[flags].cycles += cycles;
}
void SchedulerHistogram::Dispatcher( int label, unsigned long long time )
{
    dispatcherCount[label]++;
    dispatcherCycles[label] += time;
}

int SchedulerHistogram::HistCompare( const void* p1, const void* p2 ) 
{
    const HistEntry* e1 = static_cast<const HistEntry*>(p1);
    const HistEntry* e2 = static_cast<const HistEntry*>(p2);
    if( e1->cycles == e2->cycles ) {
	return 0;
    } else if( e1->cycles < e2->cycles ) {
	return 1;
    } else {
	return -1;
    }
}

unsigned long long SchedulerHistogram::TotalCycles()
{
    unsigned long long totalCycles = 0;
    for( int i = 0; i < MAX_FLAGS; ++i ) {
	totalCycles += entryHistogram[i].cycles;
    }
    for( int i = 0; i <= MAX_LABEL; ++i ) {
	totalCycles += dispatcherCycles[i];
    }
    return totalCycles;
}

void SchedulerHistogram::Dump( unsigned long long totalTicks )
{
    if( !printStats ) {
	return;
    }

    qsort( entryHistogram, NUM_TREE_CODES, sizeof(HistEntry), HistCompare );
    printf_V( "\nScheduler Histogram:\n" );
    printf_V( "%-12s: %20s %20s %20s %5s\n", "flags", "total", "total cycles", "avg cycles", "% " );
    printf_V( "------------------------------------------------------------------------\n" );
    for( int i = 0; i < MAX_FLAGS; ++i ) {
	unsigned flags = entryHistogram[i].flags;
	long long int total = entryHistogram[i].total;
	long long int cycles = entryHistogram[i].cycles;
	if( total == 0 ) {
	     continue;
	}
	printf_V( "%-12x: %20lld %20lld %20lld %5.2f    ", 
		flags, total, cycles, cycles/total, (double)cycles*100./totalTicks );
	unsigned v = flags;
	unsigned p = 1;
	while( v ) {
	    if( v&1 ) {
		printf_V( "%s ", MarkerFlagName( (enum marker_flags)p ) );
	    }
	    v = v >> 1;
	    p = p << 1;
	}
	printf_V( "\n" );
	
    }
    printf_V( "\nDispatcher Stats:\n" );
    printf_V( "%-8s: %20s %20s %20s %5s\n", "label", "total", "total cycles", "avg cycles", "%" );
    printf_V( "------------------\n" );
    unsigned long long totalCount = 0;
    unsigned long long totalCycles = 0;
    for( int i = 0; i <= MAX_LABEL; ++i ) {
	totalCount += dispatcherCount[i];
	totalCycles += dispatcherCycles[i];
	if( dispatcherCount[i] ) {
            printf_V( "%-8d %20lld %20lld %20lld %5.2f\n", 
		i, dispatcherCount[i], dispatcherCycles[i], 
	        dispatcherCycles[i]/dispatcherCount[i], (double)dispatcherCycles[i]*100./totalTicks );
	}
    }
    if( totalCount ) {
        printf_V( "%-8s %20lld %20lld %20lld\n", 
		"total", totalCount, totalCycles, 
	        totalCycles/totalCount );
    }
    printf_V( "\nScheduler Stats:\n" );
    printf_V( "------------------\n" );
    printf_V( "Notify events: %lld\n", notifyCount );
    printf_V( "Marker records: %lld\n", markerCount );
    printf_V( "FIXED Marker records: %lld\n", fixedMarkerCount );
    printf_V( "Avg marker chain length: %lld\n", markerCount/notifyCount );
}

void DumpTimeListStats() { timelist.Dump(); }

SCB* SCB::BuildSCB(union tree_node *start, enum which_list list)
{
    SCB *scb, *tmp, *s;
    int i;
    extern tree current_scope;

    if (!freelist) {
	SCB *scbArray = (SCB *) xmalloc(sizeof(SCB) * BLK_SIZE);
	for (i = 0; i < BLK_SIZE; i++) {
	    ADD_LIST_SCB(FREE_LIST, &freelist, scbArray + i);
	}
    }
    scb = freelist;
    REMOVE_LIST_SCB(scb);

    scb->pc = start;
    ASSIGNTIME64(&(scb->time), &CurrentTime);
    scb->prev = NULL;
    scb->next = NULL;
    scb->fork_count = 0;
    scb->fork = NULL;
    scb->here.scb = NULL;
    scb->mode = RUN;
    scb->context = NULL;
    enter_context(scb, current_scope, NULL_TREE);

    switch (list) {
    case TIME_LIST:
	ASSERT( FALSE );
	break;
    case READY_LIST:
	if (!readylist) {
	    ADD_LIST_SCB(READY_LIST, &readylist, scb);
	} else {
	    /* put new entries at end of list */
	    for (s = readylist; s->next; s = s->next);
	    ADD_LIST_SCB(READY_LIST, &s->next, scb);
	}
	break;
    case EVENT_LIST:
	ADD_LIST_SCB(EVENT_LIST, NULL, scb);
	break;	
    case NOLIST:
	ADD_LIST_SCB(NOLIST, NULL, scb);
	break;
    default:
    case FREE_LIST:
	ASSERT( FALSE );
	break;
    }

    return scb;
}

void PrintSCB(SCB * scb)
{
    printf("building SCB at %p\n", scb);
    printf("  scb->prev: %p\n", scb->prev);
    printf("  scb->list: %d\n", scb->list);
}

void BuildMarker(tree decl, struct Marker_info *marker_info)
{
    Marker *marker;

    /*
     * If decl is collapsed, resolve links to find collapsed net 
     */
    while( PORT_COLLAPSED_ATTR(decl) ) {
	decl = DECL_COLLAPSED_NET(decl);
    }

    /* First, ensure that only one marker per variable is used e.g. a = b + b; */
    for (marker = marker_info->first; marker; marker = marker->link)
	if (marker->decl == decl)
	    return;

    marker = (Marker *) xmalloc_aligned(sizeof(Marker));
    marker->next = NULL;
    marker->prev = NULL;
    marker->scb = marker_info->current_scb;
    marker->expr.expr = NULL;
    marker->decl = decl;
    marker->flags = marker_info->flags;
    marker->delay = marker_info->delay;
    marker->link = NULL;
    marker->tmp1 = NULL;
    marker->tail = NULL;
    marker->back = NULL;

    if (marker_info->last)
	marker_info->last->link = marker;	/* Link this to end of the list */
    else
	marker_info->first = marker;

    marker_info->last = marker;

    if ((marker->flags & M_FIXED))
	WaitOnEvent(marker, marker_info->current_scb);
    return;
}

tree WaitOnTime(delay_t waittime, tree pc)
{
    SCB *scb = SCB::ActiveSCB();

    scb->pc = pc;		/* save next instruction to be executed */
    ADDTIME6432(&(scb->time), &CurrentTime, waittime);
    scb->mode = WAIT;

    scb = SCB::dispatcher(TIME_LIST);
    return scb->pc;		/* as wll as the new pc */
}

tree WaitOnTime64(time64 * waittime, tree pc)
{
    SCB *scb = SCB::ActiveSCB();

    scb->pc = pc;		/* save next instruction to be executed */
    ADDTIME6464(&(scb->time), &CurrentTime, waittime);
    scb->mode = WAIT;

    scb = SCB::dispatcher(TIME_LIST);
    return scb->pc;		/* as wll as the new pc */
}

void Schedule64(time64 *time, SCB * scb, int prepend)
{
    SCB *link, *tmp;

    scb->time = *time;
    scb->mode = WAIT;

    REMOVE_LIST_SCB(scb);
    TimeList::Insert( scb, prepend );

}

void ScheduleDelta64(time64 *time, SCB * scb, int prepend)
{
    SCB *link, *tmp;

    ADDTIME6464(&(scb->time), &CurrentTime, time);
    scb->mode = WAIT;

    REMOVE_LIST_SCB(scb);
    TimeList::Insert( scb, prepend );

}

void Schedule(delay_t waittime, SCB * scb, int prepend)
{
    SCB *link, *tmp;

    ADDTIME6432(&(scb->time), &CurrentTime, waittime);
    scb->mode = WAIT;

    REMOVE_LIST_SCB(scb);
    TimeList::Insert( scb, prepend );
}

/* Schedule nets */

void ScheduleNet(Marker * marker, tree decl)
{
    SCB *scb = marker->scb;

    if( NET_DELAY(scb->pc) ) {
	logical_value state = DeclState(decl);
	delay_t delay = eval_delay(NET_DELAY(scb->pc), state);
	if( delay ) {
	    Schedule(delay, scb, 0);
	}
	return;
    }

/* If there is no delay, schedule on net list, else put on time list.
   Note that in the case of inouts, we could be working with our own
   net in which case, we're finished. */

    if (scb->list != NET_LIST) {
        if (scb == SCB::ActiveSCB()) {
            retrigger = scb;
	} else {
	    REMOVE_LIST_SCB(scb);
	    if( netlist_last ) {
	        netlist_last->next = scb;
	        scb->prev = &netlist_last->next;
	    } else {
	        netlist = scb;
	        scb->prev = &netlist;
	    }
	    netlist_last = scb;
	    scb->next = NULL;
	    scb->list = NET_LIST;
	    scb->time = CurrentTime;
	}
    }
}

void thread_marker(Marker * marker)
{
    Marker *mark, *tmp;
    tree decl = marker->decl;

    AddMarkerToListTail( DECL_EVENT_CHAIN_HEAD(decl), DECL_EVENT_CHAIN_TAIL(decl), marker );
}

/* Start with a marker and thread on all markers on the link chain */

void thread_all_markers(Marker * marker)
{
    Marker *mark;

    for (mark = marker; mark->link != marker; mark = mark->link)
	thread_marker(mark);
    thread_marker(mark);
}

/* Thread marker onto an event chain, headed by a decl node */

void WaitOnEvent(Marker * marker, SCB * scb)
{
    if (marker->prev) {
	ASSERT(scb != 0);

	/* In a task, it is possible that more than one caller can be waiting on
	   the same event.  If so, add the SCB to the list. */
	if (marker->scb != scb) {
	    scb->here.scb = marker->scb;
	    marker->scb = scb;
	}
	return;
    }

    marker->scb = scb;
    thread_marker(marker);
}

tree WaitOnEventAll(Marker * marker, tree pc)
{
    Marker *mark;
    SCB *scb;

    SCB* current = SCB::ActiveSCB();    
    current->pc = pc;
    current->mode = WAIT;

    // pathological case: wait (0); -- wait forever, no markers, skip past Waits
    if (marker)
	 {
	for (mark = marker; mark->link != marker; mark = mark->link)
	    WaitOnEvent(mark, current);
	WaitOnEvent(mark, current);
	}
    scb = SCB::dispatcher(EVENT_LIST);
    return scb->pc;
}

void WaitOnEventOther(SCB * scb, Marker * marker, tree pc)
{
    Marker *mark;
    SCB *tmp;

    scb->pc = pc;
    scb->mode = WAIT;

    for (mark = marker; mark->link != marker; mark = mark->link)
	WaitOnEvent(mark, scb);
    WaitOnEvent(mark, scb);

    REMOVE_LIST_SCB(scb);
    ADD_LIST_SCB(EVENT_LIST, NULL, scb);
    return;
}

/* Notify everything on the deferred list (used for lval concats) */

void NotifyDeferred(tree deferred_markers)
{
    tree tmp;

    while (deferred_markers) {
	NotifyEvent((Marker *) TREE_PURPOSE(deferred_markers),
		    TREE_VALUE(deferred_markers), (int) 0);

	tmp = deferred_markers;
	deferred_markers = TREE_CHAIN(deferred_markers);
	free_link_list(tmp);
    }
}

/* Place marker's SCB onto ready list; unthread event chain if not fixed */

void NotifyEvent(Marker * marker,  tree decl, int deferred)
{
    Marker *ring, *NextEntry;
    SCB *scb, *tmp, *next_scb;

    if (!marker)
	return;

    if (deferred) {
	deferred_markers = link_tree_cons((tree) marker, decl,
					  deferred_markers);
	return;
    }
#ifdef BUILD_HISTOGRAMS
    histogram->Analyse( marker );
#endif

    do {
#ifdef BUILD_HISTOGRAMS
       unsigned long long int tscStart = rdtsc();
       unsigned long long int tscDeltaSave = tscDelta;
       tscDelta = tscStart;
       enum marker_flags tscFlags = marker->flags;
#endif

	NextEntry = marker->next;
	scb = marker->scb;

	/* If this is a continuous assignment marker, do it now */

	if ((marker->delay)) {	/* if cont assignment, put at beginning */
	    if (in_initial) {
		do_net_assignment(scb->pc);
	    } else {
		logical_value state = DeclState(decl);
		Schedule(eval_delay(marker->delay, state), scb,
			 marker->flags & M_CONT);
	   }
	} else if ((marker->flags & M_CONT)) {
	    if( marker->flags & M_OPTIMIZED ) {
	        do_net_assignment_optimized(marker);
            } else {
	        do_net_assignment(marker);
	    }
	}

	/* If there is a ref to a net (from a cont assignment), schedule it
	   or do it now if we are in initialization or if we are a port. */

	else if (marker->flags & M_NET) {
	    if( marker->flags & M_OPTIMIZED ) {
	        ScheduleNet_fast(marker);
	    } else if( marker->flags & M_IMMED ) {
	        ScheduleNetImmediate(marker);
	    } else {
	        ScheduleNet(marker, decl);
	    }
	}

	/* If this is a primitive (built-in or UDP), call the processing routine */

	else if ((marker->flags & M_PRIM)) {
	    ASSERT(GATE_ROUTINE((tree) marker->scb));
	    (*GATE_ROUTINE((tree) marker->scb)) (marker);
	}

	else if (marker->flags & M_VCD)
	    dumpvars_dump(marker->decl, marker->expr.vcd_id);

	else if (marker->flags & M_IDLE)
	    tickle_monitor(&idle_strobe_queue, marker->expr.tree);	/* in systask.c; points to mon info */

	else if (marker->flags & M_STROBE)	/* Special for $monitor */
	    tickle_monitor_old(&final_strobe_queue, marker->expr.mon);	/* in systask.c; points to mon info */

	else if (marker->flags & M_VCL)
	    vcl_dispatch((struct pli_vcl *) marker->expr.vcl);

	else if (marker->flags & M_ASYNCH) {
	    pli_asynccall((PliInfo_t *) marker->expr.tree);
	}

/* If marker says test the expression (for edge triggered), then do
   so.  But, we may already be on the ready list, so we don't need to
   do the threading of the SCB.  */

	else if (!scb->OnReadyList() &&
		 (!(marker->flags & M_TEST) || test1(marker->expr.expr))) {
	    for (; scb; scb = next_scb) {
		next_scb = scb->here.scb;
		scb->here.scb = NULL;

		REMOVE_LIST_SCB(scb);

		SCB::ReadyListPushBack( scb );
	    }

/* Now, unthread our marker and all the other markers associated with
   this event, if the fixed bit is not set, of course.  Note that the
   markers are threaded by the links in a circular list.  We know to
   stop when we hit a marker that has its prev pointer set to NULL. */

	    if (!(marker->flags & M_FIXED))
		for (ring = marker; ring->prev; ring = ring->link) {
                    RemoveMarkerFromList( ring ) ;
		    ring->prev = NULL;	/* This tells us that the marker is not threaded */
		}
	}

	/* else if... */
	marker = NextEntry;
#ifdef BUILD_HISTOGRAMS
       unsigned long long int tscStop = rdtsc();
       unsigned long long int delta = tscStop-tscDelta;
       histogram->Cycles( tscFlags, delta );
       tscDelta = tscDeltaSave+tscStop-tscStart;
#endif
    }
    while (marker);
}

/* Disconnect the markers associate with this event or wait stmt; usually
     called from disable.  Watch for cases where marker was already notified.
*/

void event_undo(Marker * first)
{
    Marker *mark;

    if (!first)
	return;

    for (mark = first; mark->link != NULL && mark->link != first;
	 mark = mark->link) {
	if (mark->prev) {	/* if null, marker was already notified */
            RemoveMarkerFromList( mark );
	    mark->prev = NULL;
	}

    }				/* handle the last one */
    if (mark->prev) {
            RemoveMarkerFromList( mark ) ;
	mark->prev = NULL;
    }
}

/* Get the next process on the ready list, return a pointer to an
   executable tree; place the incoming SCB onto the wait list (possibly
   ordered by time). */

#ifdef BUILD_HISTOGRAMS
    #define DISPATCHER_BEGIN tscDeltaSave=tscDelta;tscStart=rdtsc();tscDelta=tscStart
    #define DISPATCHER_END(L) tscStop=rdtsc();delta = tscStop-tscDelta;histogram->Dispatcher(L,delta); delta=tscStop-tscStart;tscDelta=tscDeltaSave+delta
#else
    #define DISPATCHER_BEGIN
    #define DISPATCHER_END(LABEL)
#endif

SCB* SCB::dispatcher(enum which_list remove_list)
{
#ifdef BUILD_HISTOGRAMS
    unsigned long long int tscStart;
    unsigned long long int tscDeltaSave;
    unsigned long long int tscStop;
    unsigned long long int delta;
#endif

    DISPATCHER_BEGIN;

    SCB *link, *scb = readylist, *tmp;
    time64 new_time;

    switch (remove_list) {

    case NOLIST:
	REMOVE_LIST_SCB(scb);
	ADD_LIST_SCB(NOLIST, NULL, scb);
	break;

    case EVENT_LIST:
	REMOVE_LIST_SCB(scb);
	ADD_LIST_SCB(EVENT_LIST, NULL, scb);
	break;

    case TIME_LIST:
	REMOVE_LIST_SCB(scb);
	TimeList::Insert( scb, 0 );
	break;

    case FREE_LIST:
	REMOVE_LIST_SCB(scb);
	ADD_LIST_SCB(FREE_LIST, &freelist, scb);
	if (scb->context)
	    unwind_context(scb->context);
	break;

    case NOSCB:
	break;

    case READY_LIST:
    case LOCAL_LIST:
    default:
	ASSERT(FALSE);
    }

  /* 
   * Ready List Check: if we have something on the Ready list - go with it
   */
  go:
    if (readylist) {
        DISPATCHER_END(0);
	return readylist;
    }
    /* 
     * Nothing in the ready list, see if any gates need to
     * be processed. 
     */

    if( PeekGate() ){
	tree gate;

	while (TRUE) {
	    gate = RemoveNextReadyGate();
	    if (gate == NULL_TREE) {
		break;
	    }
	    DEBUGPRINT("Inside gatelist\n");
	    handle_gate(gate);
	}
        if( readylist ) {
            DISPATCHER_END(1);
	    return readylist;
	}
    }

    /* 
     * See if there is a net change pending 
     */
    if (netlist) {
        if (!readylist) {
	    SPLICE_LIST_SCB(READY_LIST, &readylist, netlist);
	} else {
	    SPLICE_LIST_SCB(READY_LIST, &readylist_last->next, netlist);
	}
	netlist = NULL;
	netlist_last = NULL;
        DISPATCHER_END(2);
	return readylist;
    }

/* Nothing on the ready list; check the time list for SCBs that
   are waiting for time to pass. */

    if (scb = TimeList::Peek()) {
	if((CMPTIME64(&(scb->time), &CurrentTime) == 0)) {
	    scb = TimeList::PopGroup();
	    if (!readylist) {
		SPLICE_LIST_SCB(READY_LIST, &readylist, scb);
	    } else {
		SPLICE_LIST_SCB(READY_LIST, &readylist_last->next, scb);
	    }
            DISPATCHER_END(3);
	    return (readylist);
	}

/* Before moving stuff on the readylist, check for things that want to happen
   at the end of the current time unit. */

	check_strobe(&idle_strobe_queue);
	if (readylist) {		/* check_strobe may have indirectly caused new events */
            DISPATCHER_END(4);
	    return readylist;
	}

	if((scb = TimeList::Peek())
	       && (CMPTIME64(&(scb->time), &CurrentTime) == 0)) {
	    scb = TimeList::PopGroup();
	    if (!readylist) {
		SPLICE_LIST_SCB(READY_LIST, &readylist, scb);
	    } else {
		SPLICE_LIST_SCB(READY_LIST, &readylist_last->next, scb);
	    }
            DISPATCHER_END(5);
	    return (readylist);
	}


	if (netlist) {
	    goto go;
	}

	if (PeekGate() && IsGateReady()) {
	    goto go;
	}
	if (!empty_strobe(&idle_strobe_queue)) {
	    goto go;
	}

	check_strobe(&final_strobe_queue);
	ASSERT(readylist == NULL);

	if (PeekGate() && GateConditionalAdvanceTime(&(TimeList::Peek()->time))) {
	    goto go;
	}

	ASSIGNTIME64(&new_time, &(TimeList::Peek()->time));
        dumpvar.new_time = 1;
        scb = TimeList::PopGroup();
        if (!readylist) {
            SPLICE_LIST_SCB(READY_LIST, &readylist, scb);
        } else {
            SPLICE_LIST_SCB(READY_LIST, &readylist_last->next, scb);
        }

/* Nothing on Ready list or Time list, exit. */
    } else {
	extern SCB *finish_scb;
	check_strobe(&idle_strobe_queue);
	/*
	 * there is a chance that pli code has added some 
	 * events. If so continue on.
	 */
	if (TimeList::Peek() != NULL || PeekGate() != NULL) {
	    goto go;
	}
	if (!empty_strobe(&idle_strobe_queue)) {
	    goto go;
	}
	check_strobe(&final_strobe_queue);
        DISPATCHER_END(6);
	return finish_scb;
    }
    DISPATCHER_END(7);
    return readylist;
}

tree dispatch_pc(enum which_list list)
{
    return SCB::dispatcher(list)->pc;
}

/* Note that neither dump_current_process, nor dispatch, must change the
   current value of the evaluation stack. */

tree dump_current_process(void)
{
    SCB *scb = SCB::dispatcher(FREE_LIST);
    return scb->pc;
}

void interject_readylist(SCB * scb, tree pc)
{
    SCB *current = SCB::ActiveSCB();

    if (current) {
	current->pc = pc;
    }

    REMOVE_LIST_SCB(scb);
    SCB::ReadyListPushFront( scb );
}

void FreeSCB(SCB * scb)
{
    SCB *tmp;

    REMOVE_LIST_SCB(scb);
    ADD_LIST_SCB(FREE_LIST, &freelist, scb);
    if (scb->context)
	unwind_context(scb->context);
}


/*****************************************************************
 *	
 *	init_sched
 *		- initialize global and static variables for sched.c
 *
 *****************************************************************
 */

void init_sched()
{
    netlist = NULL;		/* List of SCBs containing pending net changes */
    netlist_last = NULL;	/* Points to last SCB on net list */
    freelist = NULL;		/* List of available SCBs (eg. fork/join) */
    deferred_markers = NULL_TREE;
    SET_TIME(&CurrentTime,0,0);
#ifdef BUILD_HISTOGRAMS
    histogram = new SchedulerHistogram;
#endif
}

}
