/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* SCHED.H - Declarations for scheduler and queues */

#ifndef SCHED_H
#define SCHED_H

#include <stddef.h> 
#include "flags.h"
#include "io.h"

// public methods

namespace veriwell {

struct monitor_info;

#define ADD_LIST(POINT, NEW)			\
  {tmp = POINT;						\
  POINT = NEW;						\
  (NEW)->prev = &(POINT);				\
  (NEW)->next = tmp;					\
  if (tmp) tmp->prev = &(NEW)->next;}

#define REMOVE_LIST(PTR)			\
	{if ((*((PTR)->prev) = (PTR)->next) != 0)	\
	  (PTR)->next->prev = (PTR)->prev;}


//enum { SmallCharge, MediumCharge, LargeCharge };
enum which_list { NOSCB = 0, NOLIST,
	READY_LIST, EVENT_LIST, TIME_LIST, NET_LIST, FREE_LIST, LOCAL_LIST };
enum marker_flags {
    M_OPTIMIZED = 0x2000,	/* used for dynamic optimization */
    M_FORCE = 0x1000,		/* used by self referential force */
    M_PRIM_FAST = 0x800,	/* Fast Primitive (no need to re-eval) */
    M_PRIM = 0x400,		/* Primitive marker */
    M_IDLE = 0x200,		/* Schedule for idle at end of simulation time */
    M_VCL = 0x100,		/* VCL (PLI) marker */
    M_VCD = 0x80,		/* VCD (dumpvars) marker */
    M_ASYNCH = 0x40,		/* PLI asynch marker */
    M_IMMED = 0x20,		/* evaluate immediately */
    M_STROBE = 0x10,		/* Schedule for end of current simulation time */
    M_NET = 0x8,		/* Evaluate NETs immediately */
    M_CONT = 0x4,		/* Evaluate continuous assignment immediately */
    M_FIXED = 0x2,		/* if set, don't unthread markers after event */
    M_TEST = 0x1,		/* if set, test event expression at notification */
    M_PLAIN = 0
};

inline const char* MarkerFlagName( enum marker_flags flags ) 
{
     switch( flags ) {
     case M_OPTIMIZED:
	return "OPTIMIZED";
     case M_FORCE:
	return "FORCE";
     case M_PRIM_FAST: 
	return "PRIM_FAST";
     case M_PRIM:
	return "PRIM";
     case M_IDLE:
	return "IDLE";
     case M_VCL: 
	return "VCL";
     case M_VCD:
	return "VCD";
     case M_ASYNCH:
	return "ASYNCH";
     case M_IMMED:
	return "IMMED";
     case M_STROBE:
	return "STROBE";
     case M_NET:
	return "NET";
     case M_CONT:
	return "CONT";
     case M_FIXED:
	return "FIXED";
     case M_TEST:
	return "TEST";
     case M_PLAIN:
	return "PLAIN";
     default:
	return "???";
     }
}

extern time64 CurrentTime;
extern struct SCB *netlist;
extern struct SCB *freelist;
extern struct SCB *retrigger;

enum eMode { RUN, WAIT };

typedef struct SCB {
    /*
     * private global structures managed by SCB
     */
private:	
    static struct SCB *readylist;
    static struct SCB *readylist_last;
    /*
     * private per SCB structures
     */
private:
    struct SCB *next;		/* Next SCB on this linked list */
    struct SCB **prev;		/* Previous SCB on linked list */
    struct SCB *tl_next;	/* Next SCB on the timelist */
    struct SCB **tl_prev;	/* Previous SCB on the timelist */
    enum which_list list;
// ??? mdh - these are a hack for now
friend void SPLICE_LIST_SCB(enum which_list list, SCB ** POINT, SCB * NEW);
friend class TimeList;
friend void ADD_LIST_SCB(enum which_list list, SCB ** POINT, SCB * NEW);
friend void REMOVE_LIST_SCB(SCB * PTR);
friend void PrintSCB(SCB * scb);
friend void ScheduleNet(Marker * marker, tree decl);
friend void ScheduleNetImmediate(Marker * marker);
public:
    /*
     * public per SCB structures
     */
    time64 time;		/* For use in Delayed events */
    union tree_node *pc;	/* A pointer to executable code */
    enum eMode mode;		/* Used to determine if entering or exiting events */
    union {
	struct SCB *scb;	/* In case there are many SCBs waiting on the same event */
	struct context_member *fork_list;	/* If at fork, point to members */
	union tree_node *decl;	/* for nonblock assigns */
	Marker*  marker;	/* used for gate intialization */
    } here;			/* This is valid while waiting at a particular place */
    struct context_member *context;
    struct SCB *fork;		/* If inside fork/join, points to "main" SCB */
    unsigned fork_count;	/* Count of number of SCBs still in fork/join */
    int bit_offset;		/* used in delayed assignments */
    int group_offset;		/* used in delayed assignements */
    // class methods
    int OnReadyList() { return list == READY_LIST; }
    // global construction/destruction methods
    static SCB* BuildSCB(union tree_node *start, enum which_list list);
    static SCB *dispatcher(enum which_list remove_list);
    // global readylist methods
    static SCB* ActiveSCB() { return readylist; }
    static void ReadyListPushBack( SCB* scb );
    static void ReadyListPushFront( SCB* scb );
    static void ReadyListInsertAfter( SCB* insertionPoint, SCB* scb );
    static SCB* ReadyListPeek() { return readylist; }
    static SCB* ReadyListPeekAhead1() { return readylist?readylist->next:NULL; }
} SCB;

inline void SPLICE_LIST_SCB(enum which_list list, SCB ** POINT, SCB * NEW)
{
    SCB *tmp = *POINT;
    SCB* scb;
    for( scb=NEW; scb->next; scb = scb->next ) {
	scb->list = list;
    }
    switch (list) {
    case READY_LIST:
	if( *POINT == 0 ) {
	   SCB::readylist_last = scb;
	}
	break;
    default:;
	ASSERT( FALSE );
    }
    *POINT = NEW;
    (NEW)->prev = POINT;
    scb->list = list;
    (scb)->next = tmp;
    if (tmp) {
	tmp->prev = &(scb)->next;
    }
    return;
}

inline void ADD_LIST_SCB(enum which_list list, SCB ** POINT, SCB * NEW)
{
    switch (list) {
    case READY_LIST:
	if (*POINT == 0) {
	    SCB::readylist_last = NEW;
	}
	break;
    case TIME_LIST:
	ASSERT( FALSE );
	break;
    case EVENT_LIST:
        NEW->list = list;
	return;
    case NOLIST:
        NEW->list = list;
	return;
    default:;
    }
    SCB *tmp = *POINT;
    NEW->list = list;
    *POINT = NEW;
    (NEW)->prev = POINT;
    (NEW)->next = tmp;
    if (tmp)
	tmp->prev = &(NEW)->next;
    return;
}

inline void SCB::ReadyListPushBack( SCB* scb ) {
    if(readylist) {
        ADD_LIST_SCB(READY_LIST, &readylist_last->next, scb);
    } else {
        ADD_LIST_SCB(READY_LIST, &readylist, scb);
    }
}


inline void SCB::ReadyListPushFront( SCB* scb ) {
    ADD_LIST_SCB(READY_LIST, &readylist, scb);
}

inline void SCB::ReadyListInsertAfter( SCB* insertionPoint, SCB* scb ) {
    ADD_LIST_SCB(READY_LIST, &insertionPoint->next, scb);
}

typedef struct Marker {
    struct Marker *next;	/* ...(all from the same decl) */
    struct SCB *scb;
    union tree_node *delay;	/* if not null, put scb on timelist */
    union {
	union tree_node **expr;	/* The event expression to evaluate */
	union tree_node *arg;	/* For gates, point to the argument */
	struct monitor_info *mon;	/* For $monitor and $wwaves, etc. */
	union tree_node *tree;	/* Catch-all for typeless usage */
	int vcd_id;		/* For VCD output */
	void *vcl;		/* For PLI VCL */
    } expr;
    tree tmp1;			/* used by runtime to cache info */
    enum marker_flags flags;
    struct Marker *link;	/* Links markers for an event */
    union tree_node *decl;	/* pointer to the associated decl tree entry */
    struct Marker **prev;	/* pointer for event chain... */
    struct Marker* back;       /* points to previous marker */
    struct Marker** tail;	/* points to queue tail */
} Marker;

/* global info used during pass3 to build and thread markers as needed */

struct Marker_info {
    SCB *current_scb;
    int in_event;		/* If set, references will build markers */
    Marker *first;		/* Keep track of links during threading at pass3 */
    Marker *last;
    enum marker_flags flags;
    union tree_node *delay;
};

tree dispatch_pc(enum which_list);
tree dump_current_process(void);
void BuildMarker(tree decl, struct Marker_info *marker_info);
tree WaitOnTime(delay_t waittime, tree pc);
tree WaitOnTime64(time64 *waittime, tree pc);
void WaitOnEvent(Marker * marker, SCB * scb);
union tree_node *WaitOnEventAll(Marker *, union tree_node *);
void WaitOnEventOther(SCB *, Marker *, union tree_node *);
void NotifyEvent(Marker *, tree decl, int);
void NotifyDeferred(union tree_node *);
void FreeSCB(SCB *);
void event_undo(Marker * first);
void cleanup_SCB(SCB *);
void interject_readylist(SCB *, union tree_node *);
void thread_marker(Marker *);
void thread_all_markers(Marker *);
void Schedule(delay_t waittime, SCB * scb, int prepend);
void ScheduleDelta64(time64 *time, SCB * scb, int prepend);
void Schedule64(time64 *time, SCB * scb, int prepend);
void ScheduleNet(Marker * marker, tree decl);
void PrintSCB(SCB * scb);
tree WaitOnEventAll(Marker * marker, tree pc);
void interject_readylist(SCB * scb, tree pc);
void init_sched(void);


//#undef offsetof
//#define offsetof(c,m) ((size_t)(((c*)64)->m)-64)
#define container_of( ptr, type, member) ({                      \
        const typeof( (( type *)0)->member ) *__mptr = (ptr);    \
        ( type *)( (char *)__mptr - offsetof(type,member) );})


/******************************************************
 * time list management
 *****************************************************/
void DumpTimeListStats();
class TimeList {
    static SCB *timelist;
    static unsigned long long removeCycles;
    static unsigned long long removeCount;
    static unsigned long long removeGroupCycles;
    static unsigned long long removeGroupCount;
    static unsigned long long insertCycles;
    static unsigned long long insertCount;
public:
    TimeList() {
	atexit(DumpTimeListStats);
    }
    void Dump() {
        if( !printStats ) {
	    return;
        }
	printf_V( "\nTimeList Stats\n" );
	printf_V( "%-12s: %20s %20s\n", "operation", 
				"total cycles", "avg cycles" );
	printf_V( "-------------------------------\n" );
	printf_V( "%-12s: %20lld %20lld\n", "insert", insertCycles, insertCount?insertCycles/insertCount:0 ); 
	printf_V( "%-12s: %20lld %20lld\n", "remove", removeCycles, removeCount?removeCycles/removeCount:0 ); 
	printf_V( "%-12s: %20lld %20lld\n", "remove group", removeGroupCycles, removeGroupCount?removeGroupCycles/removeGroupCount:0 ); 
    }
		
    static void Insert( SCB* scb, int prepend ) {
#ifdef BUILD_HISTOGRAMS
	 unsigned long long int start = rdtsc();
#endif
	 SCB* link;
	 SCB* hol;
	 time64 stime = scb->time;
	/*
 	 * 5 cases:
 	 * 1) list is empty
 	 * 2) new layer1 entry is prepended
 	 * 3) existing layer1 entry utilized(prepend)
 	 * 4) existing layer1 entry utilized(postpend)
 	 * 5) new layer1 entry is inserted
 	 */
	 if( !timelist || (CMPTIME64(&stime, &timelist->time) < 0)) {
	     // case 1 and 2 
	     scb->next = scb;
	     scb->prev = &scb->next;
	     scb->tl_next = timelist;
	     scb->tl_prev = &timelist;
	     if( timelist ) {
		timelist->tl_prev = &scb->tl_next;
	     }
	     timelist = scb;
	     goto DONE;
	 } else {
	    SCB* n;
	    for(link = timelist; (n = link->tl_next)
	        && (CMPTIME64(&stime, &n->time) >= 0);
	        link = n);
            if( CMPTIME64(&stime,&link->time) == 0 ) {
	        // case 3 and 4
	        hol = link;
	    } else {
	        // case 5
	        scb->next = scb;
	        scb->prev = &scb->next;
		SCB* tl_next = link->tl_next;
	        scb->tl_next = tl_next;
	        scb->tl_prev = &link->tl_next;
	        if( tl_next ) {
	    	    tl_next->tl_prev = &scb->tl_next;
	        }
	        link->tl_next = scb;
	        goto DONE;
	    }
         }

	/*
 	 * head of chain is next element in time quanta
 	 */
	if( prepend ) {
	    // case 3
            scb->next = hol;
            scb->prev = hol->prev;
	    hol->prev = &scb->next;
	    *scb->prev = scb;

	    SCB* tl_next = hol->tl_next;
	    scb->tl_next = tl_next;
	    scb->tl_prev = hol->tl_prev;
	    hol->tl_next = NULL;
	    hol->tl_prev = NULL;
	    *scb->tl_prev = scb;
	    if( tl_next ) {
		tl_next->tl_prev = &(scb->next);
	    }
	} else {
	    // case 4
            scb->next = hol;
            scb->prev = hol->prev;
	    *scb->prev = scb;
	    hol->prev = &scb->next;
	    scb->tl_next = NULL;
	    scb->tl_prev = NULL;
	}
DONE:
        scb->list = TIME_LIST;

#ifdef BUILD_HISTOGRAMS
	 insertCycles += rdtsc()-start;
	 insertCount++;
#endif
        return;
    }
    static void Remove( SCB* scb ) {
#ifdef BUILD_HISTOGRAMS
	 unsigned long long int start = rdtsc();
#endif
	/*
  	 * cases:
  	 * 1) node is only on layer2 list
  	 * 2) node is on layer1 list and last node on layer2
  	 * 3) node is on layer1 list and not last node on layer2
	 */
	// cases 2 and 3
	SCB** tl_prev = scb->tl_prev;
	SCB* tl_next = scb->tl_next;
	if( tl_prev ) {
	    if( scb->next == scb ) {
                if((*(tl_prev) = tl_next) != 0)
	             tl_next->tl_prev = tl_prev;
	    } else {
		SCB* hol = scb->next;
		hol->tl_next = tl_next;
		hol->tl_prev = tl_prev;
		*tl_prev = hol;
		if( tl_next ) {
		    tl_next->tl_prev = &hol->tl_next;
		}
	    }
	}
	
	// remove from layer 2 list
	SCB** prev = scb->prev;
	SCB* next = scb->next;
        if ((*(prev) = next) != 0)
	     next->prev = prev;

#ifdef BUILD_HISTOGRAMS
	 removeCycles += rdtsc()-start;
	 removeCount++;
#endif
    }
    static SCB* Pop() {
	SCB* scb = timelist;
	if( !scb ) {
	    return NULL;
	}
	Remove(scb);
	return scb;
    }
    static SCB* Peek() {
	return timelist;
    }
    static SCB* PopGroup() {
#ifdef BUILD_HISTOGRAMS
	 unsigned long long int start = rdtsc();
#endif
	SCB* scb = timelist;

	if( scb ) {
            if((*(scb->tl_prev) = scb->tl_next) != 0) {
	        scb->tl_next->tl_prev = scb->tl_prev;
	    }
	    *scb->prev = NULL;
	    scb->prev = NULL;
	}
	
#ifdef BUILD_HISTOGRAMS
	 removeGroupCycles += rdtsc()-start;
	 removeGroupCount++;
#endif
         return scb;
    }
};
extern TimeList timelist;

inline void REMOVE_LIST_SCB(SCB * PTR)
{
    switch (PTR->list) {
    case READY_LIST:
        if ((*((PTR)->prev) = (PTR)->next) != 0)
	    (PTR)->next->prev = (PTR)->prev;
	if (SCB::readylist_last == PTR) {
	    if( !SCB::readylist ) {
		SCB::readylist_last = NULL;
	    } else {
		for( SCB::readylist_last=SCB::readylist; SCB::readylist_last->next; 
			SCB::readylist_last = SCB::readylist_last->next );
	    }
	}
	break;
    case TIME_LIST:
	timelist.Remove( PTR );
	return;
    case NOLIST:
	return;
    case EVENT_LIST:
	return;
    default:;
        if ((*((PTR)->prev) = (PTR)->next) != 0)
	    (PTR)->next->prev = (PTR)->prev;
    }
}



inline tree suspend_current_process(void)
{
    SCB *scb = SCB::dispatcher(NOLIST);
    return scb->pc;
}

void net_exec(tree pc);
void do_net_eval(tree node, int exclude_inout, tree exclude_tree);
void store(tree lval, tree pc);
extern Group **R;
extern nbits_t R_nbits;
inline void ScheduleNet_fast(Marker * marker)
{
    tree pc = marker->tmp1;
    /*
     * accelerate single driver/single location WIRE case
     */
    tree source = NET_SOURCE(pc);
    R_nbits = TREE_NBITS(pc);
    **R = *(DECL_STORAGE(source));
    *++R = *R+1;                    // + 1;

   // TRACE(pc, pc, *(R - 1), 1, 0);
    store(pc, pc);


}

inline void ScheduleNetImmediate(Marker * marker)
{
    SCB *scb = marker->scb, *tmp;
    tree pc = scb->pc;

/* If a port/net has just been changed, propigate immediately (also during
   continuous assignment initialization) */

    do_net_eval(pc, 0, NULL);
    store(pc, pc);
    if ( PORT_MIRRORED_ATTR(pc) ) {
        do_net_eval(pc, 1, NULL_TREE);
        store(DECL_THREAD(pc), pc);
    } else {
        if( !marker->tmp1 ) {
            tree source = NET_SOURCE(pc);
	    if( TREE_NBITS(pc) <= 32 && 
	        TREE_TYPE(pc) == NET_WIRE_TYPE &&
                source && NET_SOURCE(source) == NULL ) {
	        marker->flags = (enum marker_flags)
			(marker->flags | M_OPTIMIZED);
	        marker->tmp1 = pc;
 	    } else {
	        marker->tmp1 = (tree)1;
	    }
	}
    }

}

inline void AddMarkerToListTail( Marker*& head, Marker*& tail, Marker* marker ) 
{
    Marker* tmp;
    if( !tail ) {
	ADD_LIST(head, marker);
	marker->back = NULL;
    } else {
	ADD_LIST(tail->next, marker);
	marker->back = tail;
    }
    tail = marker;
    marker->tail = &tail;
}

inline void RemoveMarkerFromList( Marker* marker ) 
{
    if( marker->next ) {
	marker->next->back = marker->back;
    }
    if( *marker->tail == marker ) {
	*marker->tail = marker->back;
    }
    REMOVE_LIST(marker);
}
}
#endif				// SCHED_H
