/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* STORE.C - Routines for storing results of expressions */

#define STORE_C

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <stack>
#include "vtypes.h"
#include "tree.h"
#include "schedule.h"
#include "eval.h"
#include "runtime.h"
#include "flags.h"
#include "store.h"
#include "veriwellinterface.h"
#include "stats.h"

using std::stack;

namespace veriwell {
int store_computeOnly = 0;
int store_useCached = 0;
SCB *store_scb = 0;
extern unsigned long long int tscDelta;

/************************************************************
	SToreHistogram
	- histogram class for tracking store execution
 ***********************************************************/
class StoreHistogram : public Stats {
    static const int MAX_LENGTH = 1000;
    struct HistEntry {
	int code;
	long long int total;
	long long int noaccel;
 	long long int scalar;
	long long int mdw;
	long long int cycles;
    } treeCodeHistogram[NUM_TREE_CODES];
public:
    StoreHistogram();
    virtual const char* Name() { return "store"; }
    virtual unsigned long long TotalCycles();
    virtual void Dump( unsigned long long totalTicks );
    void Count( enum tree_code code, int width ) { 
	treeCodeHistogram[code].total++; 
	if( width==1 ) {
	    treeCodeHistogram[code].scalar++;
	}
	if( width > 32 ) {
	    treeCodeHistogram[code].mdw++;
	}
    }
    void CountNonAccel( enum tree_code code ) { treeCodeHistogram[code].noaccel++; }
    void Cycles( enum tree_code code, unsigned long long int t ) { treeCodeHistogram[code].cycles +=t; }
    static int HistCompare( const void* p1, const void* p2 ); 
};
#ifdef BUILD_HISTOGRAMS
    static StoreHistogram* histogram;
#endif

StoreHistogram::StoreHistogram()
{
    memset( treeCodeHistogram, 0, sizeof(treeCodeHistogram) );
    for( int i = 0; i < NUM_TREE_CODES; ++i ) {
	treeCodeHistogram[i].code = i;
    }
    masterStats.Add(this);
}
int StoreHistogram::HistCompare( const void* p1, const void* p2 ) 
{
    const HistEntry* e1 = static_cast<const HistEntry*>(p1);
    const HistEntry* e2 = static_cast<const HistEntry*>(p2);
    if( e1->cycles == e2->cycles ) {
	return 0;
    } else if( e1->cycles < e2->cycles ) {
	return 1;
    } else {
	return -1;
    }
}

unsigned long long StoreHistogram::TotalCycles()
{
    unsigned long long storeCycles = 0;
    for( int i = 0; i < NUM_TREE_CODES; ++i ) {
	storeCycles += treeCodeHistogram[i].cycles;
    }
    return storeCycles;
}

void StoreHistogram::Dump( unsigned long long totalTicks )
{
    if( !printStats ) {
	return;
    }

    qsort( treeCodeHistogram, NUM_TREE_CODES, sizeof(HistEntry), HistCompare );
    printf_V( "\nStore Histogram:\n" );
    printf_V( "%-25s  %10s %10s %10s %10s %20s, %10s %4s\n", "expression", "total", 
			"non-accel", "scalar", "multi-dw", "total cycles", "avg cycles", "%" );
    printf_V( "----------------------------------------------\n" );
    for( int i = 0; i < NUM_TREE_CODES; ++i ) {
	int code = treeCodeHistogram[i].code;
	long long int total = treeCodeHistogram[i].total;
	long long int noaccel = treeCodeHistogram[i].noaccel;
	long long int scalar = treeCodeHistogram[i].scalar;
	long long int mdw = treeCodeHistogram[i].mdw;
	long long int cycles = treeCodeHistogram[i].cycles;
	if( total == 0 ) {
	     continue;
	}
	printf_V( "%-25s: %10lld %10lld %10lld %10lld %20lld %10lld %4.2f\n", 
		tree_code_name[code], total, noaccel, scalar, mdw, cycles, cycles/total, (double)cycles*100./totalTicks );
    }
}

void store_array(tree decl, Bit index)
{
//  tree decl = ARRAY_REF_DECL (lval);
    ngroups_t array_hi = ARRAY_HI(decl);
    ngroups_t array_lo = ARRAY_LO(decl);
//  Bit aval = AVAL (g1);
    Group *g1, *g2;
    ngroups_t ngroups = 1 + bits_to_groups(TREE_NBITS(decl));
    ngroups_t offset, i;
    nbits_t nbits;
    int cond = 0;

    if (VECTOR_DIRECTION_ATTR(decl)) {
	if (index < array_lo || index > array_hi)
	    cond = X;
	else {
	    offset = (index - array_lo) * ngroups;
	}
    } else {
	if (index < array_hi || index > array_lo)
	    cond = X;
	else {
	    offset = (index - array_hi) * ngroups;
	}
    }

    g2 = *--R;
    if (!cond)
	g1 = DECL_STORAGE(decl) + offset;
    else
	return;

//  nbits = TREE_NBITS (lval);
    nbits = TREE_NBITS(decl);
    ngroups = bits_to_groups(nbits);

    for (i = 1; i <= ngroups; i++, g1++, g2++) {
	if (!cond) {
	    AVAL(g1) = AVAL(g2);
	    BVAL(g1) = BVAL(g2);
	} else {
	    AVAL(g1) = (Bit) - 1;
	    BVAL(g1) = (Bit) - 1;
	}
    }
    if (!cond) {
	AVAL(g1) = AVAL(g2) & MaskRight1(nbits);
	BVAL(g1) = BVAL(g2) & MaskRight1(nbits);
    } else {
	AVAL(g1) = (Bit) - 1 & MaskRight1(nbits);
	BVAL(g1) = (Bit) - 1 & MaskRight1(nbits);
    }
}

int deferred = 0;	/* set during the storing of concats */

/*
 * CASE(L)
 * - use instead of normal case label to accelerate
 *   switch execution
 */
/*
#ifdef BUILD_HISTOGRAMS				  
     #define CASE(L) case L: TREE_STORE_GOTO(lval)=&&L; L : tsc_code = TREE_CODE(lval); tscDeltaSave=tscDelta;tscStart=rdtsc();tscDelta=tscStart; L##_1	 
#else
    #define CASE(L) case L: TREE_STORE_GOTO(pc)=&&L; L 
#endif

#ifdef BUILD_HISTOGRAMS
    #define ENDCASE tscStop=rdtsc();delta = tscStop-tscDelta;histogram->Cycles(tsc_code,delta); delta=tscStop-tscStart;tscDelta=tscDeltaSave+delta; break
#else
    #define ENDCASE break
#endif
*/
#ifdef BUILD_HISTOGRAMS				  
     #define CASE(L) case L: tsc_code = TREE_CODE(lval); tscDeltaSave=tscDelta;tscStart=rdtsc();tscDelta=tscStart; L##_1	 
#else
    #define CASE(L) case L
#endif

#ifdef BUILD_HISTOGRAMS
    #define ENDCASE tscStop=rdtsc();delta = tscStop-tscDelta;histogram->Cycles(tsc_code,delta); delta=tscStop-tscStart;tscDelta=tscDeltaSave+delta; break
#else
    #define ENDCASE break
#endif

// ??? mdh - delay part selects and arrays need to be fixed
void store(tree lval, tree pc)
{
    enum tree_code lval_code;
    Group *g1, *g2;
    Bit a, b, current_a, current_b;
    int cond;
    int changed;
    ngroups_t i, ngroups;
    nbits_t nbits, save_R_nbits;
    tree t;
    enum logical_value old_state, new_state;
    struct part_info *p;
    int offset;
    int saveUseCached = store_useCached;
    store_useCached = 0;
#ifdef BUILD_HISTOGRAMS
    unsigned long long int delta;
    unsigned long long int tscDeltaSave;
    unsigned long long int tscStart;
    unsigned long long int tscStop;
    enum tree_code tsc_code;
#endif

#ifdef BUILD_HISTOGRAMS
	histogram->Count( TREE_CODE(lval), TREE_NBITS(lval) );
#endif
    /*
     * accelerate execution by embedding goto
     * addresses directly into tree
     */
/*
    void* jmpPtr = TREE_STORE_GOTO(lval);
    if( jmpPtr ) {
        goto *jmpPtr;
    }
*/
    /*
     * build statement execution histogram by frequency
     */
#ifdef BUILD_HISTOGRAMS
    histogram->CountNonAccel(TREE_CODE(lval));
#endif

    lval_code = TREE_CODE(lval);

    switch (lval_code) {
/***************************************************************************/
/* REG scalar DECL                                                         */
/***************************************************************************/

    case REG_SCALAR_DECL:
    CASE(NET_SCALAR_DECL): {
	if (store_computeOnly) {
	    ENDCASE;
	}
	Group* g1 = DECL_STORAGE(lval);
	Group* g2 = *--R;
	Bit aval = AVAL(g2) & 1;
	Bit bval = BVAL(g2) & 1;
	int changed = ((AVAL(g1) ^ aval) |(BVAL(g1) ^ bval));
	AVAL(g1) = aval;
	BVAL(g1) = bval;
	notify(lval, changed!=0, pc); } ENDCASE;

/***************************************************************************/
/* REG vector DECL                                                         */
/***************************************************************************/

    case INTEGER_DECL:
    case REG_VECTOR_DECL:
    case NET_VECTOR_DECL:
    case TMP_DECL:
    case TIME_DECL:
    case PARAM_DECL:
    CASE(SPECPARAM_DECL): {
	if (store_computeOnly) {
	    ENDCASE;
	}
	int nbits = TREE_NBITS(lval);
	int ngroups = bits_to_groups(nbits);
	Group* g1 = DECL_STORAGE(lval);
	Group* g2 = *--R;
/* MAKE THIS changed = 0 AFTER DEBUGGING!!! */
	Bit changed = 0;
	for (int i = 1; i <= ngroups; i++, g1++, g2++) {
	    changed |= (AVAL(g1) ^ AVAL(g2)) | (BVAL(g1) ^ BVAL(g2));
	    *g1 = *g2;
	}
	Bit mask = MaskRight1(nbits);
	Bit aval = AVAL(g2) & mask;
	Bit bval = BVAL(g2) & mask;
	changed |= ((AVAL(g1) ^ aval) |(BVAL(g1) ^ bval));
	AVAL(g1) = aval;
	BVAL(g1) = bval;
	notify(lval, changed!=0, pc);
	} ENDCASE;

/***************************************************************************/
/* Part reference                                                          */
/***************************************************************************/

    CASE(PART_REF):
	if (store_computeOnly) {
	    ENDCASE;
	}
	g1 = PART_STORAGE(lval);
	g2 = *--R;

	changed = part_lref(g1, g2, PART_INFO(lval));
	notify(PART_DECL(lval), changed, pc);
	ENDCASE;

/***************************************************************************/
/* Bit reference                                                           */
/***************************************************************************/

    CASE(BIT_REF):
	/* evaluate the index */
	nbits = R_nbits;	/* save these */
	ngroups = R_NGROUPS;
	eval(BIT_EXPR_CODE(lval));
	g2 = DECL_STORAGE(BIT_REF_DECL(lval));
//      g1 = *--R; /* points to index */

	cond = 0;
	changed = 0;
	new_state = ZERO;
	/* if index is a big number, make sure all high bits are zero */
	if (R_NGROUPS)		/* R_NGROUPS is from the index eval */
	    for (i = 1, g1 = *(R - 1) + 1; i < R_NGROUPS; i++, g1++) {
		if (BVAL(g1) || AVAL(g1)) {
		    cond = X;
		    break;
		}
	    }
	g1 = *--R;		/* Points to index */
	if (BVAL(g1) || cond)
	    cond = X;
	{
	    tree decl = BIT_REF_DECL(lval);
	    nbits_t decl_lsb = LSB(decl);
	    nbits_t decl_msb = MSB(decl);
	    Bit aval = AVAL(g1);
	    nbits_t bit_offset;
	    ngroups_t group_offset;

	    /* Make sure index is within bounds, else ignore */
	    if (VECTOR_DIRECTION_ATTR(decl)) {
		if (aval < decl_lsb || aval > decl_msb)
		    cond = Z;	/* will make store not happen */
		else {
		    bit_offset = (aval - decl_lsb) % BITS_IN_GROUP;
		    group_offset = bits_to_groups(aval - decl_lsb + 1);
		}
	    } else {
		if (aval < decl_msb || aval > decl_lsb)
		    cond = Z;	/* Will make store not happen */
		else {
//    bit_offset = (aval - decl_msb) % BITS_IN_GROUP;
//    group_offset = bits_to_groups (aval - decl_msb + 1);
		    bit_offset = (decl_lsb - aval) % BITS_IN_GROUP;
		    group_offset = bits_to_groups(decl_lsb - aval + 1);
		}
	    }

	    if (store_computeOnly) {
		// handle cond==Z case
		store_scb->bit_offset = bit_offset;
		if (cond == Z) {
		    store_scb->group_offset = 0xffffffff;
		} else {
		    store_scb->group_offset = group_offset;
		}
		SET_R_NBITS(nbits); /* restore original values in case trace need them. */
		ENDCASE;
	    }
	    if (saveUseCached) {
		SCB* scb = SCB::ActiveSCB();
		bit_offset = scb->bit_offset;
		group_offset = scb->group_offset;
		cond = 0;
		if (group_offset == 0xffffffff) {
		    cond = Z;
		}
	    }

	    /* take out the least significant bit and pop the stack */
	    a = AVAL(*--R) & 1;
	    b = BVAL(*R) & 1;
	    if (cond == Z)
		goto bit_ref_end;	/* Don't do store, but clean up */


	    current_a = AVAL(g2 + group_offset) & ((Bit) 1 << bit_offset);
	    current_b = BVAL(g2 + group_offset) & ((Bit) 1 << bit_offset);

	    if (a || cond == X) {	/* set bit */
		if (!current_a) {
		    AVAL(g2 + group_offset) |= (Bit) 1 << bit_offset;
		    changed = 1;
		    new_state = ONE;
		}
	    } else {		/* clear bit */

		if (current_a) {
		    AVAL(g2 + group_offset) &= ~((Bit) 1 << bit_offset);
		    changed = 1;
		} else
		    new_state = ONE;	/* never changed; reflects current val */
	    }

	    if (b || cond == X) {	/* set bit */
		if (!current_b) {
		    BVAL(g2 + group_offset) |= (Bit) 1 << bit_offset;
		    changed = 1;
		    new_state = (new_state == ZERO) ? Z : X;
		}
	    } else {		/* clear bit */

		if (current_b) {
		    Bit bval = BVAL(g2 + group_offset);
		    bval &= ~((Bit) 1 << bit_offset);
		    BVAL(g2 + group_offset) = bval;
//    BVAL (g2 + group_offset) &= ~((Bit) 1 << bit_offset);
		    changed = 1;
		} else
		    new_state = (new_state == ZERO) ? Z : X;
	    }
	}

      bit_ref_end:
	SET_R_NBITS(nbits); /* restore original values in case trace need them. */
	notify(BIT_REF_DECL(lval), changed, pc);
	ENDCASE;

/****************************************************************************/
/* Array reference                                                          */
/****************************************************************************/

    CASE(ARRAY_REF):
//{ int test = farheapcheck ();
//  if (test < 0)
//    fatal ("Corrupted before array store");
//}
	/* evaluate the index */
	save_R_nbits = R_nbits;	/* save these */

	eval(ARRAY_REF_EXPR_CODE(lval));


	SET_R_NBITS(save_R_nbits);
	cond = 0;

	g1 = *--R;		/* Points to index */
	if (BVAL(g1)) {
	    cond = Z;
	}

	offset = AVAL(g1);
	if (store_computeOnly) {
	    // handle cond==Z case
	    if (cond == Z) {
		store_scb->group_offset = 0xffffffff;
	    } else {
		store_scb->group_offset = offset;
	    }
	    ENDCASE;;
	}
	if (saveUseCached) {
	    offset = SCB::ActiveSCB()->group_offset;
	    cond = 0;
	    if (offset == 0xffffffff) {
		cond = Z;
	    }
	}
	if (cond) {
	    --R;
	    ENDCASE;
	}

	store_array(ARRAY_REF_DECL(lval), offset);

	notify(ARRAY_REF_DECL(lval), 1, pc);
	ENDCASE;

/***************************************************************************/
/* Concatenation                                                           */
/***************************************************************************/

    CASE(CONCAT_REF):
	store_useCached = saveUseCached;
	if (store_computeOnly) {
	    for (t = CONCAT_LIST(lval); t; t = TREE_CHAIN(t)) {
		p = (struct part_info *) TREE_VALUE(TREE_VALUE(t));
		store(*(tree *) TREE_PURPOSE(TREE_VALUE(t)), pc);
	    }
	    ENDCASE;
	}
	g1 = *--R;
	g2 = g1;
	nbits = R_nbits;	/* save these */
	ngroups = R_NGROUPS;
	i = 0;
	deferred++;		/* Don't trigger con't assgn yet */
	/* assume TREE_LIST is in reverse order (lsb first) */
	for (t = CONCAT_LIST(lval); t; t = TREE_CHAIN(t)) {
	    p = (struct part_info *) TREE_VALUE(TREE_VALUE(t));
	    i += PART_NEXT(p);
	    current_a = AVAL(g1 + i);
	    current_b = BVAL(g1 + i);
//        g2 += part_rref (g1, g2, (struct part_info *) TREE_VALUE (TREE_VALUE (t)));
	    g2 += part_rref(g1, g2, p);
	    R++;
	    store(*(tree *) TREE_PURPOSE(TREE_VALUE(t)), pc);
	    AVAL(g1 + i) = current_a;
	    BVAL(g1 + i) = current_b;
	}
	{			/* Save the global deferred_marker and clear.  This is done
				   because NotifyDeferred is not automic.  The case in
				   that excersizes this is where two module have
				   concatenated argument ports which connect to
				   each other and are defined as INOUTs:
				   foo f0 ({a0, a1});
				   bar b0 ({a0, a1}); */
	    extern tree deferred_markers;
	    tree local_deferred = deferred_markers;
	    deferred_markers = NULL;
	    if (!--deferred)
		NotifyDeferred(local_deferred);
	}
	SET_R_NBITS(nbits); /* restore original values in case trace need them. */
	saveUseCached = store_useCached;
	ENDCASE;

/***************************************************************************/
/* Shadow lval reference (from hierarchical ref)                                */
/***************************************************************************/

    CASE(SHADOW_LVAL_REF):
	if (store_computeOnly) {
	    ENDCASE;
	}
	store(TREE_CHAIN(lval), pc);
	ENDCASE;

/***************************************************************************/
/* Real DECL                                                               */
/***************************************************************************/

    CASE(REAL_DECL):
	if (store_computeOnly) {
	    ENDCASE;
	}
	g1 = DECL_STORAGE(lval);
	g2 = *--R;
	changed = 0;
	if (AVAL(g1) != AVAL(g2)) {
	    AVAL(g1) = AVAL(g2);
	    changed = 1;
	}
	if (BVAL(g1) != BVAL(g2)) {
	    BVAL(g1) = BVAL(g2);
	    changed = 1;
	}
	notify(lval, changed, pc);
	ENDCASE;

    default:
//    printf ("Line: %ld, File %s: \n", STMT_SOURCE_LINE (pc), STMT_SOURCE_FILE (pc));
	fatal("Internal error: Unimplemented lval reference: %s",
	      tree_code_name[lval_code]);
//          abort ();
    }				/* switch */
    store_useCached = saveUseCached;
}

void init_store()
{
    deferred = 0;
#ifdef BUILD_HISTOGRAMS
    histogram = new StoreHistogram;
#endif
}

}
