/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/**********************************************************
 *
 *      usertask.c
 *	    manage user tasks
 *
 **********************************************************
 */

#include <string.h>
#include <list>
#include "usertask.h"
#include "vtypes.h"
#include "pliext.h"
#include "io.h"
#include "flags.h"
#include "stats.h"

using std::list;
/*
 * existing compiled in list of user tasks
 */
extern s_tfcell veriusertfs[];

namespace veriwell {
list<p_tfcell> tfcellList;

/**********************************************
   Reason2Name 
   - convert pli reason code to ascii name
 **********************************************/
static const char* Reason2Name( int reason )
{
    switch(reason) {
    case reason_checktf:
	return "checktf";
    case reason_sizetf:
	return "sizetf";
    case reason_calltf:
	return "calltf";
    case reason_save:
	return "save";
    case reason_restart:
	return "restart";
    case reason_disable:
	return "disable";
    case reason_paramvc:
	return "paramvc";
    case reason_synch:
	return "synch";
    case reason_finish:
	return "finish";
    case reason_reactivate:
	return "reactivate";
    case reason_rosynch:
	return "rosync";
    case reason_paramdrc:
	return "paramdrc";
    case reason_interactive:
	return "interactive";
    case reason_endofcompile:
	return "endofcompile";
    case reason_force:
	return "force";
    case reason_release:
	return "release";
    case reason_scope:
	return "scope";
    default:
	return "???";
    }
}
/**********************************************
   UsertaskStats
   - dump usertask stats
 **********************************************/
class UsertaskStats : public Stats {
    static const int MIN_REASON=1;
    static const int MAX_REASON=18;
public:
    UsertaskStats() {
	masterStats.Add(this);
    }
    virtual const char* Name() { return "usertasks"; }
    virtual unsigned long long TotalCycles() 
    {
	unsigned long long totalCycles = 0;
        list<p_tfcell>::iterator ptr;
        for( ptr = tfcellList.begin(); ptr != tfcellList.end(); ++ptr ) {
    	    for( int i = MIN_REASON; i <= MAX_REASON; ++i ) {
    	        totalCycles += (*ptr)->call_stats[i].cycles;
            }
	}
        return totalCycles;
    }
    virtual void Dump( unsigned long long totalTicks )
    {
        if( !printStats ) {
    	    return;
        }
        const int MIN_REASON=1;
        const int MAX_REASON=18;
        printf_V( "\nUser Task Stats\n" );
        printf_V( "%-20s: %12s %20s %20s %10s %5s\n", "systask", 
    			"reason", "total", "total cycles", "avg cycles", "%" );
        printf_V( "--------------------------\n" );
        unsigned long long totalCycles = 0;
        unsigned long long totalCount = 0;
        list<p_tfcell>::iterator ptr;
        for( ptr = tfcellList.begin(); ptr != tfcellList.end(); ++ptr ) {
    	    for( int i = MIN_REASON; i <= MAX_REASON; ++i ) {
    	        totalCycles += (*ptr)->call_stats[i].cycles;
    	        totalCount += (*ptr)->call_stats[i].count;
    	        unsigned long long count = (*ptr)->call_stats[i].count;
    	        unsigned long long cycles = (*ptr)->call_stats[i].cycles;
    	        if( count ) {
    	            printf_V( "%-20s: %12s %20lld %20lld %10lld %5.2f%%\n", 
    	    		(*ptr)->tfname, Reason2Name(i), count, cycles, cycles/count, 
	    		(double)cycles*100./totalTicks );
    	        }
    	    }
        }
        if( totalCount ) {
            printf_V( "%-20s: %12s  %20lld %20lld %10lld\n", "total", "", 
    		totalCount, totalCycles, totalCycles/totalCount );
        }
    }
};
/**********************************************
  Usertask_Init
    - initialize usertask module. pass in 
      command line arguments to be parsed.
 **********************************************/
void Usertask_Init(int argc, char *argv[])
{
#ifdef BUILD_HISTOGRAMS
    new UsertaskStats();
#endif
}

/**********************************************
  Usertask_Lookup
    - Lookup user task/function name. Return
      task data structure. Return 0 if task/function
      is not found.
 **********************************************/
p_tfcell Usertask_Lookup(char *name)
{
    p_tfcell result = 0;

    result = FindPliTask( name );
    if( result ) {
        goto DONE;
    }
    
    for (int i = 0; veriusertfs[i].type; i++) {
	if (!strcmp(name, veriusertfs[i].tfname)) {
	    result = &veriusertfs[i];
	    break;
	}
    }
DONE:
#ifdef BUILD_HISTOGRAMS
    if( result && !result->statsInit ) { 
	result->statsInit = 1;
	tfcellList.push_back(result);
    }
#endif
    return result;
}

}
