/*-----------------------------------------------------------------------------
 * 	Copyright (c) 1997-2002 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   cstats.cpp
 *		- methods for status reporting tool
 *
 ******************************************************************************
 */
						

#include <config.h> 
#include "glue.h"
#include "cstats.h"
#include "cnode.h"
#include "cdecl.h"
#include "cvar.h"
#include "cnet.h"
#include "cmodule.h"
#include "cinstance.h"
#include "cdatatype.h"
#include <set>


static set<CModule*> leafModules;
static set<CModule*> allModules;
static set<CModule*> topLevelModules;
static set<CModule*> undefinedModules;
static StatsModule* top;
static FILE*	     f;	
extern char*	     outputPath;
extern int	     errorCount;
static Message*	     mMISUD;

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
#ifndef CYGWIN
extern "C" {
CBackend* CreateToolInstance()
{
    return new CStats();
}
}
#endif

/************************************************
 	Constructor
 ************************************************/
CStats::CStats( ) 
{
    mMISUD = Message::RegisterWarning( this, Message::eWARNING,
	 	"MISUD", "module '%s' is undefined\n", "<module>" );
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CStats::GetToolName( void )
{
	static char	toolName[] = "stats";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CStats::GetToolDescription( void )
{
	static char	toolDescription[] = "Print summary of hiearchy";

	return toolDescription;
}

/************************************************
	HasEdge
	- return TRUE if code tree
	  is an event control with a edge expression
 ************************************************/

static int HasEdge( CNode* n )
{
	int edge = 0;

	switch( n->GetOp() ) {
	case eEVOR:
		edge |= HasEdge( n->Arg<CNode*>(0) );
		edge |= HasEdge( n->Arg<CNode*>(1) );
		break; 			
	case eNEGEDGE:
	case ePOSEDGE:
		edge = 1;
		break;
	case eARRAY:
	case eNET_REF:
	case eVAR_REF:
	case eEXTERNAL_REF:		
		break;
	default:
		error( n->GetCoord(), "Illegal edge type %d", n->GetOp() );
		MASSERT( FALSE );
	};
	
	return edge;

}

/************************************************
	CollectRegRef
	- builds a list of regs that are referenced
	  in expressions
 ************************************************/
static int CollectRegRef( CNode* n, void* param ) 
{
	set<CVar*>& regList = *(set<CVar*>*)param;

	switch( n->GetOp() ) {
	case eVAR_REF:
		regList.insert( n->Arg<CVar*>(0) );		
		break;
	default:	
		break;
	}
	return TRUE;
		
}

/************************************************
	CollectLvals
	- builds a list of regs that appear as lvals
	  of procedural assignments 
 ************************************************/
static int CollectLvals( CNode* n, void* param ) 
{
	switch( n->GetOp() ) {
	case eNBASSIGN:
	case eASSIGN:
		n->Arg<CNode*>(1)->PreVisit1( CollectRegRef, param );
		break;
	default:
		break;
	}
	return TRUE;
		
}
/************************************************
	CollectAlwaysBlocks
	- builds a list of always blocks
 ************************************************/

static int CollectAlwaysBlocks( CNode* n, void* param ) 
{
	list<CNode*>& alwaysList = *(list<CNode*>*) param;

	switch( n->GetOp() ) {
	case eALWAYS:
		alwaysList.push_back(n);
		return FALSE;
	case eINIT:
		return FALSE;
	default:
		break;
	}
	return TRUE;
}
/************************************************
	CollectRegisters
	- builds a list of register elements 
 ************************************************/
static void CollectRegisters( set<CVar*>& regList, CNode* n )
{
	if( n == NULL ) {
		return;
	}
	/*
	 * find all always blocks
	 */
	list<CNode*> alwaysBlocks;
	n->PreVisit1( &CollectAlwaysBlocks, (void*)&alwaysBlocks );
	/*
	 * determine type of always block:
	 * 	register
	 *	latch
	 *	combinational
	 */
	list<CNode*>::iterator p;
	for( p = alwaysBlocks.begin(); p != alwaysBlocks.end(); ++p ) {
		CNode* l = (*p)->Arg<CNode*>(0);
		if( l->GetOp() != eEVENT ) {
			break;
		}
		CNode* event = l->Arg<CNode*>(0);
		if( event && HasEdge( event ) ) {
			/*
			 * register, so collect all lvals
			 */
			l->Arg<CNode*>(1)->PreVisit1( CollectLvals, &regList );	
		} else {
			/*
			 * could be latch or combinational
			 */
		}
		
	}
}

/************************************************
	GetFullWidth
	- calculate full width
 ************************************************/
static int GetFullWidth ( CDecl* decl ) 
{
    CDataType* dt = decl->GetDataType();
    if( !dt->IsPackedWidthEvaluateable() ) {
	return 0;
    }	
    int width = dt->GetPackedWidth();
    for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
	CNode* range = decl->GetUnpackedRange(i);
        if( range->IsWidthEvaluateable() ) {
	    width *= range->GetWidth();
        } else {
	    return 0;
	} 
    }
    return width;
}


/************************************************
	BuildHierarchy
	- build structural hierarchy 
 ************************************************/

static StatsModule* BuildHierarchy( CModule* m, const char* name, StatsModule* parent )
{
	StatsModule* s = new StatsModule;
	s->cmodule  = m;
	s->name     = name;
	s->parent   = parent;
	s->children = NULL;
	
	set<CVar*> regList;
	CollectRegisters( regList, m->GetCodeList() );
	set<CVar*>::iterator rp;
	s->registerBitCount = 0;
	s->cumulativeRegisterBitCount = 0;
	for( rp = regList.begin(); rp != regList.end(); ++rp ) {
		int width = GetFullWidth( *rp );
		s->registerBitCount += width;
		s->cumulativeRegisterBitCount +=  width;
	}
	if( parent != NULL ) {
		s->next     = parent->children;
		parent->children = s;
	} else {
		s->next = NULL;
	}
	vector<CInstance*>& inst = *m->GetInstanceList();
	vector<CInstance*>::iterator p;
	for( p = inst.begin(); p != inst.end(); ++p ) {
		BuildHierarchy( (*p)->GetDefinition(),
					      (*p)->GetName(), s );
	}
	for( StatsModule* child = s->children; child; child = child->next ) {
		s->cumulativeRegisterBitCount += 
					child->cumulativeRegisterBitCount; 
	}
	
	return s;
}

/************************************************
	PrintDecl
	- print declaration
 ************************************************/
static void PrintDeclaration ( CDecl* decl ) 
{
    printf( "%s", Identifier(decl->GetName()) );
    CDataType* dt = decl->GetDataType();
    if( dt->IsVector() && dt->IsPackedWidthEvaluateable() &&
	dt->GetPackedWidth() == 1 ) {
    } else if( dt->IsPackedWidthEvaluateable() ) { 
	if( dt->GetPackedMsb() && dt->GetPackedLsb() ) {
            printf( "[%ld:%ld]", 
		dt->GetPackedMsb()->EvalINT32(),
		dt->GetPackedLsb()->EvalINT32() );
	}
    } else {
        printf( "[???]" );
    }
    for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
        if( decl->GetUnpackedRange(i)->IsWidthEvaluateable() ) {
            printf( "[%ld:%ld]", 
    		decl->GetUnpackedMsi(i)->EvalINT32(),
    		decl->GetUnpackedLsi(i)->EvalINT32() );
        } else {
            printf( "[???]" );
        } 
    }
}
/************************************************
	PrintHierarchy
	- print structural hierarchy 
 ************************************************/

static void PrintHierarchy( StatsModule* s, int indent )
{
	
	printf( "%.4d:", indent );
	for( int i = 0; i < indent; ++i ) {
		printf( "  " );
	}
	printf( "%s[%s]: %d %d\n", s->name, s->cmodule->GetName(),
			s->registerBitCount,
			s->cumulativeRegisterBitCount );
	vector<CVar*>::iterator rp;
        for( rp = s->cmodule->GetVarList()->begin(); 
	     rp != s->cmodule->GetVarList()->end(); ++rp ) {
	    CDataType* dt = (*rp)->GetDataType();
	    if( !dt->IsPacked() ) {
		continue;
	    }
	    printf( "reg: " );
	    for( int i = 0; i < indent+1; ++i ) {
		printf( "  " );
	    }
	    PrintDeclaration( *rp );
	    printf( "\n" );
	}
	vector<CNet*>::iterator np;
        for( np = s->cmodule->GetNetList()->begin(); 
	     np != s->cmodule->GetNetList()->end(); ++np ) {
	    printf( "net: " );
	    for( int i = 0; i < indent+1; ++i ) {
		printf( "  " );
	    }
	    PrintDeclaration( *np );
	    printf( "\n" );
	}

	for( StatsModule* c = s->children; c ; c = c->next ) {
		PrintHierarchy( c, indent+1 );
	}
}


/************************************************
	CreateModuleList
	- mapping of name to module definition
 ************************************************/

static void CreateModuleList( CNode* n )
{
	if( n == NULL ) {
		return;
	}
        switch( n->GetOp() ) {
        case eLIST:
                CreateModuleList( n->Arg<CNode*>(0) );
                CreateModuleList( n->Arg<CNode*>(1) );
                break;
        case eMODULE_DEF:
		{
			CModule* m = n->Arg<CModule*>(0);
			allModules.insert(m);
			vector<CInstance*>& inst = *m->GetInstanceList();
			vector<CInstance*>::iterator p;
			for( p = inst.begin(); p != inst.end(); ++p ) {
				CModule* def = (*p)->GetDefinition();
				leafModules.insert(def);
			}
		}
                break;
	case ePRAGMA:
        case eCOMMENT:
	case eVRQ:
                break;
        default:
		MASSERT( FALSE );
        }       
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CStats::Process( list<CElement>& inputList,
		       list<CElement>& outputList )
{
    CNode* code = NULL;
    /*
     * merge all input streams together
     */
    list<CElement>::iterator ptr;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
        code = cLINK( code, ptr->Code() );
    }

    CreateModuleList( code );
	
    set_difference( allModules.begin(), allModules.end(),
		    leafModules.begin(), leafModules.end(),
		    inserter(topLevelModules, topLevelModules.end()) );

    set_difference( leafModules.begin(), leafModules.end(),
		    allModules.begin(), allModules.end(),
		    inserter(undefinedModules, undefinedModules.end()) );
	
    set<CModule*>::iterator p;
    for( p = undefinedModules.begin(); p != undefinedModules.end(); ++p ) {
	message( NULL, mMISUD, (*p)->GetName() );
    }	
    for( p = topLevelModules.begin(); p != topLevelModules.end(); ++p ) {
	StatsModule* m = BuildHierarchy(*p, (*p)->GetName(), NULL);
	m->next = top;
	top = m;
    }	
    int	totalRegisterBits = 0;
    for( StatsModule* s = top; s; s = s->next ) {
	PrintHierarchy( s, 0 );
	totalRegisterBits += s->cumulativeRegisterBitCount;
    }
    printf( "Total Register Bits: %d\n", totalRegisterBits );
}


