/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cattr.hpp
 *		- class definition of attribute
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CATTR_HPP
#define CATTR_HPP

#include <stdio.h>
#include <string>
#include <map>
#include "glue.h"
#include "cdecl.h"

using namespace std;


class CNode;


/**
 * Declaration object for holding lists of verilog attributes and their
 * corresponding expressions.
 */
class CAttr: public CDecl
{
private:
    struct Entry {
	string s;
	CNode* exp;
    };
    vector<Entry> table;
public:
	/**
 	 * Create a attribute object
 	 */
	CAttr( Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Add an attribute to list
 	 * \param name attribute name.
 	 * \param aExp attribute expression.
 	 */
	void 	AddAttribute( const char* name, CNode* aExp ) { 
	    Entry entry;
	    entry.s = name;
	    entry.exp = aExp;
	    table.push_back( entry );
	}
	/**
 	 * Get the number of attributes in list.
 	 * \return number of attributes in list.
 	 */
	int	GetAttributeCount() { return table.size(); }
	/**
 	 * Get an attribute by index.
 	 * \param index index of attribute (0 based).
	 * \return attribute name.
	 */
	const char*	GetAttributeName( int index ) 
	    { return table[index].s.c_str(); }
	/**
  	 * Get expression for attribute by index
	 * \param index index of attribute.
	 * \return attribute expression.
  	 */
	CNode*	GetAttributeExpression( int index ) { return table[index].exp; }
	/**
  	 * Get expression for attribute by name
	 * \param name name of attribute.
	 * \return attribute expression. NULL if not present.
  	 */
	CNode*	GetAttributeExpression( const char* name );
	/**
 	 * Check if attribute is in list.
 	 * \param name name of attribute.
 	 * \return non-zero if attribute is in list.
 	 */
	int	HasAttribute( const char* name );
	/**
 	 * Dump attributes to file.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
private:
	/**
 	 * Deep copy from declaration.
 	 * \param heap heap to use for allocation.
 	 * \param d declaration to copy.
 	 */
	void Copy( CObstack* heap, CAttr& d );
private:
	/*
 	 * Disable copy constructor.
 	 */
	CAttr( const CAttr& );
};

#endif // CATTR_HPP
