/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cblock.cpp
 *		- methods for blocks 
 *
 ******************************************************************************
 */

#include "glue.h"
#include "cnode.h"
#include "csymbol.h"
#include "cblock.h"
#include "csymtab.h"
#include "cportdir.h"
#include "cfref.h"
#include "cnet.h"
#include "cvar.h"
#include "cgenvar.h"
#include "cparam.h"

extern	CSymtab<CDecl> cuSymtab;
extern	CObstack*      unitStack;

static Message* mDODNM = Message::RegisterWarning( NULL, 
		 Message::eWARNING, "DODNM", 
                 "declaration of %s does not match width of port" 
		 " specification, overriding with latter declare range",
                        "<declaration>" );

/*************************************************
	Constructor
 *************************************************/
/* this constuctor is for named blocks only */
CBlock::CBlock( CSymbol* aName, Coord_t* aLoc ) : 
	CDecl( aName, aLoc, eBLOCK, NULL, eFLAG_NONE )
{
	cuSymtab.Add( aName, this );
	codeList = NULL;
	parent = NULL;
	rebalance = 0;
	trailingLabel = CSymbol::Lookup("");
}
/* this constructor is for other types of blocks */
CBlock::CBlock( CSymbol* aName, Coord_t* aLoc, Decl_t dtype ) : 
	CDecl( aName, aLoc, dtype, NULL, eFLAG_NONE )
{
	codeList = NULL;
	parent = NULL;
	rebalance = 0;
	trailingLabel = CSymbol::Lookup("");
}
/*************************************************
	SetParent
	- set blocks parent
 *************************************************/
void CBlock::SetParent( CBlock* aParent ) 
{
	parent = aParent;
}
/*************************************************
	GetParent
	- get blocks parent
 *************************************************/
CBlock* CBlock::GetParent( void ) 
{
	return parent;
}

/*************************************************
	Add
	- various flavors of add
 *************************************************/
	
void CBlock::Add( CNet* net )
{
	netList.push_back( net );
}
void CBlock::Add( CParam* param )
{
	paramList.push_back( param );
}
void CBlock::Add( CVar* var )
{
	varList.push_back( var );
}
void CBlock::Add( CGenvar* genvar )
{
	genvarList.push_back( genvar );
}
void CBlock::Add( CNode* code  )
{
	if( codeList == NULL ) {
		codeList = code;
	} else {
		codeList = cLINK( codeList, code );
	} 
	rebalance = 1;
	CodeListChanged();
}
void CBlock::Add( CBlock* child )
{
	children.push_back( child );
}
void CBlock::Add( CFref* fref )
{
	MASSERT( GetType() != eBLOCK ||
		 GetType() != eMODULE );
	frefList.push_back( fref );
}
void CBlock::Add( CInstance* instance )
{
	instanceList.push_back( instance );
}
/*************************************************
	GetNetList
	- get list of nets defined in block 
 *************************************************/
vector<CNet*>* CBlock::GetNetList()
{
	return &netList;
}
/*************************************************
	GetGenvarList
	- get list of genvars defined in block 
 *************************************************/
vector<CGenvar*>* CBlock::GetGenvarList()
{
	return &genvarList;
}
/*************************************************
	GetVarList
	- get list of variables defined in block 
 *************************************************/
vector<CVar*>* CBlock::GetVarList()
{
	return &varList;
}
/*************************************************
	GetParamList
	- get list of parameters defined in block 
 *************************************************/
vector<CParam*>*	CBlock::GetParamList()
{
	return &paramList;
}
/*******************************************
	GetInstanceList
	- return module's instance list 
 *******************************************/

vector<CInstance*>* CBlock::GetInstanceList()
{
	return &instanceList;
}
/*************************************************
	GetCodeList
	- get list of statements defined in block 
 *************************************************/
CNode*	CBlock::GetCodeList()
{
	if( rebalance ) {
	    codeList = RebalanceRight( codeList );
	    rebalance = 0;
	}
	return codeList;
}
/*************************************************
	SetCodeList
	- set list of statements defined in block 
 *************************************************/
void	CBlock::SetCodeList( CNode* list )
{
	codeList = list;
	rebalance = 1;
	CodeListChanged();
}
/*************************************************
	GetChildList
	- get list of children defined in block 
 *************************************************/
vector<CBlock*>*	CBlock::GetChildList()
{
	return &children;
}
/*******************************************
	SetSymtab
	- Set module symbol table 
 *******************************************/
void	CBlock::SetSymtab( CSymtab<CDecl>& aSymtab )
{
	symtab = aSymtab;
}
/*************************************************
	Dump	
	- dump block info 
 *************************************************/

class Dump_ {
	FILE* f;
public:
	Dump_(FILE* f): f(f){};
	void operator()( CDecl* r ) { r->Dump(f); }
};

void CBlock::Dump( FILE* f )
{
	if( GetSymbol() ) {
	    fprintf( f, "%s: %s defined in", declName[GetType()], GetName() );
	} else {
	    fprintf( f, "%s: %s defined in", declName[GetType()], "-" );
	}
	CDecl::Dump( f );
	fprintf( f, "\n" );

	fprintf( f, "symbol table:\n" );
	GetSymtab().Dump(f, 0);

	for_each( paramList.begin(), paramList.end(), Dump_(f) );
	for_each( varList.begin(), varList.end(), Dump_(f) );
	for_each( netList.begin(), netList.end(), Dump_(f) );
	for_each( genvarList.begin(), genvarList.end(), Dump_(f) );
	vector<CBlock*>* blocks = GetChildList();
	for_each( blocks->begin(), blocks->end(), Dump_(f) );
}
/*******************************************
	PreVisit
	- Support routine for PreVisit 
 	  tree transveral routine
 *******************************************/

void CBlock::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	codeList->PreVisit1( func, data );	
    }
    CDecl::PreVisit1( func, data );
}
/*******************************************
	PostVisit
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CBlock::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	codeList->PostVisit1( func, data );	
    }
    CDecl::PostVisit1( func, data );
}
/*******************************************
	PostSubVisit
	- Support routine for PostSubVisit 
 	  tree transveral routine
 *******************************************/

void CBlock::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	SetCodeList( codeList->PostSubVisit1( func, data ) );	
    }
    CDecl::PostSubVisit1( func, data );
}

/*************************************************
	Reconnect	
	- replace all PORT_REF and FWD_REF
	  expression with true REF
 *************************************************/
static void __Reconnect( CNode* n, void* )
{
        CPortDir*       portdir;
        CFref*          fref;
        CDecl*          decl;

        switch( n->GetOp() ) {
        case ePORT_REF:
                portdir = n->Arg<CPortDir*>(0); 
                decl = portdir->GetDecl();
                break;
        case eFWD_REF:
                fref = n->Arg<CFref*>(0);       
                decl = fref->GetDecl();
                break;
        default:
                return;
        }
        /*
         * decl may be null if errors are present
         */
        if( decl == NULL ) {
            return;
        }
        switch( decl->GetClass() ) {
        case eVAR:
            n->SetOp( eVAR_REF );
	    break;
        case ePARAM:
                n->SetOp( ePARAM_REF );
                break;
        case eNET:
                n->SetOp( eNET_REF );
                break;
	// this only occurs if errors are present
	case eFREF:
		break;
        default:
                MASSERT( FALSE );
        }
        n->Arg<CDecl*>(0) = decl;
}

void CBlock::Reconnect( CNode *n ) 
{
    if( !n ) {
	return;
    }
    n->PostVisit1( __Reconnect, NULL );
}

/*************************************************
	RedeclarePorts	
	- back patch all port declarations
	  and link to true declarations
 *************************************************/
class _RedeclarePorts : public binary_function<CPortDir*,CBlock*,void> {
public:
void operator()( CPortDir* port, CBlock* block ) const
{
        CSymbol* sym;
        Decl_t   type;
        CNode*   dMsb;
        CNode*   dLsb;
        CNode*   pMsb;
        CNode*   pLsb;
	CSymtab<CDecl>& symtab = block->GetSymtab();

	int isFunctionOrTask = block->GetType() == eFUNCTION;
        sym = port->GetSymbol();
        /*
         * if port decl is a FREF or NULL check for redefinition
         */
        CDecl*   decl = port->GetDecl();
        CFref*   fref = NULL;
        if( decl && decl->GetType() == eFREF ) {
            fref = static_cast<CFref*>(decl);
        }
        if( !decl || decl->GetType() == eFREF ) {
            decl = symtab.LookupTop( sym );
        }

        /*
         * if port decl has been redefined, link in new definition
         */  
        if( decl && decl->GetType() != eFREF ) {
            port->SetDecl(decl);
            goto CHECK;
        }
        /*
         * if port decl has not been redefined, 
         * redefine it as a reg if this is a function or udp
         * else as a net
         */
	if( !isFunctionOrTask ) {
            CNet* net = new(unitStack) CNet( sym, port->GetCoord(), 
			eWIRE, port->GetDataType()->Clone(unitStack), TRUE );
            net->PortDir( port );
            symtab.Add( sym, net );
            block->Add( net );
            decl = net;
	} else {
	    CDataType* dataType = port->GetDataType()->Clone(unitStack);
            CVar* reg = new(unitStack) CVar( sym, port->GetCoord(), 
			dataType, TRUE );
            symtab.Add( sym, reg );
            block->Add( reg );
            decl = reg;
	}
        port->SetDecl( decl );
        if( fref ) {
            fref->SetDecl( decl );
        }

        /*
         * check validity of new definition
         */
CHECK:
        /*      
         * registers can only redeclare output ports
         * in modules and UDPs
         */
        type = decl->GetType();
//        ??? mdh disabling this check as it appear sv allows this.
//        if( !isFunctionOrTask && type == eREG &&
//            port->GetType() != eOUTPUT ) {
//                error( decl->GetCoord(),
//                        "input/inout port %s cannot be declared as reg",
//                        decl->GetName() );
//        }
        /*
         * only nets, registers and reals are allowed to connect to ports
         */

        switch( type ) {
	case eNET:
	case eVAR:
	    switch( decl->GetDataType()->GetVarDataType() ) {
	    case eSHORTREAL:
	        dMsb = cINT32(31);    
	        dLsb = cINT32(0);    
	        break;
	    case eREAL:
	    case eREALTIME:
	        dMsb = cINT32(63);    
	        dLsb = cINT32(0);    
	        break;
            case eBYTE:
            case eTIME:
	    case eLONGINT:    
	    case eSHORTINT:    
	    case eINTEGER:      
            case eINT:
            case eREG:
	    case eLOGIC:          
	    case eBIT:           
	    case eNO_VAR_TYPE: {
			CDataType* dataType = decl->GetDataType();	
			if( dataType->IsScalar() || !dataType->IsPacked() ) {
                            dMsb = NULL;
                            dLsb = NULL;
			} else {
                            dMsb = dataType->GetPackedMsb();
                       	    dLsb = dataType->GetPackedLsb();
			}
                    } break;
	    case eINTERFACE_TYPE:
	    case ePROXY:
	    case eDEVENT:
		    dMsb = NULL;
		    dLsb = NULL;
		    break;
	    default:
		MASSERT( FALSE );
	    }
	    break;
        default:
                error( decl->GetCoord(),
                        "port %s cannot be declared as type %s",
                        decl->GetName(), declName[decl->GetType()] );

        }

	if( port->GetDataType() ) {
	    switch( port->GetDataType()->GetVarDataType() ) {
	    case eSHORTREAL:
	        pMsb = cINT32(31);    
	        pLsb = cINT32(0);    
	        break;
	    case eREAL:
	    case eREALTIME:
	        pMsb = cINT32(63);    
	        pLsb = cINT32(0);    
	        break;
            case eTIME:
	    case eLONGINT:    
	    case eSHORTINT:
	    case eINTEGER:      
	    case eBYTE:
            case eINT:
            case eREG:
	    case eLOGIC:          
	    case eBIT:           
	    case eNO_VAR_TYPE: {
			CDataType* dataType = port->GetDataType();	
			if( dataType->IsScalar() || !dataType->IsPacked() ) {
                            pMsb = NULL;
                            pLsb = NULL;
			} else {
                            pMsb = dataType->GetPackedMsb();
                            pLsb = dataType->GetPackedLsb();
			}
                    } break;
	    case eINTERFACE_TYPE:
	    case ePROXY:
	    case eDEVENT:
		    pMsb = NULL;
		    pLsb = NULL;
		    break;
	    default:
		MASSERT( FALSE );
	    }
	} else {
	    return;
	}

	if( !pMsb || !pLsb ) {
	    return;
	}

        if( ((dMsb == NULL) != (pMsb == NULL)) ||
            ((dLsb == NULL) != (pLsb == NULL)) ) {
                goto MISS_MATCH;
        }

        if( dMsb != NULL &&
            !dMsb->IsVolatile() &&
            !pMsb->IsVolatile() &&
            dMsb->EvalINT32() != pMsb->EvalINT32() ) {
                goto MISS_MATCH;
        }

        if( dLsb != NULL &&
            !dLsb->IsVolatile() &&
            !pLsb->IsVolatile() &&
            dLsb->EvalINT32() != pLsb->EvalINT32() ) {
                goto MISS_MATCH;
        }
        return;

MISS_MATCH:
	port->SetDataType( decl->GetDataType()->Clone(unitStack) );
        // "declaration of %s does not match width of port" 
	// " specification, overriding with latter declare range",
        message( decl->GetCoord(), mDODNM, decl->GetName() );
	

}
};

	
void CBlock::RedeclarePorts()
{
    for_each( GetPortDirList()->begin(), GetPortDirList()->end(),
              bind2nd(_RedeclarePorts(),this) );
}

/*************************************************
	LinkPortFref	
	- Link port fref to  port decl. 
 *************************************************/
void CBlock::LinkPortFref( CSymtab<CDecl>* symtab1,
	       CSymtab<CDecl>* symtab2,
	       CFref* fref )
{
	CDecl*		decl;
	
	if( fref->GetDecl() ) {
		return;
	}

	decl = symtab1->LookupTop( fref->GetSymbol() );
	if( decl == NULL ) {
	     return;
	}
	Decl_t	  type = decl->GetType();
	switch( type ) {
	case eINPUT:
	case eOUTPUT:
	case eINOUT:
	case eNODIR:
	    break;
	default:
	    return;
	}
	if( !fref->IsTypeValid( type ) ) {
	   return;
	}
	CPortDir* pdir = (CPortDir*) decl;
	MASSERT( pdir->GetDecl() != NULL );
	fref->SetDecl( pdir->GetDecl() );	
	fref->Direction( pdir->GetType() );
}

/*************************************************
	InferFref	
	- Create decls for unbound frefs where allowed
 *************************************************/
void CBlock::InferFref( CSymtab<CDecl>* symtab,
	       CFref* fref, int inferVector )
{
	CDecl*		decl;
	CBlock* 	root = this;
	
	while( root->GetParent() ) {
	    root = root->GetParent();
	}

	
	if( fref->GetDecl() ) {
		return;
	}
	decl = symtab->LookupTop( fref->GetSymbol() );
	// not sure why this isn't true
	// MASSERT( decl );
	if( decl == NULL || decl->GetType() == eFREF ) {
	    if( fref->IsTypeValid(eNET) && 
		( (inferVector /*&& fref->Lval()*/) ||
		  (!inferVector && !fref->IsWidthVolatile() &&
		   fref->IsWidthEvaluateable() && fref->GetWidth() == 1) ) ) {
                if( !fref->IndicesRangeValid() ) {
		    return;
		}
		CNode* range = fref->GetRange();
		list<CNode*> rangeList;
		if( range ) {
		    rangeList.push_back(range->Clone(unitStack));
		}
		CSimpleType* stype = new(unitStack) CSimpleType(eNO_VAR_TYPE,rangeList);
		CNet* net = new(unitStack) CNet( fref->GetSymbol(), 
				fref->GetCoord(), eWIRE, stype, TRUE );
		fref->SetDecl( net );
		symtab->Add( fref->GetSymbol(), net );
		root->Add( net );
	     } else if( inferVector && 
			fref->IsTypeValid(eVAR) /*&& fref->Lval()*/ ) {
                if( !fref->IndicesRangeValid() ) {
		    return;
		}
		CNode* range = fref->GetRange();
		list<CNode*> rangeList;
		if( range ) {
		    rangeList.push_back(range->Clone(unitStack));
		}
		CSimpleType* stype = new(unitStack) CSimpleType(eREG, rangeList);
		CVar* reg = new(unitStack) CVar( fref->GetSymbol(), 
				fref->GetCoord(), stype, TRUE );
		fref->SetDecl( reg );
		symtab->Add( fref->GetSymbol(), reg );
		root->Add( reg );
	     }
	}
}

/*************************************************
	LinkFref	
	- Link frefs to real decls. Report
	  errors on unlinked frefs
 *************************************************/
void CBlock::LinkFref( CSymtab<CDecl>* symtab,
	               CFref* fref, int inferVector )
{
	CDecl*		decl;
	CBlock* 	root = this;

	
	while( root->GetParent() ) {
	    root = root->GetParent();
	}
	
	if( fref->GetDecl() ) {
		return;
	}
	decl = symtab->LookupTop( fref->GetSymbol() );
	if( decl == NULL || decl->GetType() == eFREF ) {
            if( inferVector ) {
                 error( fref->GetCoord(),
                        "cannot infer declaration for %s", fref->GetName() );
            } else {
                error( fref->GetCoord(),
                    "%s is missing a declaration statement",
                    fref->GetName() );
            }
	    // create a bogus decl to keep things happy
	   if( fref->IsTypeValid( eVAR ) ) {
	       CSimpleType* stype = new(unitStack) CSimpleType(eREG);
	       CVar* reg = new(unitStack) CVar( fref->GetSymbol(), 
					fref->GetCoord(), stype, TRUE );
	       fref->SetDecl( reg );
	   } else {
	       CSimpleType* stype = new(unitStack) CSimpleType(eNO_VAR_TYPE);
	       CNet* net = new(unitStack) CNet( fref->GetSymbol(), 
					fref->GetCoord(), eWIRE, stype, TRUE );
	       fref->SetDecl( net );
	   }
	    return;
	}
	Decl_t	  type = decl->GetType();
	if( !fref->IsTypeValid( type ) ) {
	    if( type != eFREF ) {
	        error( fref->GetCoord(), 
		   "%s is incompatible with declaration statement",
		   fref->GetName() );
            } else if( fref->IsTypeValid( eINPUT ) ||
                       fref->IsTypeValid( eINOUT ) ||
                       fref->IsTypeValid( eNODIR ) ||
                       fref->IsTypeValid( eOUTPUT ) ) {
                error( fref->GetCoord(),
                       "%s is missing a port declaration statement",
                       fref->GetName() );
	    } else {
	        error( fref->GetCoord(), 
		   "%s is missing a declaration statement",
		   fref->GetName() );
	    }
	    // create a bogus decl to keep thing happy
	   if( fref->IsTypeValid( eVAR ) ) {
	       CSimpleType* stype = new(unitStack) CSimpleType(eREG);
	       CVar* reg = new(unitStack) CVar( fref->GetSymbol(), 
					fref->GetCoord(), stype, TRUE );
	       fref->SetDecl( reg );
	   } else {
	       CSimpleType* stype = new(unitStack) CSimpleType(eNO_VAR_TYPE);
	       CNet* net = new(unitStack) CNet( fref->GetSymbol(), 
					fref->GetCoord(), eWIRE, stype, TRUE );
	       fref->SetDecl( net );
	   }
		
	   return;
	}
	switch( type ) {
	case eINPUT:
	case eOUTPUT:
	case eINOUT:
	case eNODIR:
	    {
		CPortDir* pdir = (CPortDir*) decl;
		MASSERT( pdir->GetDecl() != NULL );
		fref->SetDecl( pdir->GetDecl() );	
		fref->Direction( pdir->GetType() );
	    }
	    break;
	default:
		fref->SetDecl( decl );	
		break;
	}
}
/*************************************************
	LinkFrefs	
	- Link frefs to real decls. Create decls
	  where allowed
 *************************************************/
void CBlock::LinkFrefs( CSymtab<CDecl>& portSymtab,
			CSymtab<CDecl>& cuSymtab,
			vector<CFref*>& frefList, 
			int inferVectors )
{
	/*
 	 * make multiple passes to connect frefs:
 	 * 1) connect ports
 	 * 	o Don't infer any declarations
 	 * 	o Don't fail on type mismatch
 	 * 	o Don't fail on undefined
 	 * 2) infer decls
 	 * 	o scalar nets
 	 * 	o vector nets and regs if -inferVector switch is enabled
 	 * 	  but only for lvals
 	 * 	o Don't fail on type mismatch
 	 * 	o Don't fail on undefined
 	 * Note: need to add width expansion 
 	 * 3) connect remaining frefs
 	 * 	o Fail on undefined
 	 * 	o Fail on type mismatch
 	 * 	o Infer on error
 	 * Note: need to add width validity checks 
 	 */
        vector<CFref*>::iterator ptr;
        for( ptr = frefList.begin(); ptr != frefList.end(); ++ptr ) {
                LinkPortFref( &portSymtab, &cuSymtab, *ptr );
	}
        for( ptr = frefList.begin(); ptr != frefList.end(); ++ptr ) {
                InferFref( &cuSymtab, *ptr, inferVectors );
	}
        for( ptr = frefList.begin(); ptr != frefList.end(); ++ptr ) {
                LinkFref( &cuSymtab, *ptr, inferVectors );
        }
}
