/*****************************************************************************
 * Copyright (C) 2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

#include "ccontext.h"
#include "cvector.h"

/**
 * Class for holding environment settings,
 * state and API associated with procedural execution.
 */

/********************************************
  Constructor
 *******************************************/
CContext::CContext() :
    paramsVolatile(1),
    trace(0),
    pc(NULL),
    heap("contextHeap")
{
}
/********************************************
  Destructor
 *******************************************/
CContext::~CContext()
{}
/********************************************
  Execute
    - execute code tree in this context,
      throws exceptions if error occurs  
 *******************************************/
void CContext::Execute( CNode* n ) throw (CExecuteException)
{
    pc = n;
    if( Trace() ) {
	DumpContext();
    }
    Coord_t coord = { 0, "???" };
    throw CExecuteUnimplementedConstruct(coord);
}
/********************************************
  Reset
    - reset and release decls in context back to initial state
 *******************************************/
void CContext::Reset()
{
}
/********************************************
  DumpContext
    - Dump context to log file.
 *******************************************/
void CContext::DumpContext()
{
	logprintf( "next pc to execute: %p\n", pc );
	logprintf( "paramsVolatile=%d\n", paramsVolatile );
	logprintf( "trace=%d\n", trace );
	logprintf( "Context:\n" );
}

/********************************************
  GetDeclRecord
    - return ptr to declarations record structure
 *******************************************/
CContext::DeclRecord* CContext::GetDeclRecord( CDecl* decl ) 
{ 
    if( decl2Record.find(decl) == decl2Record.end() ) {
	decl2Record[decl].decl = decl;
	decl2Record[decl].referenceCount = 0;
	decl2Record[decl].assignmentCount = 0;
	decl2Record[decl].vectorPtr = NULL;
	decl2Record[decl].realPtr = NULL;
	switch( decl->GetType() ) {
	case eREAL:
	case eREALTIME:
	    decl2Record[decl].vector = 0;
	    decl2Record[decl].realPtr = (double*)heap.Alloc(sizeof(double));
	    break;
	case eSHORTREAL:
	default:
	    decl2Record[decl].vector = 1;
            if( !decl->IsWidthEvaluateable() || decl->IsWidthVolatile() ) {
                throw CExecuteUnimplementedConstruct(*decl->GetCoord());
	    }
	    decl2Record[decl].vectorPtr = 
			CVector::AllocFromHeap( &heap, decl->GetWidth() );
	    /*
 	     * set initial vector to all x's
 	     */
	    decl2Record[decl].vectorPtr->SetToX();
	    break;
	}
    }
    return &decl2Record[decl];
}

/********************************************
  Referenced
    - return non-zero if declaration has been referenced.
 *******************************************/
int  CContext::Referenced( CDecl* decl )
{ 
    DeclRecord* record = GetDeclRecord( decl );
    return record->referenceCount > 0;
}

/********************************************
  MarkReferenced
    - mark declaration as referenced
 *******************************************/
void CContext::MarkReferenced( CDecl* decl )
{ 
    DeclRecord* record = GetDeclRecord( decl );
    record->referenceCount++;
}

/********************************************
  Assigned
    - return non-zero if declaration has been assigned
 *******************************************/
int  CContext::Assigned( CDecl* decl )
{ 
    DeclRecord* record = GetDeclRecord( decl );
    return record->assignmentCount > 0;
}

/********************************************
  MarkAssigned
    - mark declaration as assigned
 *******************************************/
void CContext::MarkAssigned( CDecl* decl )
{ 
    DeclRecord* record = GetDeclRecord( decl );
    record->assignmentCount++;
}

/********************************************
  VPtr
    - return pointer to declaration's storage.
 *******************************************/
CVector* CContext::VPtr( CDecl* decl ) 
{ 
    DeclRecord* record = GetDeclRecord( decl );
    MASSERT( record->vectorPtr );
    return record->vectorPtr;
}

/********************************************
  RPtr
    - return pointer to declaration's storage.
 *******************************************/
double*  CContext::RPtr( CDecl* decl )
{ 
    DeclRecord* record = GetDeclRecord( decl );
    MASSERT( !record->vector );
    return record->realPtr;
}
