/*****************************************************************************
 * Copyright (C) 2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   ccontext
 *		- container for execution state
 *
 *
 ******************************************************************************
 */

#ifndef CCONTEXT_HPP
#define CCONTEXT_HPP

class CVector;
class CNode;

#include "cdecl.h"

/**
 * Execution exception classes
 */
class CExecuteException {
Coord_t& coord;
public:
    CExecuteException( Coord_t& coord ) : coord(coord) {};
    void Coord( Coord_t& aCoord ) { aCoord = coord; }
    virtual const char* Message() = 0;
};

class CExecuteUnimplementedConstruct : public CExecuteException {
public:
    CExecuteUnimplementedConstruct( Coord_t& coord ) : CExecuteException( coord ) {};
    virtual const char* Message() { return "unimplemented construct"; }
};
class CExecuteUninitializedAccess    : public CExecuteException {
public:
    CExecuteUninitializedAccess( Coord_t& coord ) : CExecuteException( coord ) {};
    virtual const char* Message() { return "access to uninitialized variable"; }
};
class CExecuteUnmatchedPort    : public CExecuteException {
public:
    CExecuteUnmatchedPort( Coord_t& coord ) : CExecuteException( coord ) {};
    virtual const char* Message() { return "unmatch port"; }
};
class CExecuteArgEvaluationError    : public CExecuteException {
public:
    CExecuteArgEvaluationError( Coord_t& coord ) : CExecuteException( coord ) {};
    virtual const char* Message() { return "argument evaluation error"; }
};

/**
 * Class for holding environment settings,
 * state and API associated with procedural execution.
 */
class CContext {
    struct DeclRecord {
	CDecl* decl;
	int referenceCount;
	int assignmentCount;
	int vector;
	CVector* vectorPtr;
	double*  realPtr;
    };
    int paramsVolatile;	// non-zero if parameters should be 
			// treated as non-volatile
    int trace;	        // non-zero if execution should be trace for debugging
    CNode* pc;		// next node to execute
    map<CDecl*,DeclRecord> decl2Record;
    CObstack heap;
public:
    CContext();
    ~CContext();
    /*
     * execute code tree in this context
     */
    void Execute( CNode* n ) throw (CExecuteException);
    /*
     * reset and release decls in context back to initial state
     */
    void Reset();
    /*
     * decl management routines
     */
    int  Referenced( CDecl* decl );
    void MarkReferenced( CDecl* decl );
    int  Assigned( CDecl* decl );
    void MarkAssigned( CDecl* decl );
    CVector* VPtr( CDecl* decl );
    double* RPtr( CDecl* decl );
    /*
     * environment management
     */
    int  ParamsVolatile() { return paramsVolatile; }
    void ParamsVolatile( int value ) { paramsVolatile = value; }
    int  Trace() { return trace; }
    void Trace( int value ) { trace = value; }
private:
    DeclRecord* GetDeclRecord( CDecl* decl );
    void DumpContext();
};

#endif // CCONTEXT_HPP


