/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdatatype.cc
 *		- abstract class for data types 
 *
 *
 ******************************************************************************
 */

#define CDATATYPE_CC

#include "cdatatype.h"
#include "cnode.h"

/**
 * Create data type instance
 * This should never be call directly.
 */
CDataType::CDataType( VarDataType_t t )
    : const_(FALSE),
      varDataType(t),
      numberOfPackedDimensions(0)
{}

CDataType::CDataType( VarDataType_t t, list<CNode*> rangeList )
    : const_(FALSE),
      varDataType(t)
{
   packedRange.clear();
   list<CNode*>::iterator rPtr;
   for( rPtr = rangeList.begin(); rPtr != rangeList.end(); ++rPtr ) {
       CNode* range = *rPtr;
       /*
        * ensure ranges that are specified with macros, use defined macros
        */
       MASSERT( !range || 
		range->GetOp() != eMACRO_EXPR || 
		range->Arg<CNode*>(1) );
       packedRange.push_back(range); 
       
   }
   numberOfPackedDimensions = rangeList.size();
}
/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CDataType::Copy( CObstack* heap, const CDataType& o )
{
    const_   = o.const_;
    varDataType = o.varDataType;
    numberOfPackedDimensions = o.numberOfPackedDimensions;
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {
        packedRange.push_back( o.packedRange[i] ? o.packedRange[i]->Clone(heap) : NULL );
    }
}

/**
 * Determine if complete data structure is a scalar.
 * Scalars are single-bit intergal types. 
 * \return non-zero if complete data structure is a vector.
 */
int CDataType::IsScalar() const 
{ 
    return IsVector() && 
	   CDataType::GetNumberOfPackedDimensions() == 0 &&
	   GetBaseWidth() == 1; 
}

/**
 * Get expression for declaration's msb. Valid only for
 * Packed and Vector declarations.
 * \return msb expression, NULL if scalar.
 */
CNode* CDataType::GetPackedMsb() const
{
    MASSERT( IsPacked() );
    if( GetNumberOfPackedDimensions() == 1 ) {
	CNode* range = GetPackedRange(0);
	MASSERT( range );
	while( range->GetOp() == eMACRO_EXPR ) {
	    range = range->Arg<CNode*>(1);
        }
	if( !range->Arg<CNode*>(1) ) {
	    return cSUB(range->Arg<CNode*>(0),cINT32(1));
	}
	return range->Arg<CNode*>(0);
    }
    if( IsPackedWidthEvaluateable() && 
	!IsPackedWidthVolatile() &&
	GetPackedWidth() == 1 ) {
	return NULL;
    }
    return cSUB(GetPackedWidthExp(),cINT32(1));
}
/**
 * Get expression for declaration's lsb. Valid only for
 * Packed and Vector declarations.
 * \return lsb expression, NULL if scalar.
 */
CNode* CDataType::GetPackedLsb() const
{
    if( GetNumberOfPackedDimensions() == 1 ) {
	CNode* range = GetPackedRange(0);
	MASSERT( range );
	while( range->GetOp() == eMACRO_EXPR ) {
	    range = range->Arg<CNode*>(1);
        }
	if( !range->Arg<CNode*>(1) ) {
	    return cINT32(0);
	}
	return range->Arg<CNode*>(1);
    }
    if( IsPackedWidthEvaluateable() && 
	!IsPackedWidthVolatile() &&
	GetPackedWidth() == 1 ) {
	return NULL;
    }
    return cINT32(0);
}

/**
 * Get expression for datatype's overall packed or vector 
 * range (msb/lsb)
 * \return range expression, NULL if scalar.
 */
CNode* CDataType::GetPackedRange() const
{
    if( GetNumberOfPackedDimensions() == 1 ) {
	CNode* range = GetPackedRange(0);
	MASSERT( range );
	while( range->GetOp() == eMACRO_EXPR ) {
	    range = range->Arg<CNode*>(1);
        }
	return range;
    }
    if( IsPackedWidthEvaluateable() && 
	!IsPackedWidthVolatile() &&
	GetPackedWidth() == 1 ) {
	return NULL;
    }
    return cRANGE(GetPackedMsb(), GetPackedLsb());
}

/**
 *  Evaluate current packed or vector decl width. 
 *  Expression must be constant.  
 * \return evaluated packed or vector width expression.
 */
INT32 CDataType::GetPackedWidth( void ) const {
	MASSERT( IsPacked() ) ;
	int width = GetBaseWidth();
	for( int i = 0; i < numberOfPackedDimensions; ++i ) {
	     CNode* range = packedRange[i];
	     MASSERT( range );
	     if( range->GetOp() == eMACRO_EXPR ) {
		range = range->Arg<CNode*>(1);
	     }
	     CNode* msb = range->Arg<CNode*>(0);
	     CNode* lsb = range->Arg<CNode*>(1);
	     MASSERT( msb && lsb );
 	     width *= range->GetWidth();
	}
        return width;
}


/**
 * Get expression for datatype's overall packed or vector width
 * \return width expression
 */
CNode* CDataType::GetPackedWidthExp() const
{
    MASSERT( IsPacked() ) ;
    if( !IsPackedWidthVolatile() && IsPackedWidthEvaluateable() ) {
        return cINT32(GetPackedWidth());
    }

    MASSERT( IsPacked() );
    CNode* result = NULL;
    if( IsBaseWidthVolatile() || !IsBaseWidthEvaluateable() || 
	GetBaseWidth() != 1 ) {
        result = GetBaseWidthExp();
    }
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {
	if( packedRange[i]->IsWidthEvaluateable() &&
	    packedRange[i]->GetWidth() == 1 ) {
	    continue;
	}
	if( result ) {
	    result = cMUL( result, packedRange[i]->GetWidthExp() );
	} else {
	    result = packedRange[i]->GetWidthExp();
	}
    }
    if( !result ) {
	return cINT32(1);
    }
    return result;
}

/***********************************************************
	PackedWidthDirection
	- evaluate current decl width direction.
	  Expression must be constant.  
	  1 - big endian, -1 - little endian, 0 - cannot be determined
 **********************************************************/

int CDataType::PackedWidthDirection( void ) const 
{
    MASSERT( IsPacked() );
    if( numberOfPackedDimensions != 1 ) {
	return 1;
    }
    CNode* lsb;
    CNode* msb;
    CNode* range = packedRange[0];
    MASSERT( range );
    if( !range->IsWidthEvaluateable() ) {
	return 0;
    }
    if( range->GetOp() == eMACRO_EXPR ) {
        range = range->Arg<CNode*>(1);
    }
    lsb = range->Arg<CNode*>(1);
    msb = range->Arg<CNode*>(0);
    INT32 l = lsb->EvalINT32();
    INT32 m = msb->EvalINT32();
    return m>=l ? 1 : -1;
}

/**
 * Determine if packed width of declaration is constant, 
 * ie dependent upon only constants and parameters.
 * \return non-zero of packed or vector width of declaration is constant.
 */
int  CDataType::IsPackedWidthConstant( void ) const {
    int result = IsBaseWidthConstant();
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {
	result &= packedRange[i]->IsWidthConstant();
    }
    return result;
} 

/**
 * Determine if packed width of declaration is volatile, ie depend
 * upon parameters or variables.
 * \return non-zero of packed or vector width of declaration is volatile.
 */
int  CDataType::IsPackedWidthVolatile( void ) const {
    int result = IsBaseWidthVolatile();
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {
	result |= packedRange[i]->IsWidthVolatile();
    }
    return result;
} 

/**
 * Determine if packed width of declaration can be evaluated.
 * \return non-zero of packed or vector width of declaration can be evaluated.
 */
int  CDataType::IsPackedWidthEvaluateable( void ) const {
    if( !IsPacked() ) {
	return FALSE; 
    }
    int result = IsBaseWidthEvaluateable();
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {
	result &= packedRange[i]->IsWidthEvaluateable();
    }
    return result;
} 
/**
 * Get expression tree for upper limit of given packed array dimension.
 * \param dim dimension.
 * \return expression tree for packed upper limit.
 */
CNode* CDataType::GetPackedMsi( INT32 dim ) const 
{ 
    return packedRange[dim]->Arg<CNode*>(0); 
}
/**
 * Get expression tree for lower limit of packed array dimension.
 * \param dim dimension.
 * \return expression tree for packed lower limit.
 */
CNode* CDataType::GetPackedLsi( INT32 dim ) const
{ 
    return packedRange[dim]->Arg<CNode*>(1); 
}

/**
 * Walk tree invoking callback on each node before children have 
 * been visited. callback can terminate transversal early.
 * callback returns 0 to terminate transversal early.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CDataType::PreVisit1( int (*callback)(CNode*,void*), void* data )
{
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {  
	if( packedRange[i] ) {
	    packedRange[i]->PreVisit1( callback, data ); 
	}
    }
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CDataType::PostVisit1( void (*callback)(CNode*,void*), void* data )
{
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {  
	if( packedRange[i] ) {
	    packedRange[i]->PostVisit1( callback, data ); 
	}
    }
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited. callback may modify tree and return a modified
 * subtree.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 * \return modified child subtree.
 */
void CDataType::PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data )
{
    for( int i = 0; i < numberOfPackedDimensions; ++i ) {  
	if( packedRange[i] ) {
	    packedRange[i] = packedRange[i]->PostSubVisit1( callback, data ); 
	}
    }
}

