/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdatatype.h
 *		- abstract class for data types 
 *
 *
 ******************************************************************************
 */

#ifndef CDATATYPE_H
#define CDATATYPE_H

#include <list>
#include "cobject.h"

class CNode;

/**
 * Variable data types
 */
enum VarDataType_t {
	eNO_VAR_TYPE = 0,///< no type declaration
	eREAL,		 ///< real declaration
	eSHORTREAL,	 ///< shortreal declaration
	eREALTIME,	 ///< realtime declaration
	eREG,	         ///< register declaration
	eTIME,		 ///< time declaration
	eLOGIC,		 ///< logic declaration
	eINTEGER,	 ///< integer declaration
	eBIT,		 ///< integer declaration
	eINT,		 ///< int declaration
	eBYTE,		 ///< byte declaration
	eSHORTINT,	 ///< shortint declaration
	eLONGINT,	 ///< longint declaration
	eDEVENT,	 ///< event declaration
	eSTRUCT,	 ///< struct declaration
	eUNION,	 	 ///< union declaration
	eUNION_TAGGED,	 ///< tagged union declaration
	eENUM,	 	 ///< enum declaration
	eCLASS,	 	 ///< class declaration
	ePROXY,	 	 ///< proxy declaration
	eINTERFACE_TYPE, ///< interface declaration
	eSTRING,         ///< string declaration
};
#if defined CDATATYPE_CC
/**
 * Array to convert DataType_t to character string
 */
const char* varDataTypeName[] = {
	"",
	"real",
	"shortreal",
	"realtime",
	"reg",
	"time",
	"logic",
	"integer",
	"bit",
	"int",
	"byte",
	"shortint",
	"longint",
	"event",
	"struct",
	"union",
	"union tagged",
	"enum",
	"class",
	"proxy",
	"interface type",
	"string",
};
#else // CDATATYPE_CC
extern const char* varDataTypeName[];
#endif // CDATATYPE_CC

/**
 * Expression node type
 */
enum NodeType_t {
	eU,	///< undefined
	eR,	///< real - have width 0
	eS,	///< signed bit vector, includes integer
	eE,	///< event - have width 0
	eB	///< unsigned bit vector
};

/**
 * Base class for describing data types
 */
class CDataType : public CObject
{
public:
private:
	int	      const_;	        ///< declaration is a const type
	VarDataType_t varDataType;      ///< variable data type
        int numberOfPackedDimensions;   ///< number of packed dimensions
        vector<CNode*> packedRange;     ///< packed dimension ranges
public:
        /**
         * Create data type instance
         */
        CDataType( VarDataType_t t );
        /**
         * Create data type instance
         */
        CDataType( VarDataType_t t, list<CNode*> packedList );
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap ) = 0;
	/**
 	 * Set declaration's variable data type
 	 * \param v variable data type
 	 */
	virtual void SetVarDataType( VarDataType_t v ) { varDataType = v; }
	/**
 	 * Get declaration's variable data type
 	 * \return variable data type
 	 */
	virtual VarDataType_t  GetVarDataType() const { return varDataType; }
	/**
 	 * Get declaration's variable data type as a string
 	 * \return value of signed property.
 	 */
	virtual const char*  GetVarDataTypeName() const { return varDataTypeName[varDataType]; }
	/**
 	 * Set declaration's const property.
 	 * Use to indicate type is a constant variable.
 	 * \param v value to const property.
 	 */
	virtual void SetConst( int v ) { const_ = v; }
	/**
 	 * Get declaration's const property.
 	 * \return value of const property.
 	 */
	virtual int  GetConst() const { return const_; }
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const = 0;
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const = 0;
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const = 0;
	/**
         * Validate data structure 
	 * \param errorMessage description of validation failure
	 * \return true if structure is valid
	 */
	virtual bool Validate( string& errorMessage ) { return true; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero if vector or packed width of declaration 
 	 * is constant.
 	 */
	virtual int  IsBaseWidthConstant( void ) const = 0;
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is volatile, 
 	 * ie depend upon parameters or variables.
 	 * \return non-zero of vector or packed width of declaration 
 	 * is volatile.
 	 */
	virtual int  IsBaseWidthVolatile( void ) const = 0;
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration can be evaluated.
 	 * \return non-zero of packed or vector width of declaration can 
 	 * be evaluated.
 	 */
	virtual int  IsBaseWidthEvaluateable( void ) const = 0;
        /**
         * Evaluate base width (sans packed dimensions) of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetBaseWidth( void ) const = 0;
        /**
         * Get expression for datatype's base width (sans packed dimensions)
         * \return width expression
         */
        virtual CNode* GetBaseWidthExp() const = 0;
        /**
         * Determine if complete data structure is a scalar.
         * Scalars are single-bit intergal types. 
         * \return non-zero if complete data structure is a vector.
         */
        virtual int IsScalar() const;
        /**
         * Determine if complete data structure is a vector.
         * Vectors are multi-bit intergal types.
         * \return non-zero if complete data structure is a vector.
         */
        virtual int IsVector() const = 0;
        /**
         * Determine if complete data structure is packed.
         * Packed data structures can be treated as a single vector.
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const = 0;
	/**
 	 * Determine if packed or vector width of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero if vector or packed width of declaration 
 	 * is constant.
 	 */
	virtual int  IsPackedWidthConstant( void ) const;
	/**
 	 * Determine if packed or vector width of declaration is volatile, 
 	 * ie depend upon parameters or variables.
 	 * \return non-zero of vector or packed width of declaration 
 	 * is volatile.
 	 */
	virtual int  IsPackedWidthVolatile( void ) const;
	/**
 	 * Determine if packed or vector width of declaration can be evaluated.
 	 * \return non-zero of packed or vector width of declaration can 
 	 * be evaluated.
 	 */
	virtual int  IsPackedWidthEvaluateable( void ) const;
        /**
         * Get expression for declaration's msb. Valid only for
         * Packed and Vector declarations.
         * \return msb expression, NULL if scalar.
         */
        virtual CNode* GetPackedMsb() const;
        /**
         * Get expression for declaration's lsb. Valid only for
         * Packed and Vector declarations.
         * \return lsb expression, NULL if scalar.
         */
        virtual CNode* GetPackedLsb() const;
        /**
         * Get expression for datatype's overall packed or vector 
         * range (msb/lsb)
         * \return range expression, NULL if scalar.
         */
        virtual CNode* GetPackedRange() const;
        /**
         * Evaluate packed or vector width of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetPackedWidth( void ) const;
        /**
         * Get expression for datatype's overall packed or vector width
         * \return width expression
         */
        virtual CNode* GetPackedWidthExp() const;
	/**
 	 * Evaluate current decl width direction. Expression must be constant.  
	 * \return 1 - big endian, -1 - little endian, 0 - cannot be determined
	 */
	virtual int   PackedWidthDirection( void ) const;
	/**
 	 * Get number of packed dimensions of declaration.
 	 * \return number of dimensions, 0 if non-array.
 	 */
	virtual INT32 GetNumberOfPackedDimensions( void ) const 
	{ return numberOfPackedDimensions;}
	/**
 	 * Get expression tree for upper limit of given packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for packed upper limit.
	 */
	virtual CNode* GetPackedMsi( INT32 dim ) const; 
	/**
 	 * Get expression tree for lower limit of packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for packed lower limit.
	 */
	virtual CNode* GetPackedLsi( INT32 dim ) const;
	/**
 	 * Get expression for range of packed array for dimension.
 	 * \param dim dimension to get.
 	 * \return range expression for limit.
 	 */
	virtual CNode*  GetPackedRange( INT32 dim ) const 
	{ return packedRange[dim]; }


        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const = 0;
        /**
         * Walk tree invoking callback on each node before children have 
         * been visited. callback can terminate transversal early.
         * callback returns 0 to terminate transversal early.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PreVisit1( int (*callback)(CNode*,void*), void* data );
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PostVisit1( void (*callback)(CNode*,void*), void* data );
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited. callback may modify tree and return a modified
         * subtree.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         * \return modified child subtree.
         */
	virtual void PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data );
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CDataType& o );
private:
	/*
 	 * disable copy constructors
 	 */
	CDataType( const CDataType& o );

};

#endif // CDATATYPE_H

