/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdecl.cpp
 *		- methods for declarations 
 *
 *
 ******************************************************************************
 */

#define CDECL_CPP

#include "cdecl.h"
#include "cnode.h"
#include "cattr.h"
#include "cdatatype.h"

/************************************************************
  	Copy
	- perform deep copy from given object to this object
 ***********************************************************/

void CDecl::Copy( CObstack* heap, const CDecl& o )
{
    loc           = o.loc;
    symbol        = o.symbol;
    type          = o.type;
    declStatement = o.declStatement;
    attributes    = o.attributes ? o.attributes->Clone(heap) : NULL;
    pragmas       = o.pragmas ? o.pragmas->Clone(heap) : NULL;
    numberOfUnpackedDimensions = o.numberOfUnpackedDimensions;
    for( int i = 0; i < numberOfUnpackedDimensions; ++i ) {
        unpackedRange[i] = o.unpackedRange[i] ? o.unpackedRange[i]->Clone(heap) : NULL;
    }
    flags = o.flags;
    automatic = o.automatic;
    constAttr = o.constAttr;
    wireAttr = o.wireAttr;
    varAttr = o.varAttr;
    dataType = o.dataType ? o.dataType->Clone(heap) : NULL;
}
/***********************************************************
	GetUnpackedMsi
 	- Get expression tree for upper limit of unpacked array dimension.
 **********************************************************/
CNode* CDecl::GetUnpackedMsi( INT32 dim ) 
{ 
   return unpackedRange[dim] ? 
		(CNode*)unpackedRange[dim]->Arg<CNode*>(0) : NULL; 
}
/***********************************************************
	GetUnpackedLsi
 	- Get expression tree for lower limit of unpacked array dimension.
 **********************************************************/
CNode* CDecl::GetUnpackedLsi( INT32 dim ) 
{ 
   return unpackedRange[dim] ? 
		(CNode*)unpackedRange[dim]->Arg<CNode*>(1) : NULL; 
}

/***********************************************************
	HasAttribute
	- returns 1 if attribute is present
 **********************************************************/
int CDecl::HasAttribute( char* name, CNode* n, int init )
{
    int result = 0;
    if( !name ) {
	return 0;
    }
    if( init ) {
	n = GetAttributes();
    }
    if( !n ) {
	return 0;
    }
    switch( n->GetOp() ) {
    case eLIST:
	result |= HasAttribute( name, n->Arg<CNode*>(0), 0 );
	result |= HasAttribute( name, n->Arg<CNode*>(1), 0 );
	break;
    case eATTRIBUTE:
	{
	    CAttr* attr = n->Arg<CAttr*>(0);
	    return attr->HasAttribute( name );
	}
    default:
	return 0;
    }
   
    return result; 
}

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CDecl::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    for( int i = 0; i < GetNumberOfUnpackedDimensions(); ++i ) {
	if( GetUnpackedRange( i ) ) {
	    GetUnpackedRange( i )->PreVisit1( func, data );
	}
    }
    if( pragmas ) {
	pragmas->PreVisit1( func, data );
    }
    if( attributes ) {
	attributes->PreVisit1( func, data );
    }
    if( dataType ) {
	dataType->PreVisit1( func, data );
    }
}

/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CDecl::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    for( int i = 0; i < GetNumberOfUnpackedDimensions(); ++i ) {
	if( GetUnpackedRange( i ) ) {
	    GetUnpackedRange( i )->PostVisit1( func, data );
	}
    }
    if( pragmas ) {
	pragmas->PostVisit1( func, data );
    }
    if( attributes ) {
	attributes->PostVisit1( func, data );
    }
    if( dataType ) {
	dataType->PostVisit1( func, data );
    }
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CDecl::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    for( int i = 0; i < GetNumberOfUnpackedDimensions(); ++i ) {
	if( GetUnpackedRange( i ) ) {
	    SetUnpackedRange( i, GetUnpackedRange( i )->PostSubVisit1( func, data ) );
	}
    }
    if( pragmas ) {
	pragmas = pragmas->PostSubVisit1( func, data );
    }
    if( attributes ) {
	attributes = attributes->PostSubVisit1( func, data );
    }
    if( dataType ) {
	dataType->PostSubVisit1( func, data );
    }
}

