/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cenumtype.h
 *		- abstract class for enum types 
 *
 *
 ******************************************************************************
 */

#ifndef CENUMTYPE_H
#define CENUMTYPE_H

#include <list>
#include "cdatatype.h"
#include "cnode.h"


/**
 * class for describing enumeration types
 */
class CEnumType : public CDataType
{
    CDataType* baseType;	///< base type of enum (may be NULL)
    CNode*     enumList;	///< list of enum definitions
public:
        /**
         * Create enum type instance
         */
        CEnumType( VarDataType_t t, CDataType* baseType, 
	 	   CNode* enumList, list<CNode*>& packedList ) 
	: CDataType(t,packedList), 
	  baseType(baseType),
	  enumList(enumList) {}
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap );
	/**
 	 * Get enum's element description list
 	 * \return enum's element description list
 	 */
	virtual CNode*  GetEnumList() const { return enumList; }
	/**
 	 * Get enum's base data type. NULL implies integer.
 	 * \return enum's base data type
 	 */
	virtual CDataType*  GetBaseType() const { return baseType; }
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const 
	{ return  baseType ? baseType->GetTwoState() : FALSE; }
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const 
	{ return baseType ? baseType->GetSigned() : TRUE; }
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const 
	{ return baseType ? baseType->GetNodeType() : eS; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero if vector or packed width of declaration 
 	 * is constant.
 	 */
	virtual int  IsBaseWidthConstant( void ) const
	{ return baseType ? baseType->IsPackedWidthConstant() : TRUE; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is volatile, 
 	 * ie depend upon parameters or variables.
 	 * \return non-zero of vector or packed width of declaration 
 	 * is volatile.
 	 */
	virtual int  IsBaseWidthVolatile( void ) const
	{ return baseType ? baseType->IsPackedWidthVolatile() : FALSE; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration can be evaluated.
 	 * \return non-zero of packed or vector width of declaration can 
 	 * be evaluated.
 	 */
	virtual int  IsBaseWidthEvaluateable( void ) const
	{ return baseType ? baseType->IsPackedWidthEvaluateable() : TRUE; }
        /**
         * Evaluate base width (sans packed dimensions) of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetBaseWidth( void ) const
	{ 
	    return baseType ? baseType->GetPackedWidth() : 32; 
	}
        /**
         * Get expression for datatype's base width (sans packed dimensions)
         * \return width expression
         */
        virtual CNode* GetBaseWidthExp() const
	{ return cINT32(GetBaseWidth()); }
        /**
         * Determine if complete data structure is a vector.
         * Vectors are multi-bit intergal types.
         * \return non-zero if complete data structure is a vector.
         */
        virtual int IsVector() const 
	{ return baseType ? baseType->IsVector() : TRUE; }
        /**
         * Determine if complete data structure is packed.
         * Packed data structures can be treated as a single vector.
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const 
	{ return baseType ? baseType->IsPacked() : TRUE; }
        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const;
        /**
         * Walk tree invoking callback on each node before children have 
         * been visited. callback can terminate transversal early.
         * callback returns 0 to terminate transversal early.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PreVisit1( int (*callback)(CNode*,void*), void* data );
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         */
        virtual void    PostVisit1( void (*callback)(CNode*,void*), void* data );
        /**
         * Walk tree invoking callback on each node after children have 
         * been visited. callback may modify tree and return a modified
         * subtree.
         * \param callback function to be called for each node
         * \param data opaque argument passed to callback function
         * \return modified child subtree.
         */
	virtual void PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data );
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CEnumType& o );
private:
	/*
 	 * disable copy constructors
 	 */
	CEnumType( const CEnumType& o );

};

#endif // CENUMTYPE_H

