/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cfunction.cpp
 *		- class definition of function
 *		  definition nodes
 *
 ******************************************************************************
 */


#include "main.h"
#include "cfunction.h"
#include "cportdir.h"
#include "cvar.h"
#include "cfref.h"
#include "cparam.h"
#include "cnode.h"
#include "cnet.h"
#include "ccontext.h"

extern CSymtab<CDecl> cuSymtab;
extern CObstack*      unitStack;

/*************************************
	Constructor
 *************************************/
CFunction::CFunction( CSymbol* name, Coord_t* aLoc, int automatic ) :
 	CBlock( name, aLoc, eFUNCTION ),
	portDirListValid( FALSE ),
	output( NULL ),
	automatic( automatic ),
	ports( NULL ),
	inlinePortDecls( FALSE )
{
}
/*************************************
	GetPortDirList
	- return portdir list
 *************************************/
vector<CPortDir*>* CFunction::GetPortDirList()
{
	if( !portDirListValid ) {
	    portDirListValid = TRUE;
	    portDirList.erase( portDirList.begin(), portDirList.end() );
	    ArgList2Vector<CPortDir*>(GetCodeList(),ePORT_DECL,0,portDirList);
	}
	return &portDirList;
}
/*******************************************
	PreVisit
	- Support routine for PreVisit 
 	  tree transveral routine
 *******************************************/

void CFunction::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    if( ports ) {
    	ports->PreVisit1( func, data );	
    }
    CBlock::PreVisit1( func, data );
}
/*******************************************
	PostVisit
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CFunction::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    if( ports ) {
    	ports->PostVisit1( func, data );	
    }
    CBlock::PostVisit1( func, data );
}
/*******************************************
	PostSubVisit
	- Support routine for PostSubVisit 
 	  tree transveral routine
 *******************************************/

void CFunction::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    if( ports ) {
    	ports = ports->PostSubVisit1( func, data );	
    }
    CBlock::PostSubVisit1( func, data );
}
/*************************************
	Dump
	- dump contents of function structure 
 *************************************/
void CFunction::Dump( FILE* f )
{
	if( output ) {
       	    fprintf( f, "FUNCTION: " );
	} else {
            fprintf( f, "TASK: " );
	}
        fprintf( f, "%s, defined in  ", GetName() );
        CBlock::Dump( f );
        fprintf( f, "\n" );
}
/*************************************
	SetFrefSymtab
	- set fref symbol table 
 *************************************/
void CFunction::SetPortSymtab( CSymtab<CDecl>& symtab )
{
	portSymtab = symtab;
}
/*************************************
	FixupPorts
	- declare registers for input ports 
 *************************************/

void CFunction::FixupPorts()
{
	/*
	 * declare all untyped ports as reg
	 * connect all port and forward references
	 * to their real decl
	 */
	RedeclarePorts();
	LinkFrefs( portSymtab, cuSymtab, frefList, inferVectors );
	Reconnect( GetCodeList() );
	Reconnect( ports );
}	

/**************************************
    Constant function evaluation routines
 *************************************/
CFunction* CFunction::FindFunction( CSymbol* symbol,
				CNode* args, CBlock* block )
{
    if( !block ) {
	return NULL;
    }
    CDecl* decl = block->GetSymtab().Lookup( symbol );
    if( !decl || decl->GetType() != eFUNCTION ) {
        return NULL;
    }
    return static_cast<CFunction*>(decl);
}
/**************************************
    Width
      - Get width of function.
	Return sane default if function cannot be resolved.
 *************************************/
int CFunction::Width( CSymbol* symbol, CNode* args, CBlock* block ) 
{ 
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return 32;
    }
    CDecl* output = function->GetOutput();
    return output->GetWidth(); 
}
/**************************************
    WidthExp
      - Get width of function as an expression.
	Return NULL default if function cannot be resolved.
 *************************************/
CNode* CFunction::WidthExp( CSymbol* symbol, CNode* args, CBlock* block ) 
{ 
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return NULL;
    }
    CDecl* output = function->GetOutput();
    return output->GetWidthExp(); 
}
/**************************************
    WidthConstant
      - Determine if width of function is constant. 
	ie dependent upon
 	constants and parameters only.
 *************************************/
int CFunction::WidthConstant( CSymbol* symbol, CNode* args, CBlock* block ) 
{ 
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return FALSE;
    }
    CDecl* output = function->GetOutput();
    return output->IsWidthConstant(); 
}
/**************************************
    WidthVolatile
      - Determine if width of function is volatile. ie dependent upon
 	variables and parameters.
 *************************************/
int CFunction::WidthVolatile( CSymbol* symbol, CNode* args, CBlock* block ) 
{ 
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return TRUE;
    }
    CDecl* output = function->GetOutput();
    return output->IsWidthVolatile(); 
}
/**************************************
    WidthEvaluateable
      - Determine if width of function is evaluateable.
 *************************************/
int CFunction::WidthEvaluateable( CSymbol* symbol, CNode* args, CBlock* block ) 
{ 
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return FALSE;
    }
    CDecl* output = function->GetOutput();
    return output->IsWidthEvaluateable(); 
}
/**************************************
    Type
      - Determine type of result of function.
 *************************************/
NodeType_t CFunction::Type( CSymbol* symbol, CNode* args, CBlock* block ) 
{ 
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return eB;
    }
    CDecl* output = function->GetOutput();
    return output->GetNodeType(); 
}

/**************************************
    Evaluateable
      - Determine if function can be evaluated
 *************************************/
int CFunction::Evaluateable( CSymbol* symbol, CNode* args, CBlock* block )
{
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	return FALSE;
    }
    NodeType_t nodeType = Type( symbol, args, block );
    if( nodeType != eR ) {
	if( !WidthEvaluateable( symbol, args, block ) ) {
	    return FALSE;
	}
	int width = Width( symbol, args, block );
	CVector v(width);
        try {
	    EvalVector( v, symbol, args, block );
	} catch(...) {
	    return FALSE;
	}
    } else {
	try {
	    EvalReal( symbol, args, block );
	} catch(...) {
	    return FALSE;
	}
    }
    return TRUE;
}
/**************************************
    EvaluateVector
      - Evaluate function as a vector and return result
 *************************************/
void CFunction::EvalVector( CVector& v, CSymbol* symbol, 
			    CNode* args, CBlock* block )
{
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	Coord_t coord = { 0, "???" };
	throw CExecuteUnimplementedConstruct( coord );
    }
    /*
     * - create a new context for function
     * - initialize output value to x
     * - attempt to execute function's code
     */
    CContext context; 
    CDecl* output = function->GetOutput();
    CVector* out = context.VPtr( output );
    function->EvaluateArgs( &context, args );
    context.Execute( function->GetCodeList() );
    /*
     * return function output value
     */
    v = *out;
}
/**************************************
    EvaluateReal
      - Evaluate function as a real and return result
 *************************************/
double CFunction::EvalReal( CSymbol* symbol, CNode* args, CBlock* block )
{
    CFunction* function = FindFunction( symbol, args, block );
    if( !function ) {
	Coord_t coord = { 0, "???" };
	throw CExecuteUnimplementedConstruct( coord );
    }
    /*
     * create a new context for function and attempt
     * to execute function's code
     */
    CContext context;
    context.Execute( function->GetCodeList() );
}
/**************************************
    EvaluateArgs
      - Evaluate function arguments and save values
        in port variables.
 *************************************/
void CFunction::EvaluateArgs( CContext* context, CNode* args )
{
    set<NodeOp_t>  exclude;
    exclude.insert( eVRQ );
    exclude.insert( ePRAGMA );
    exclude.insert( eCOMMENT );
    vector<CNode*> argList;
    EList2VectorExclude( args, exclude, argList );
    vector<CNode*>::iterator argPtr;
    for( argPtr = argList.begin(); argPtr != argList.end(); ++argPtr ) {
	if( (*argPtr)->IsVolatile() || !(*argPtr)->IsEvaluateable() ) {
	    Coord_t coord = { 0, "???" };
	    throw CExecuteArgEvaluationError( coord );
	}
    }
}

