/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cfunction.hpp
 *		- class definition of function
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CFUNCTION_HPP
#define CFUNCTION_HPP

#include <stdio.h>
#include <vector> 
#include <list> 
#include "glue.h"
#include "cdecl.h"
#include "csymtab.h"
#include "cblock.h"
#include "cnode.h"
#include "cfref.h"

class CContext;


/**
 * Declaration object for functions and tasks.
 */
class CFunction: public CBlock
{
private:
	CSymtab<CDecl>	  portSymtab; 	    ///< symbol table for ports
	int 		  portDirListValid; ///< flag indicating portDirList
					    ///< needs to be rebuilt.
	vector<CPortDir*> portDirList;	    ///< cached list of portdirs's
	CDecl*		  output;	    ///< function output declaration,
					    ///< NULL for tasks.
	int 		  automatic;	    ///< non-zero if function is automatic.
	CNode*		  ports;	    ///< list of ports
	int		  inlinePortDecls;  ///< non-zero if port list is inline.
public:
	/**
 	 * Create function/task declaration object.
 	 * \param symbol name of function/task.
 	 * \param aLoc file coordinates of declaration.
 	 * \param automatic non-zero if function is automatic.
 	 */
	CFunction( CSymbol* symbol, Coord_t* aLoc, int automatic );
	/**
 	 * Dump function info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Get array of port declarations
 	 * \return vector of portdir declarations.
 	 */
	virtual vector<CPortDir*>* GetPortDirList();
	/**
 	 * Set function port list.
 	 * \param n port list.
 	 */
	void	SetPorts( CNode* n ) { ports = n; }
	/**
 	 * Get function port list.
 	 * \return port list.
 	 */
	CNode*	GetPorts() { return ports; }
	/**
 	 * Get inline declaration attribute.
 	 * \return non-zero if port declarations are inline.
 	 */
	int	InlinePortDecls() { return inlinePortDecls; }
	/**
 	 * Set inline declaration attribute.
 	 * \param v non-zero if port declarations are inline.
 	 */
	void	InlinePortDecls( int v ) { inlinePortDecls = v; }
	/**
 	 * Set output declaration for function.
 	 * \param output output declaration.
 	 */
	void	SetOutput( CDecl* output ) { this->output = output; }
	/**
 	 * Get output declaration.
 	 * \return output declaration.
 	 */
	CDecl*	GetOutput() { return output; }
	/**
 	 * Get automatic attribute.
 	 * \return non-zero if function has been declare with 
 	 *         automatic attribute.
 	 */
	int	GetAutomatic() { return automatic; }

	/**
 	 * Get width of function.
 	 * \todo This routine needs to be implemented to support proper 
 	 * evaluation of constant functions.
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return width of function.
 	 */
	static int Width( CSymbol* symbol, CNode* args, CBlock* block ) ;
	/**
 	 * Get width of function as an expression.
 	 * \todo This routine needs to be implemented to support proper 
 	 * evaluation of constant functions.
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return width expression.
 	 */
	static CNode* WidthExp( CSymbol* symbol, CNode* args, CBlock* block ) ;
	/**
 	 * Determine if width of function is constant. ie dependent upon
 	 * constants and parameters only.
 	 * \todo This routine needs to be implemented to support proper 
 	 * evaluation of constant functions.
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return non-zero of function's width is constant.
 	 */
	static int WidthConstant( CSymbol* symbol, CNode* args, 
							CBlock* block );
	/**
 	 * Determine if width of function is volatile. ie dependent upon
 	 * variables and parameters.
 	 * \todo This routine needs to be implemented to support proper 
 	 * evaluation of constant functions.
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return non-zero of function's width is volatile.
 	 */
	static int WidthVolatile( CSymbol* symbol, CNode* args, 
							CBlock* block );
	/**
 	 * Determine if width of function is evaluateable.
 	 * \todo This routine needs to be implemented to support proper 
 	 * evaluation of constant functions.
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return non-zero of function's width is evaluateable.
 	 */
	static int WidthEvaluateable( CSymbol* symbol, CNode* args, 
		                      CBlock* block );
	/**
 	 * Determine if type of result of function.
 	 * \todo This routine needs to be implemented to support proper 
 	 * evaluation of constant functions.
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return type of return result of function.
 	 */
	static NodeType_t Type( CSymbol* symbol, CNode* args, CBlock* block );

	/**
 	 * Determine if function can be evaluated
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return non-zero if function can be evaluated
 	 */
	static int Evaluateable( CSymbol* symbol, CNode* args, CBlock* block );

	/**
 	 * Evaluate function as a vector and return result
 	 * \param v container for result
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 */
	static void EvalVector( CVector& v, CSymbol* symbol, 
					CNode* args, CBlock* block );

	/**
 	 * Evaluate function as a real and return result
 	 * \param symbol name of function.
 	 * \param args function argument list
 	 * \param block scope of function instantation
 	 * \return result of function evaluation
 	 */
	static double EvalReal( CSymbol* symbol, CNode* args, CBlock* block );

	/**
 	 * These should never be called by plugins.
 	 * \cond
 	 */
	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
	void	FixupPorts();
	void 	SetPortSymtab( CSymtab<CDecl>& symtab );
protected:
	virtual void CodeListChanged() { portDirListValid = FALSE; }
private:
	static CFunction* FindFunction( CSymbol* symbol,
				CNode* args, CBlock* block );
        void EvaluateArgs( CContext* context, CNode* args );
	/**
 	 * \endcond
 	 */
};

#endif // CFUNCTION_HPP
