/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cgate.hpp
 *		- class definition a gate instance
 *
 ******************************************************************************
 */

#ifndef CGATE_HPP
#define CGATE_HPP

#include "glue.h"
#include "cdecl.h"
#include "cnode.h"



/**
 * Gate declaration object
 */
class CGate: public CDecl
{
private:
	CSymbol*	type;		///< name of gate type
	CNode*		delayControl;  	///< delay control for gate
	CNode*		arguments;	///< gate arguments
	Strength_t	strength0;	///< 0 drive strength
	Strength_t	strength1;	///< 1 drive strength
public:
	/**
 	 * Create a gate declaration
 	 * \param aSymbol gate name.
 	 * \param type gate type.
 	 * \param s0 0 drive strength.
 	 * \param s1 1 drive strength.
 	 * \param delayControl delay control expression.
 	 * \param arguments gate arguments.
 	 * \param aLoc file coordinates of declaration.
 	 */
	CGate( CSymbol* aSymbol, CSymbol* type, 
	       Strength_t s0, Strength_t s1,
	       CNode* delayControl, CNode* arguments, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Get gate's type.
 	 * \return gate's type.
 	 */
	CSymbol* GetType( void );
	/**
 	 * Get gate's delay control.
 	 * \return delay control, may be NULL.
 	 */
	CNode*	 GetDelayControl( void );
	/**
 	 * Get gate's arguments.
 	 * \return argument list, may be NULL.
 	 */
	CNode*	 GetArguments( void );
	/**
 	 * Set gate's arguments
 	 * \param n arguments.
 	 */
	void	 SetArguments( CNode* n ) { arguments = n; }
	/**
  	 * Dump gate info to file descriptor
  	 * \param f file descriptor.
  	 */
	virtual void Dump( FILE* f );
	/**
 	 * Get 0 drive stength.
 	 * \return drive strength.
 	 */
	Strength_t Strength0() { return strength0; }
	/**
 	 * Set 0 drive strength.
 	 * \param pair drive strength.
 	 */
	void     Strength0( Strength_t pair ) { strength0 = pair; }
	/**
 	 * Get 1 drive stength.
 	 * \return drive strength.
 	 */
	Strength_t Strength1() { return strength1; }
	/**
 	 * Set 1 drive strength.
 	 * \param pair drive strength.
 	 */
	void     Strength1( Strength_t pair ) { strength1 = pair; }
	/**
 	 * \cond
 	 * These should never be directly accessed.
 	 */
	virtual void PreVisit1( int (*func)(CNode*,void*), void* data );
	virtual void PostVisit1( void (*func)(CNode*, void*), void* data );
	virtual void PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
protected:
	/**
 	 * Deep copy of declaration.
 	 * \param heap heap to use for allocation.
 	 * \param d declaration to copy.
 	 */
	void Copy( CObstack* heap, CGate& d );
private:
	/*
 	 * disable copy constructor
 	 */
	CGate( const CGate& );
	/**
 	 * \endcond
 	 */	
};

#endif // CGATE_HPP
