/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cinstance.cpp
 *		- class for managing instances of
 *		  module, macromodule and	
 *		  primitive definitions 
 *
 *
 ******************************************************************************
 */


#include "cnode.h"
#include "cinstance.h"
#include "csymbol.h"
#include "cobstack.h"
#include "cmodule.h"
#include "csymtab.h"

extern CObstack*      permStack;
extern CSymtab<CDecl> definitionSymtab;
extern CSymtab<CDecl> cuSymtab;



/*******************************************
	Constructor
 *******************************************/
	CInstance::CInstance( CSymbol* name, 
		CModule* aDefinition, 
		CNode* aParameters,
		CNode* aArguments,
		Coord_t* aLoc ) : 
		CDecl( name, aLoc, eINSTANCE, NULL, eFLAG_ARRAY )
{

	timestamp = 0;
	cachedTimestamp = timestamp-1;
	definition = aDefinition;


	isGate = FALSE;
	const char* dName = aDefinition->GetName();
	if( !strcmp( dName, "and" ) ||
	    !strcmp( dName, "nand" ) ||
	    !strcmp( dName, "or" ) ||
	    !strcmp( dName, "nor" ) ||
	    !strcmp( dName, "xor" ) ||
	    !strcmp( dName, "xnor" ) ||
	    !strcmp( dName, "buf" ) ||
	    !strcmp( dName, "not" ) ||
	    !strcmp( dName, "cmos" ) ||
	    !strcmp( dName, "nmos" ) ||
	    !strcmp( dName, "pmos" ) ||
	    !strcmp( dName, "rcmos" ) ||
	    !strcmp( dName, "rnmos" ) ||
	    !strcmp( dName, "rpmos" ) ||
	    !strcmp( dName, "tran" ) ||
	    !strcmp( dName, "tranif0" ) ||
	    !strcmp( dName, "tranif1" ) ||
	    !strcmp( dName, "rtran" ) ||
	    !strcmp( dName, "rtranif0" ) ||
	    !strcmp( dName, "rtranif1" ) ||
	    !strcmp( dName, "bufif0" ) ||
	    !strcmp( dName, "bufif1" ) ||
	    !strcmp( dName, "notif0" ) ||
	    !strcmp( dName, "notif1" ) ) {
	    isGate = TRUE;
	}

	parameters = aParameters;
	SetArguments( aArguments );
}

/****************************************************
	Clone
	- return a deep copy of declaration
*****************************************************/
CDecl* CInstance::Clone( CObstack* heap )
{
     CInstance* clone = new(heap) CInstance( GetSymbol(), 
		GetDefinition(), 
    	        parameters ? parameters->Clone(heap) : NULL,
     		arguments ? arguments->Clone(heap) : NULL,
		GetCoord() );
     return clone;
}

/*******************************************
	GetDefinition	
	- return defining module object 
 *******************************************/

CModule* CInstance::GetDefinition( void )
{
	return definition;
}
/*******************************************
	GetParameters	
	- return parameters 
 *******************************************/

CNode* CInstance::GetParameters( void )
{
	return parameters;
}
/*******************************************
	SetParameters	
	- set parameters
 *******************************************/

void CInstance::SetParameters( CNode* n )
{
	parameters = n;
}

/*******************************************
	PushArgs	
	- build argument list
 *******************************************/
void CInstance::PushArgs( CNode* n, int level )
{
	if( n == NULL ) {
	    if( level != 0 ) {
		argumentVector.push_back( n );
	    }
	    return;
	}
	switch( n->GetOp() ) {
	case eLIST:
	    PushArgs( n->Arg<CNode*>(0), level+1 );
	    PushArgs( n->Arg<CNode*>(1), level+1 );
	    break;
	case eARG:
	    argumentVector.push_back( n );
	    break;
	default:
	    break;
	}	    
}
/*******************************************
	SetArguments	
	- set arguments 
 *******************************************/
void CInstance::SetArguments( CNode* n )
{
    arguments = n;
    timestamp++;
}
/*******************************************
	GetArguments	
	- return arguments 
 *******************************************/

CNode* CInstance::GetArguments( void )
{
	return arguments;
}
/*******************************************
	GetArgumentVector	
	- return argument vector rebuilding
	  the list if needed.
 *******************************************/

vector<CNode*>* CInstance::GetArgumentVector( void )
{
    if( timestamp != cachedTimestamp ) {
        argumentVector.erase( argumentVector.begin(),
			  argumentVector.end() );
        PushArgs( arguments, 0 );
	cachedTimestamp = timestamp;
    }
    return &argumentVector;
}

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CInstance::PreVisit1( int (*func)(CNode*,void*), void* data )
{
   if( GetParameters() ) {
       GetParameters()->PreVisit1( func, data );
   }
   if( GetArguments() ) {
       GetArguments()->PreVisit1( func, data );
   }
   CDecl::PreVisit1( func, data );
}

/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CInstance::PostVisit1( void (*func)(CNode*,void*), void* data )
{
   if( GetParameters() ) {
       GetParameters()->PostVisit1( func, data );
   }
   if( GetArguments() ) {
       GetArguments()->PostVisit1( func, data );
   }
   CDecl::PostVisit1( func, data );
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CInstance::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
   if( GetParameters() ) {
       SetParameters( GetParameters()->PostSubVisit1( func, data ) );
   }
   if( GetArguments() ) {
       SetArguments( GetArguments()->PostSubVisit1( func, data ) );
   }
   CDecl::PostSubVisit1( func, data );
}
/*******************************************
	Dump	
	- print out instance info 
 *******************************************/
void	CInstance::Dump( FILE* f )
{
	
	fprintf( f, "INSTANCE<%s>: ", definition->GetName() );
	CDecl::Dump( f );
	fprintf( f, "\n" );
}


