/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cinstance.hpp
 *		- class definition an instance
 *		  of module, macromodules, and primitive 
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CINSTANCE_HPP
#define CINSTANCE_HPP

#include <vector>
#include "glue.h"
#include "cdecl.h"


class CModule;

/**
 * Declaration object for module and gate instances.
 */
class CInstance: public CDecl
{
private:
	CModule*	definition;	///< module definition
	CNode*		parameters;	///< list of module parameters
	CNode*		arguments;	///< list of module arguments
	int		isGate;		///< non-zero if instance is a gate
	vector<CNode*>  argumentVector; ///< used for supplying list of arguments.
	unsigned long long timestamp;   ///< current timestamp used to 
					///< track changes in the argument list.
	unsigned long long cachedTimestamp; ///< timestamp last time arg list
					    ///< was generated.
public:
	/**
 	 * Create an instance.
 	 * \param aSymbol name of instance.
 	 * \param definition module decl for udp/module instantiation
 	 * \param parameters list of parameters.
 	 * \param arguments list of arguments.
 	 * \param aLoc file coordinates of instance.
 	 */
	CInstance( CSymbol* aSymbol, CModule* definition, 
			   CNode* parameters, CNode* arguments, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Get module definition.
 	 * \return module definition for instance.
 	 */
	CModule* GetDefinition( void );
	/**
 	 * Get instance parameters.
 	 * \return list of parameters.
 	 */
	CNode*	 GetParameters( void );
	/**
 	 * Set instance parameters.
 	 * \param param instance parameters.
 	 */
	void	 SetParameters( CNode* param );
	/**
 	 * Get instance arguments.
 	 * \return instance arguments.
 	 */
	CNode*	 GetArguments( void );
	/**
 	 * Set instance arguments.
 	 * \param n instance arguments.
 	 */
        void	 SetArguments( CNode* n );
	/**
 	 * Get array of argument nodes.
 	 * \return pointer to vector containing argument nodes.
 	 */
	vector<CNode*>* GetArgumentVector( void );
	/**
 	 * Determine if instance is a gate.
 	 * \return non-zero if instance is a gate.
 	 */
	int	 IsGate() { return isGate; }
	/**
 	 * Dump instance info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * These should never be call directly
 	 * \cond
 	 */
	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
private:
	void    PushArgs( CNode* n, int level );
	CInstance( const CInstance& );
	/**
 	 * \endcond
 	 */
};

#endif // CINSTANCE_HPP
