/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cinterfacetype.cc
 *		- abstract class for interface types 
 *
 *
 ******************************************************************************
 */

#include <list>
#include "cinterfacetype.h"
#include "cvar.h"
#include "cnode.h"

/**
 * Create interface type instance
 */
CInterfaceType::CInterfaceType( CSymtab<CDecl>& symtab,
	CSymbol* symbol, CSymbol* modportSymbol, list<CNode*>& rangeList )
  : CDataType(eINTERFACE_TYPE, rangeList),
    symtab(symtab),
    symbol(symbol),
    modportSymbol(modportSymbol)
{
}
/**
 * Create a new copy with a deep copy. 
 * \param heap heap to use for allocation.
 */
CDataType* CInterfaceType::Clone( CObstack* heap )
{
    list<CNode*> emptyList;
    CInterfaceType* clone = new(heap) CInterfaceType(symtab, GetSymbol(), GetModportSymbol(), emptyList);
    clone->Copy( heap, *this ); 
    return clone;
}

/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CInterfaceType::Copy( CObstack* heap, const CInterfaceType& o )
{
    CDataType::Copy( heap, o );
}

/**
 * Dump data type info to file descriptor.
 * \param f file descriptor.
 */
void CInterfaceType::Dump( FILE* f ) const
{
    fprintf( f, "interface " );
}

/**
 * Get declaration's 2 state property.
 * \return value of 2 state property.
 */
int CInterfaceType::GetTwoState() const
{
    return false;
}
/**
 * Determine if packed or vector width of declaration is constant, 
 * ie dependent upon only constants and parameters.
 * \return non-zero if vector or packed width of declaration 
 * is constant.
 */
int CInterfaceType::IsBaseWidthConstant() const
{
    return false;
}

/**
 * Determine if packed or vector width of declaration is volatile, 
 * ie depend upon parameters or variables.
 * \return non-zero of vector or packed width of declaration 
 * is volatile.
 */
int CInterfaceType::IsBaseWidthVolatile() const
{
    return false;
}
/**
 * Determine if packed or vector width of declaration can be evaluated.
 * \return non-zero of packed or vector width of declaration can 
 * be evaluated.
 */
int CInterfaceType::IsBaseWidthEvaluateable() const
{
    return false;
}

/**
 * Evaluate packed or vector width of declaration.
 * \return packed or vector width of declaration.
 */
INT32 CInterfaceType::GetBaseWidth( void ) const
{
    return 0;
}

/**
 * Get expression for datatype's overall packed or vector width
 * \return width expression
 */
CNode* CInterfaceType::GetBaseWidthExp() const
{
    return NULL;
}
/**
 * Find declaration for member
 * \param member  member to search for.
 * \return matching declaration, NULL if not found
 */

CDecl* CInterfaceType::FindMember( CSymbol* member )
{
    CDecl* decl = symtab.Lookup( member );
    return decl;
}
/**
 * Walk tree invoking callback on each node before children have 
 * been visited. callback can terminate transversal early.
 * callback returns 0 to terminate transversal early.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CInterfaceType::PreVisit1( int (*callback)(CNode*,void*), void* data )
{
}
/**
 * Walk tree invoking callback on each node after children have 
 * been visited.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CInterfaceType::PostVisit1( void (*callback)(CNode*,void*), void* data )
{
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited. callback may modify tree and return a modified
 * subtree.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 * \return modified child subtree.
 */
void CInterfaceType::PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data )
{
}
