/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cmacro.hpp
 *		- class definition for macro 
 *
 *
 ******************************************************************************
 */

#ifndef CMACRO_HPP
#define CMACRO_HPP

#include "glue.h"
#include "cdecl.h"



class CSymbol;

/**
 * Declaration object for preprocessor macros.
 */
class CMacro : public CDecl
{
private:
	char*	body;	    ///< body of macro
	char**	argList;    ///< array of arguments to macro
	int	complex:1;  ///< flag indicating macro has arguments
	int	defined:1;  ///< flag indicating macro has been defined
	int	zone;	    ///< preprocessor zone number
public:
	/**
 	 * Find macro definition. Create an undefined entry
 	 * if it doesn't exist.
 	 * \param aName char pointer to macro name.
 	 * \return macro object.
 	 */
	static CMacro*	LookupMacro( const char* aName );

	/**
 	 * Create a macro declaration object.
 	 * \param aSymbol name of macro.
 	 * \param aLoc file coordinate of definition.
 	 */
	CMacro( CSymbol* aSymbol, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Get body of macro.
 	 * \return macro body.
 	 */
	char*	GetBody( void );
	/**
 	 * Set body of macro.
 	 * \param aBody body of macro.
 	 */
	void	SetBody( char* aBody );
	/**
 	 * Get preprocessor zone number for last macro
 	 * definition.
 	 * \return preprocessor zone number.
 	 */
	int	GetZone( void ) { return zone; }
	/**
 	 * Set preprocessor zone number.
 	 * \param aZone preprocessor zone number.
 	 */
	void	SetZone( int aZone ) { zone = aZone; }
	/**
 	 * Set complex flag.
 	 * \param aComplex non-zero if macro has arguments.
 	 */
	void	SetComplex( int aComplex );
	/**
 	 * Get complex flag.
 	 * \return non-zero if macro has arguments.
 	 */
	int	IsComplex( void );
	/**
 	 * Get macro arguments.
 	 * \return pointer to array of macro arguments.
	 */
	char** 	GetArgList( void );
	/**
 	 * Set macro arguments.
 	 * \param argList pointer to array of macro arguments. Array is
 	 *        not copied.
	 */
	void	SetArgList( char** argList );
	/**
 	 * Dump macro info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Get defined flag.
 	 * \return non-zero if macro has been defined.
 	 */
	int	IsDefined( void );
	/**
 	 * Set defined flag.
 	 * \param aDefined non-zero if macro has been defined.
 	 */
	void	SetDefined( int aDefined );
protected:
	/**
 	 * Deep copy of declaration.
 	 * \param heap heap to use for allocation.
 	 * \param d declaration to copy.
 	 */
	void Copy( CObstack* heap, CMacro& d );
private:
	/*
 	 * disable copy constructor
 	 */
	CMacro( const CMacro& );
};

#endif // CMACRO_HPP
