/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cmodule.hpp
 *		- class definition of module, macromodules, and primitive 
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CMODULE_HPP
#define CMODULE_HPP

#include <stdio.h>
#include <vector> 
#include "glue.h"
#include "cdecl.h"
#include "csymtab.h"
#include "cblock.h"
#include "cnode.h"
extern "C" {
    #include "vpi_user.h"
}


class CInstance;
class CParam;
class CVar;
class CPortDir;
class CFref;
class CNode;


class CModule: public CBlock
{
private:
	int 		  isMacroModule;   ///< non-zero if macro module.
	bool 		  isInterface;     ///< true if interface
	bool 		  isModport;       ///< true if modport
	int		  isPrimitive;     ///< non-zero if udp.
	vector<CInstance*> instanceList;   ///< list of module instances.
	vector<CFunction*> functionList;   ///< list of function/tasks
	vector<CFref*>	  frefList;	   ///< list of forward references.
	int 		  portListValid;   ///< non-zero if portList needs to
					   ///< be rebuilt.
	vector<CPort*>	  portList;	   ///< cached list of ports.
	int		  portDirListValid;///< non-zero if portDirList needs
					   ///< to be rebuilt.
	vector<CPortDir*>  portDirList;	   ///< cached list of portDir.
	CNode*		  ports;	   ///< list of ports.
	int 		  defined;	   ///< non-zero if module has been
					   ///< defined.
	CSymtab<CDecl>	  frefSymtab; 	   ///< symbol table for frefs
	CSymtab<CDecl>	  portSymtab; 	   ///< symbol table for ports
	int		  moduleDirectlyDefined; ///< zero if file containing 
						 ///< module was found by 
						 ///< search resolution.
	int		  moduleIncluded;  ///< true if module has been 
					   ///< included in output stream.
	int		  cell;		   ///< `cell compiler directive
	int		  protect;	   ///< `protect compiler directive
	int		  hasTimescale;    ///< `timescale has been declared
	int		  precision;	   ///< timescale precision.
	int		  unit;		   ///< timescale unit.
	int		  defaultNetType;  ///< `defaultNetType compiler 
					   ///< directive
	int		  unconnectedDrive;///< `unconnectedDrive compiler 
					   ///< directive
	int		  inlineParamDecls;///< non-zero if module has inline 
					   ///< parameter declarations
	int		  inlinePortDecls; ///< non-zero if module has inline
					   ///< port declarations
	CNode*		  packageImportDecls;  ///< package import declarations
public:
	/**
 	 * Find module declaration from name. If it doesn't exist
 	 * create an undefined module object and tie it to name.
 	 * \param symtab symbol table to search or insert into.
 	 * \param aName name of module.
 	 */
	static CModule*	LookupModule( CSymtab<CDecl>& symtab, char* aName );
	/**
 	 * Create a module declaration.
 	 * \param symbol name of module.
 	 * \param aLoc file coordinates of declaration.
	 */
	CModule( CSymbol* symbol, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap ) { MASSERT(FALSE); }
	/**
 	 * Set inline parameter attribute.
 	 * \param v non-zero if parameters have been declared inline.
 	 */
	void	InlineParamDecls( int v ) { inlineParamDecls = v; }
	/**
 	 * Get inline parameter attribute.
 	 * \return non-zero if parameters have been declared inline.
 	 */
	int	InlineParamDecls() { return inlineParamDecls; }
	/**
 	 * Set inline port attribute.
 	 * \param v non-zero if port have been declared inline.
 	 */
	void	InlinePortDecls( int v ) { inlinePortDecls = v; }
	/**
 	 * Get inline port attribute.
 	 * \return non-zero if port have been declared inline.
 	 */
	int	InlinePortDecls() { return inlinePortDecls; }
	/**
 	 * Set package import declarations.
 	 * \param v list of declarations
 	 */
	void	PackageImportDecls( CNode* v ) { packageImportDecls = v; }
	/**
 	 * Get package import declarations
 	 * \return list of declarations
 	 */
	CNode*	PackageImportDecls() { return packageImportDecls; }
	/**
 	 * Set directly defined attribute.
 	 * \param v non-zero if module has been defined in a 
 	 *          input file directly provide by the command line.
 	 */
	void	ModuleDirectlyDefined( int v ) { moduleDirectlyDefined = v; }
	/**
 	 * Get directly defined attribute.
 	 * \return non-zero if module has been defined in a 
 	 *          input file directly provide by the command line.
 	 */
	int	ModuleDirectlyDefined() { return moduleDirectlyDefined; }
	/**
 	 * Get defined attribute.
 	 * \return non-zero if module has been defined.
 	 */
        int 	IsDefined( void ) { return defined; }
	/**
 	 * Set defined attribute.
 	 * \param flag non-zero if module has been defined.
 	 */
	void 	SetDefined( int flag ) { defined = flag; }
	/**
 	 * Set included attribute.
 	 * \param v non-zero if module declaration has been included in 
 	 *          a compilation unit.
 	 */
	void	ModuleIncluded( int v ) { moduleIncluded = v; }
	/**
 	 * Get included attribute.
 	 * \return non-zero if module declaration has been included in 
 	 *          a compilation unit.
 	 */
	int	ModuleIncluded() { return moduleIncluded; }
	/**
 	 * Get macro module attribute.
 	 * \return non-zero if module has been declare as a macro module.
 	 */
	int	IsMacroModule( void );
	/**
 	 * Set macro module attribute.
 	 * \param flag non-zero if module has been declare as a macro module.
 	 */
	void	SetMacroModule( int flag );
	/**
 	 * Get interface attribute.
 	 * \return true if this is an interface
 	 */
	bool	IsInterface( void );
	/**
 	 * Set interface attribute.
 	 * \param flag true if is an interface.
 	 */
	void	SetInterface( bool flag );
	/**
 	 * Get modport attribute.
 	 * \return true if this is an modport
 	 */
	bool	IsModport( void );
	/**
 	 * Set modport attribute.
 	 * \param flag true if is an modport.
 	 */
	void	SetModport( bool flag );
	/**
 	 * Get primitive attribute.
 	 * \return non-zero if module is a udp.
 	 */
	int	IsPrimitive( void );
	/**
 	 * Set primitive attribute.
 	 * \param flag non-zero if module is a udp.
 	 */
	void	SetPrimitive( int flag );
	/**
 	 * Dump macro info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Set module port list.
 	 * \param ports port list.
 	 */
	void	SetPorts( CNode* ports );
	/**
 	 * Get modules port list.
 	 * \return port list.
 	 */
	CNode*  GetPorts() { return ports; }
	/**
 	 * Get list of functions/tasks
 	 * \return vector of functions/tasks.
	 */
	virtual vector<CFunction*>* GetFunctionList();
	/**
 	 * Get list if module instances.
 	 * \return vector of instances.
	 */
	virtual vector<CInstance*>* GetInstanceList();
	/**
 	 * Get list if module ports.
 	 * \return vector of ports.
	 */
	virtual vector<CPort*>*	    GetPortList();
	/**
 	 * Get list if module ports declarations.
 	 * \return vector of ports declarations.
	 */
	virtual vector<CPortDir*>*  GetPortDirList();
	/**
 	 * Find declaration in modules scope for given symbol.
 	 * \param sym symbol to lookup.
 	 * \return declaration with given name, NULL if not found.
 	 */
	CDecl*  FindDecl( CSymbol* sym );

	/**
 	 * Get cell attribute.
 	 * \return non-zero if module has been tagged as a cell.
 	 */
	int     Cell() { return cell; }
	/**
 	 * Set cell attribute.
 	 * \param v non-zero if module has been tagged as a cell.
 	 */
	void    Cell( int v ) { cell = v; }
	/**
 	 * Get protect attribute.
 	 * \return non-zero if module has been tagged as protected.
 	 */
	int	Protect() { return protect; }
	/**
 	 * Set protect attribute.
 	 * \param v non-zero if module has been tagged as protected.
 	 */
	void	Protect( int v ) { protect = v; }
	/**
 	 * Get timescale precision.
 	 * \return precision: 0: s, -3: ms, -6: us, -9: ns, -12: ps, -15: fs
 	 */
	int	Precision() { return precision; }
	/**
 	 * Set timescale precision.
 	 * \param v precision: 0: s, -3: ms, -6: us, -9: ns, -12: ps, -15: fs
 	 */
	void	Precision( int v ) { precision = v; hasTimescale = TRUE; }
	/**
 	 * Get timescale unit.
 	 * \return unit: 0: s, -3: ms, -6: us, -9: ns, -12: ps, -15: fs
 	 */
	int	Unit() { return unit; }
	/**
 	 * Set timescale unit.
 	 * \param v  precision: 0: s, -3: ms, -6: us, -9: ns, -12: ps, -15: fs
 	 */
	void	Unit( int v ) { unit = v; hasTimescale = TRUE; }
	/**
 	 * Get timescale present attribute.
 	 * \return non-zero if timescale has been specified.
 	 */
	int	HasTimescale() { return hasTimescale; }
	/**
 	 * Get default net type attribute
 	 * \return net type: vpiWire, vpiTri, vpiTri0, vpiWand, vpiTriAnd,
 	 * 		     vpiTriAnd, vpiTri1, vpiWor, vpiTriOr, VpiTriReg
 	 */
	int	DefaultNetType() { return defaultNetType; }
	/**
 	 * Set default net type attribute
 	 * \param v net type: vpiWire, vpiTri, vpiTri0, vpiWand, vpiTriAnd,
 	 * 		     vpiTriAnd, vpiTri1, vpiWor, vpiTriOr, VpiTriReg
 	 */
	void	DefaultNetType( int v ) { defaultNetType = v; }
	/**
 	 * Get unconnected drive attribute.
 	 * \return unconnected drive attribute: vpiHighZ, vpiPull0, vpiPull1
 	 */
	int	UnconnectedDrive() { return unconnectedDrive; }
	/**
 	 * Set unconnected drive attribute.
 	 * \param v unconnected drive attribute: vpiHighZ, vpiPull0, vpiPull1
 	 */
	void	UnconnectedDrive( int v ) { unconnectedDrive = v; }
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add function/task to module.
 	 * \param function pointer to function/task object
 	 */
	virtual void Add( CFunction* function );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add instance to block.
 	 * \param instance pointer to instance object
 	 */
	virtual void Add( CInstance* instance );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add fref to block.
 	 * \param fref pointer to forward reference object
 	 */
	virtual void Add( CFref* fref  );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add variable to block.
 	 * \param var pointer to register object
 	 */
	virtual void Add( CVar* var ) { CBlock::Add(var); }
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add net to block.
 	 * \param net pointer to net object
 	 */
	virtual void Add( CNet* net ) { CBlock::Add(net); }
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add parse tree to block.
 	 * \param n parse tree.
 	 */
	virtual void Add( CNode* n );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add parameter to block.
 	 * \param param pointer to parameter object
 	 */
	virtual void Add( CParam* param ) { CBlock::Add(param); }

	/**
 	 * These should never be called by plugins.
 	 * \cond
 	 */
	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
	void	FixupPorts( void );
	void 	SetFrefSymtab( CSymtab<CDecl>& symtab );
	void 	SetPortSymtab( CSymtab<CDecl>& symtab );
	CSymtab<CDecl>& GetPortSymtab() { return portSymtab; }
private:
	virtual void CodeListChanged() { portDirListValid = FALSE; }
  	CModule( const CModule& );
	/**
 	 * \endcond
 	 */
};

#endif // CMODULE_HPP
