/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cnet.hpp
 *		- class definition of net
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CNET_HPP
#define CNET_HPP

#include <stdio.h>
#include "glue.h"
#include "cdecl.h"
#include "cnode.h"
#include "cportdir.h"


class CNode;

/**
 * Declaration object for nets
 */
class CNet: public CDecl
{
private:
	CNode*	delay;		///< net delay expression.
	CPortDir* portdir;	///< associated portdir if net is a port
	int	scalared;	///< scalared verilog attribute
	int	vectored;	///< vectored verilog attribute
	int	undefined;	///< TRUE if undefined in source
	StrengthPair_t strength;///< drive strength
public:
	/**
 	 * Create a net declaration.
 	 * \param symbol declaration name.
 	 * \param aLoc file descriptor.
 	 * \param wireType wire type.
 	 * \param dataType net data type
 	 * \param undefined non-zero if net is not explicitly defined in source.
 	 */
	CNet( CSymbol* symbol, Coord_t* aLoc, Wire_t wireType, 
	      CDataType* dataType, int undefined );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Set portdir associated with net.
 	 * \param p to associate with net.
 	 */
	void    PortDir( CPortDir* p ) { portdir = p; }
	/**
 	 * Get portdir associated with net.
 	 * \return portdir associated with net.
 	 */
	CPortDir* PortDir() { return portdir; }
	/**
 	 * Set net delay.
 	 * \param aExp net delay expression.
 	 */
	void 	SetDelay( CNode* aExp ) { delay = aExp; }
	/**
 	 * Get net delay.
 	 * \return net delay expression.
	 */
	CNode*	GetDelay( void ) { return delay; }
	/**
 	 * Get undefined attribute. 
 	 * \return non-zero if net is undeclared in source code.
 	 */
	int	GetUndefined( void ) { return undefined; }
	/**
 	 * Set undefined attribute. 
 	 * \param value non-zero if net is undeclared in source code.
 	 */
	void	SetUndefined( int value ) { undefined = value; }
	/**
 	 * Get drive strength.
 	 * \return drive strength.
 	 */
	StrengthPair_t GetStrength() { return strength; }
	/**
 	 * Set drive strength.
 	 * \param s drive strength.
 	 */
	void	SetStrength( StrengthPair_t s ) { strength = s; }
	/**
 	 * Set vectored attribute.
 	 * \param v non-zero if net is vectored.
 	 */
	virtual void SetVectored( int v ) { vectored = v; }
	/**
 	 * Get vectored attribute.
 	 * \return non-zero if net is vectored.
 	 */
	virtual int  GetVectored() { return vectored; }
	/**
 	 * Set scalared attribute.
 	 * \param v non-zero if net is scalared.
 	 */
	virtual void SetScalared( int v ) { scalared = v; }
	/**
 	 * Get scalared attribute.
 	 * \return non-zero if net is scalared.
 	 */
	virtual int  GetScalared() { return scalared; }
	/**
 	 * Dump net info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
         * \cond
         * should not be called directly */
	virtual void PreVisit1( int (*func)(CNode*,void*), void* data );
	virtual void PostVisit1( void (*func)(CNode*, void*), void* data );
	virtual void PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
	/**
 	 * \endcond
 	 */
protected:
	/**
 	 * Create a deep copy of given declaration.
 	 * \param heap heap to use for allocation.
 	 * \param net declaration to copy from.
 	 */
	void Copy( CObstack* heap, const CNet& net );
private:
	/*
 	 * disable copy constructor
 	 */
	CNet( const CNet& net );
};


#endif // CNET_HPP
