/******************************************************************************
 * 	Copyright (c) 1997-2004 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   compile.cpp
 *		- library of compile related routines
 *
 ******************************************************************************
 */
						

#include <vector>
#include <fstream>
#include "compile.h"
#include "cnode.h"
#include "cmodule.h"
#include "cinstance.h"
#include "array.h"
#include "member.h"


/************************************************
	CreateModuleList
	- mapping of name to module definition
 ************************************************/

static CNode* moduleComments;

static void CreateModuleList( 
    CNode*                n, 
    set<CModule*>&        allModules,
    set<CModule*>&        leafModules,
    map<CModule*,CNode*>& module2Comments
)
{
    if( n == NULL ) {
	return;
    }
    switch( n->GetOp() ) {
    case eLIST:
	CreateModuleList( n->Arg<CNode*>(0), allModules, leafModules, module2Comments );
	CreateModuleList( n->Arg<CNode*>(1), allModules, leafModules, module2Comments );
	break;
    case eMODULE_DEF:
	{
	    CModule* m = n->Arg<CModule*>(0);
	    module2Comments[m] = moduleComments;
	    moduleComments = NULL;
	    allModules.insert(m);
	    vector<CInstance*>& inst = *m->GetInstanceList();
	    vector<CInstance*>::iterator p;
	    for( p = inst.begin(); p != inst.end(); ++p ) {
		CModule* def = (*p)->GetDefinition();
		leafModules.insert(def);
	    }
	}
	break;
    case eVRQ:
    case ePRAGMA:
    case eATTRIBUTE:
	break;
    case eCOMMENT:
	moduleComments = cLINK( moduleComments, n );
	break;
    default:
	MASSERT( FALSE );
    }       
}

/************************************************
	AnalyseModules
	- analyse module hierarchy
 ************************************************/

void AnalyseModules( 
    CNode* codeList, 
    set<CModule*>& allModules,
    set<CModule*>& topLevelModules,
    set<CModule*>& leafModules,
    set<CModule*>& undefinedModules,
    map<CModule*,CNode*>& module2Comments
)
{
    moduleComments = NULL;
    CreateModuleList( codeList, allModules, leafModules, module2Comments );
    
    set_difference( allModules.begin(), allModules.end(),
		    leafModules.begin(), leafModules.end(),
		    inserter(topLevelModules, topLevelModules.end()) );

    set_difference( leafModules.begin(), leafModules.end(),
			allModules.begin(), allModules.end(),
			inserter(undefinedModules, undefinedModules.end()) );
	
}

/**
 * Get datatype info for array, structure and variable references
 * \param ref reference to a decl/array/member node
 * \param dataType resolved base data type
 * \param unpackedDimensions list of unresolved unpacked dimensions
 * \param packedDimensions list of unresolved packed dimensions 
 * \param residualPackedIndices list of unresolved packed indices
 * \return -1 if reference is valid but unresolvable
 * 	    0 if reference is invalid
 * 	    1 if reference is valid and was resolved
 */
int ResolveReference( 
	CNode*          ref, 
	CDataType*&     dataType,
	vector<CNode*>& unpackedDimensions,
	vector<CNode*>& packedDimensions,
	vector<CNode*>& residualPackedIndices )
{

    MASSERT( ref );
    switch( ref->GetOp() ) {
    case eARRAY:
	return Array::ResolveReference(
		ref->Arg<CNode*>(0),
		ref->Arg<CNode*>(1),
		dataType,
		unpackedDimensions,
		packedDimensions,
		residualPackedIndices );
    case eMEMBER: 
	return Member::ResolveReference(
		ref->Arg<CNode*>(0),
		ref->Arg<CSymbol*>(1),
		dataType,
		unpackedDimensions,
		packedDimensions );
    return 1;
    case eVAR_REF:
    case eNET_REF:
    case ePORT_REF:
    case ePARAM_REF:
    case eGENVAR_REF:
    case eENUM_REF:
    case eEXTERNAL_REF:
	break;
    default:
	MASSERT( FALSE );
    }   

    unpackedDimensions.clear();
    packedDimensions.clear();
    residualPackedIndices.clear();
    CDecl* decl = ref->Arg<CDecl*>(0);
    dataType = decl->GetDataType();
    for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
        packedDimensions.push_back(dataType->GetPackedRange(i));
    }
    for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
        unpackedDimensions.push_back(decl->GetUnpackedRange(i));
    }
    while( dataType->GetVarDataType() == ePROXY ) {
	CDecl* decl = dynamic_cast<CProxyType*>(dataType)->GetDecl();
        for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
            unpackedDimensions.push_back(decl->GetUnpackedRange(i));
        }
	dataType = dynamic_cast<CProxyType*>(dataType)->GetDataType();
        for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
            packedDimensions.push_back(dataType->GetPackedRange(i));
        }
    }
    return 1; 
}

