/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cparam.cpp
 *		- class definition of parameter 
 *		  definition nodes
 *
 ******************************************************************************
 */

#include "cparam.h"
#include "cnode.h"




/****************************************************
	Constructor
*****************************************************/
CParam::CParam( CSymbol* name, Coord_t* aLoc, 
		Decl_t type, CDataType* dataType ) :
	CDecl( name, aLoc, type, dataType, Or(eFLAG_PRAGMA,eFLAG_DATATYPE) ) {
	expression = NULL;
	specify    = 0;
}

/****************************************************
	Clone
	- Create a deep copy of this declaration 
*****************************************************/
CDecl* CParam::Clone( CObstack* heap )
{
    CParam* clone = new(heap) CParam( GetSymbol(), GetCoord(),
			GetType(), NULL );
    clone->Copy( heap, *this );
    return clone;
}

/****************************************************
	Copy
	- do a deep copy from given object.
*****************************************************/
void    CParam::Copy( CObstack* heap, CParam& param )
{
    CDecl::Copy( heap, param );
    expression = param.expression ? param.expression->Clone(heap) : NULL;
    specify    = param.specify;
    inlined    = param.inlined;
}

/****************************************************
	SetExpression
*****************************************************/
void 	CParam::SetExpression( CNode* aExp ) {
	expression = aExp;
}

/****************************************************
	GetExpression
*****************************************************/
CNode*	CParam::GetExpression( void ) {
	return expression;
}	

/***********************************************************
	GetNodeType
	- return parameter nodetype
***********************************************************/
NodeType_t CParam::GetNodeType( void )
{
    return expression->GetNodeType();
}

/**
 * Determine if packed width of declaration is constant, 
 * ie dependent upon only constants and parameters.
 * \return non-zero of packed width of declaration is constant.
 */
int  CParam::IsWidthConstant( void ) 
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) { 
	MASSERT( expression );
	return expression->IsWidthConstant();
    }
    return CDecl::IsWidthConstant();
}

/**
 * Determine if packed width of declaration is volatile, ie depend
 * upon parameters or variables.
 * \return non-zero of packed width of declaration is volatile.
 */
int  CParam::IsWidthVolatile( void )
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) {
	MASSERT( expression );
	return expression->IsWidthVolatile();
    }
    return CDecl::IsWidthVolatile();
}

/**
 * Determine if packed width of declaration can be evaluated.
 * \return non-zero of packed width of declaration can be evaluated.
 */
int  CParam::IsWidthEvaluateable( void ) 
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) {
	MASSERT( expression );
	return expression->IsEvaluateable();
    }
    return CDecl::IsWidthEvaluateable();
}

/**
 * Evaluate packed width of declaration.
 * \return packed width of declaration.
 */
INT32 CParam::GetWidth( void )
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) {
	MASSERT( expression );
	return expression->GetWidth();
    }
    return CDecl::GetWidth();
}

/**
 * Get expression representing width of declaration.
 * \return expression representing width of declaration.
 */
CNode* CParam::GetWidthExp( void )
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) {
	MASSERT( expression );
	return expression->GetWidthExp();
    }
    return CDecl::GetWidthExp();
}

/**
 * Get expression for declaration's msb
 * \return msb expression, NULL if scalar.
 */
CNode* CParam::GetMsb()
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) {
	MASSERT( expression );
	if( !expression->IsVolatile() && expression->IsEvaluateable() ) {
	    return cINT32(expression->GetWidth());
	}
	return cSUB(expression->GetWidthExp(),cINT32(1));
    }
    return CDecl::GetMsb();
}

/**
 * Get expression for declaration's lsb
 * \return lsb expression, NULL if scalar.
 */
CNode* CParam::GetLsb()
{
    MASSERT( GetDataType() );
    if( GetDataType()->GetVarDataType() == eNO_VAR_TYPE &&
	GetDataType()->IsVector() ) {
	MASSERT( expression );
	return cINT32(0);
    }
    return CDecl::GetLsb();
}

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CParam::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    if( GetExpression() ) {
	GetExpression()->PreVisit1( func, data );
    }
    CDecl::PreVisit1( func, data );
}
/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CParam::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    if( GetExpression() ) {
	GetExpression()->PostVisit1( func, data );
    }
    CDecl::PostVisit1( func, data );
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CParam::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    if( GetExpression() ) {
	SetExpression( GetExpression()->PostSubVisit1( func, data ) );
    }
    CDecl::PostSubVisit1( func, data );
}

/****************************************************
	Dump
*****************************************************/
void	CParam::Dump( FILE* f )
{
	fprintf( f, "Parameter: %s = " , GetName() );
	if( expression->IsVolatile() || !expression->IsEvaluateable() ) {
		expression->Dump( f );
	} else {
		switch( expression->GetNodeType() ) {
		case eS:
			{
				CVector v(expression->GetWidth());
				expression->EvalVector(v);
				if( v.IsNegative() ) {
				    CVector vm(expression->GetWidth());
				    Neg( &vm, &v );
				    fprintf( f, "-'%ldsh%s", vm.GetWidth(), vm.GetHex() );
				} else {
				    fprintf( f, "'%ldsh%s", v.GetWidth(), v.GetHex() );
				}
				fprintf( f, "%ld", v.GetINT32() );
			}
			break;
		case eR:
			fprintf( f, "%g", expression->EvalReal() );
			break;
		case eB:
			{ 
				CVector v(expression->GetWidth());
				expression->EvalVector(v);
				fprintf( f, "'%ldh%s", v.GetWidth(), v.GetHex() );
			}
			break;
		default:
			MASSERT(FALSE);
		}
	}
	fprintf( f, ", defined in " );
	CDecl::Dump( f );
	fprintf( f, "\n" );
}	

