/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cproxytype.h
 *		- abstract class for referring to other types
 *
 *
 ******************************************************************************
 */

#ifndef CPROXYTYPE_H
#define CPROXYTYPE_H

#include <list>
#include "cdatatype.h"
#include "cdecl.h"

class CSymbol;
class CNode;


/**
 * Class for representing a named data type
 */
class CProxyType : public CDataType
{
    CSymbol*   symbol;
    CDecl*     decl;
public:
        /**
         * Create type proxy instance
         */
        CProxyType( CSymbol* symbol, CDecl* decl, list<CNode*>& packedList ) 
	  : CDataType(ePROXY,packedList), 
	    symbol(symbol), 
	    decl(decl) 
	{}
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap );
	/**
 	 * Get declaration's symbol
 	 * \return symbol representing type name
 	 */
	virtual CSymbol*  GetSymbol() const { return symbol; }
	/**
 	 * Get proxy's declaration
 	 * \return proxy's decl node
 	 */
	virtual CDecl*  GetDecl() const { return decl; }
	/**
 	 * Get proxy's data type
 	 * \return proxy's data type
 	 */
	virtual CDataType*  GetDataType() const { return decl->GetDataType(); }
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const { return decl->GetDataType()->GetTwoState(); }
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const { return decl->GetDataType()->GetSigned(); }
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const 
	{ return decl->GetNodeType(); }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero if vector or packed width of declaration 
 	 * is constant.
 	 */
	virtual int  IsBaseWidthConstant( void ) const
	{ return decl->IsWidthConstant(); }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is volatile, 
 	 * ie depend upon parameters or variables.
 	 * \return non-zero of vector or packed width of declaration 
 	 * is volatile.
 	 */
	virtual int  IsBaseWidthVolatile( void ) const
	{ return decl->IsWidthVolatile(); }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration can be evaluated.
 	 * \return non-zero of packed or vector width of declaration can 
 	 * be evaluated.
 	 */
	virtual int  IsBaseWidthEvaluateable( void ) const
	{ return decl->IsWidthEvaluateable(); }
        /**
         * Evaluate base width (sans packed dimensions) of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetBaseWidth( void ) const
	{ return decl->GetWidth(); }
        /**
         * Get expression for datatype's base width (sans packed dimensions)
         * \return width expression
         */
        virtual CNode* GetBaseWidthExp() const
	{
	  return decl->GetWidthExp();
	}
        /**
         * Determine if complete data structure is a vector.
         * Vectors are multi-bit intergal types.
         * \return non-zero if complete data structure is a vector.
         */
        virtual int IsVector() const 
  	{
	  return decl->GetDataType()->IsVector() &&
		 GetNumberOfPackedDimensions() == 0 &&
		 decl->GetNumberOfUnpackedDimensions() == 0;
	}
        /**
         * Determine if complete data structure is packed.
         * Packed data structures can be treated as a single vector.
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const 
	{ 
	    return decl->GetDataType()->IsPacked() &&
		   decl->GetNumberOfUnpackedDimensions() == 0;
	}
        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const { decl->GetDataType()->Dump( f ); }
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CProxyType& o );
private:
	/*
 	 * disable copy constructors
 	 */
	CProxyType( const CProxyType& o );

};

#endif // CPROXYTYPE_H

