/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdatatype.cc
 *		- abstract class for data types 
 *
 *
 ******************************************************************************
 */

#define CSIMPLETYPE_CC

#include <list>
#include "csimpletype.h"
#include "cnode.h"

/**
 * Create data type instance
 */
CSimpleType::CSimpleType( VarDataType_t vt )
    : CDataType(vt),
      signedAttr(FALSE),
      unsignedAttr(FALSE)
{
}

/**
 * Create data type instance
 */
CSimpleType::CSimpleType( VarDataType_t vt, list<CNode*>& rangeList )
    : CDataType(vt,rangeList),
      signedAttr(FALSE),
      unsignedAttr(FALSE)
{
}

/**
 * Create a new copy with a deep copy. 
 */
CDataType* CSimpleType::Clone( CObstack* heap )
{
    CSimpleType* clone = new(heap) CSimpleType(GetVarDataType());
    clone->Copy( heap, *this ); 
    return clone;
}

/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CSimpleType::Copy( CObstack* heap, const CSimpleType& o )
{
    CDataType::Copy( heap, o );
    signedAttr   = o.signedAttr;
    unsignedAttr = o.unsignedAttr;
}
/**
 * Evaluate base width (sans packed dimensions) of declaration.
 * \return packed or vector width of declaration.
 */
INT32 CSimpleType::GetBaseWidth( void ) const
{
	switch( GetVarDataType() ) {
        case eNO_VAR_TYPE:
        case eREG:
        case eLOGIC:          
        case eBIT:            
        case eLONGINT:      
        case eTIME:
        case eBYTE:           
        case eSHORTINT:       
        case eINT:
        case eINTEGER:        
	    break;
        case eREAL:
        case eSHORTREAL:
        case eREALTIME:
	default:
	    MASSERT( FALSE );
	}
        return 1;
}
/**
 * Get expression for datatype's base width (sans packed dimensions)
 * \return width expression
 */
CNode* CSimpleType::GetBaseWidthExp() const 
{ 
	return cINT32(GetBaseWidth()); 
}

/**
 * Determine if complete data structure is a vector.
 * Vectors are a multi-bit intergal type.
 * \return non-zero if complete data structure is a vector.
 */
int CSimpleType::IsVector() const
{
    switch( GetVarDataType() ) {
    case eLONGINT:      
    case eTIME:
    case eBYTE:           
    case eSHORTINT:       
    case eINT:
    case eINTEGER:        
	return TRUE;
    case eNO_VAR_TYPE:
    case eREG:
    case eLOGIC:          
    case eBIT:            
	if( !CDataType::GetNumberOfPackedDimensions() ) {
	    return TRUE;
	}
	break;
    case eDEVENT:
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
	break;
    default:
        MASSERT( FALSE );
    }
    return FALSE;
}

/**
 * Determine if complete data structure is packed.
 * Packed data structures can be treated as a single vector.
 * \return non-zero if complete data structure is packed.
 */
int CSimpleType::IsPacked() const 
{
    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eLOGIC:          
    case eBIT:            
    case eLONGINT:      
    case eTIME:
    case eBYTE:           
    case eSHORTINT:       
    case eINT:
    case eINTEGER:        
	return TRUE;;
    case eDEVENT:
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
	break;
    default:
        MASSERT( FALSE );
    }

    return FALSE;
}

/**
 * Get expression for declaration's msb. Valid only for
 * Packed and Vector declarations.
 * \return msb expression, NULL if scalar.
 */
CNode* CSimpleType::GetPackedMsb() const 
{ 
    MASSERT( IsPacked() || IsVector() );
    if( IsVector() ) {
        return cINT32(GetVectorWidth()-1);
    }
    return CDataType::GetPackedMsb();
}
/**
 * Get expression for declaration's lsb. Valid only for
 * Packed and Vector declarations.
 * \return lsb expression, NULL if scalar.
 */
CNode* CSimpleType::GetPackedLsb() const
{ 
    MASSERT( IsPacked() || IsVector() );
    if( IsVector() ) {
        return cINT32(0);
    }
    return CDataType::GetPackedLsb();
}

/***********************************************************
	WidthDirection
	- evaluate current decl width direction.
	  Expression must be constant.  
	  1 - big endian, -1 - little endian, 0 - cannot be determined
 **********************************************************/

int CSimpleType::PackedWidthDirection( void ) const 
{
    MASSERT( IsPacked() || IsVector() );
    if( IsVector() ) {
	return 1;
    }
    return CDataType::PackedWidthDirection();
}

/**
 * Get expression for datatype's overall packed or vector 
 * range (msb/lsb)
 * \return range expression, NULL if scalar.
 */
CNode* CSimpleType::GetPackedRange() const { 
    MASSERT( IsPacked() || IsVector() );
    if( IsVector() ) {
        return cRANGE(GetPackedMsb(), GetPackedLsb());
    }
    return CDataType::GetPackedRange();
}

/**
 *  Evaluate current packed or vector decl width. 
 *  Expression must be constant.  
 * \return evaluated packed or vector width expression.
 */
INT32 CSimpleType::GetPackedWidth( void ) const {
	MASSERT( IsVector() || IsPacked() ) ;
	if( IsVector() ) {
	    return GetVectorWidth();
	}
	return CDataType::GetPackedWidth();
}


/**
 * Get expression for datatype's overall packed or vector width
 * \return width expression
 */
CNode* CSimpleType::GetPackedWidthExp() const
{
    MASSERT( IsVector() || IsPacked() ) ;
    if( !IsPackedWidthVolatile() && IsPackedWidthEvaluateable() ) {
        return cINT32(GetPackedWidth());
    }
    return CDataType::GetPackedWidthExp();
}
/**
 * Get width of vector. Only valid for vector types
 * \return width of vector in bits.
 */
int CSimpleType::GetVectorWidth() const
{
    switch( GetVarDataType() ) {
    case eLONGINT:      
    case eTIME:
	return 64;
    case eBYTE:           
	return 8;
    case eSHORTINT:       
	return 16;
    case eINT:
    case eINTEGER:        
	return 32;
    case eNO_VAR_TYPE:
    case eREG:
    case eLOGIC:          
    case eBIT:            
	return 1;
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    default:
        MASSERT( FALSE );
    }
}

/**
 * Get number of packed dimensions of declaration.
 * \return number of dimensions, 0 if non-array.
 */
INT32 CSimpleType::GetNumberOfPackedDimensions( void ) const 
{ 
    if( IsVector() ) {
	return 1;
    }
    return CDataType::GetNumberOfPackedDimensions();
}

/**
 * Get expression tree for upper limit of given packed array dimension.
 * \param dim dimension.
 * \return expression tree for packed upper limit.
 */
CNode* CSimpleType::GetPackedMsi( INT32 dim ) const {
    if( IsVector() ) {
	if( dim == 0 ) {
	    return cINT32(GetVectorWidth()-1);
	} else {
	    return NULL;
	}
    }
    return CDataType::GetPackedMsi(dim);
}

/**
 * Get expression tree for lower limit of packed array dimension.
 * \param dim dimension.
 * \return expression tree for packed lower limit.
 */
CNode* CSimpleType::GetPackedLsi( INT32 dim ) const {
    if( IsVector() ) {
	if( dim == 0 ) {
	    return cINT32(0);
	} else {
	    return NULL;
	}
    }
    return CDataType::GetPackedLsi(dim);
}
/**
 * Get expression for range of packed array for dimension.
 * \param dim dimension to get.
 * \return range expression for limit.
 */
CNode* CSimpleType::GetPackedRange( INT32 dim ) const
{ 
    if( IsVector() ) {
        if( dim == 0 ) {
	    return cRANGE( GetPackedMsi(0), GetPackedLsi(0) );
	} else {
	    return NULL;
	}
    }
    return CDataType::GetPackedRange(dim);
}

/**
 * Get declaration's 2 state property.
 * \return value of 2 state property.
 */
int  CSimpleType::GetTwoState() const
{
    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eTIME:
    case eLOGIC:          
    case eINTEGER:        
	return FALSE;
    case eDEVENT:
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    case eINT:
    case eBIT:            
    case eBYTE:           
    case eSHORTINT:       
    case eLONGINT:      
	return TRUE;
    default:
	MASSERT( FALSE );
	return FALSE;
    }
}
/**
 * Get declartion's signed property.
 * \return value of signed property.
 */
int  CSimpleType::GetSigned() const
{
    if( signedAttr ) {
	return TRUE;
    }
    if( unsignedAttr ) {
	return FALSE;
    }

    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eTIME:
    case eLOGIC:          
    case eDEVENT:
    case eBIT:            
	return FALSE;
    case eINTEGER:        
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    case eINT:
    case eBYTE:           
    case eSHORTINT:       
    case eLONGINT:      
	return TRUE;
    default:
	MASSERT( FALSE );
	return FALSE;
    }
}

/**
 * Get data type
 * \return node type
 */
NodeType_t CSimpleType::GetNodeType( void ) const
{
    switch( GetVarDataType() ) {
    case eNO_VAR_TYPE:
    case eREG:
    case eTIME:
    case eINT:
    case eLOGIC:          
    case eINTEGER:        
    case eBIT:            
    case eBYTE:           
    case eSHORTINT:       
    case eLONGINT:      
	return GetSigned() ? eS : eB;
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
	return eR;
    case eDEVENT:
	return eE;
    default:
	MASSERT( FALSE );
	return eU;
    }
}

/**
 * Dump data type info to file descriptor.
 * \param f file descriptor.
 */
void CSimpleType::Dump( FILE* f ) const
{ 
    int dumpRange = FALSE;

    switch( GetVarDataType() ) {
    case eREAL:
    case eSHORTREAL:
    case eREALTIME:
    case eINT:
    case eINTEGER:
    case eBYTE:
    case eSHORTINT:
    case eLONGINT:
    case eTIME:
        break;
    case eBIT:
    case eNO_VAR_TYPE:
    case eREG:
    case eLOGIC:
        dumpRange = TRUE;
        break;
    case eDEVENT:
	dumpRange = FALSE;
    default:
        MASSERT( FALSE );
    }
    fprintf( f, "%s ", varDataTypeName[GetVarDataType()] );
    if( GetSignedAttr() ) {
        fprintf( f, "signed " );
    }
    if( GetUnsignedAttr() ) {
        fprintf( f, "unsigned " );
    }

    if( dumpRange ) {
        if( GetPackedRange() ) {
	    GetPackedRange()->Dump( f );
            fprintf( f, " " );
        }
   }

}


