/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   csimpletype.h
 *		- abstract class for simple built in data types 
 *
 *
 ******************************************************************************
 */

#ifndef CSIMPLETYPE_H
#define CSIMPLETYPE_H

#include "cobject.h"
#include "cdatatype.h"

/**
 * class for describing simple builtin data types
 */
class CSimpleType : public CDataType
{
public:
private:
	int        signedAttr;	        ///< TRUE if signed attribute is present
	int        unsignedAttr;        ///< TRUE if unsigned attribute is present
public:
        /**
         * Create data type instance
         * \param vt variable simple data type
         */
        CSimpleType( VarDataType_t vt );
        /**
         * Create data type instance
         * \param vt variable simple data type
         * \param rangeList vector of packed ranges for declaration.
         */
        CSimpleType( VarDataType_t vt, list<CNode*>& rangeList );
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap );
	/**
 	 * Get declaration's signed attribute.
 	 * \return value of signed attribute.
 	 */
	virtual int GetSignedAttr() const { return signedAttr; }
	/**
 	 * Get declaration's unsigned attribute.
 	 * \return value of unsigned attribute.
 	 */
	virtual int GetUnsignedAttr() const { return unsignedAttr; }
	/**
 	 * Set declaration's signed attribute.
 	 * \param v value of signed attribute.
 	 */
	virtual void SetSignedAttr( int v ) { signedAttr = v; }
	/**
 	 * Set declaration's unsigned attribute.
 	 * \param v value of unsigned attribute.
 	 */
	virtual void SetUnsignedAttr( int v ) { unsignedAttr = v; }
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const;
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const;
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const;
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero if vector or packed width of declaration 
 	 * is constant.
 	 */
	virtual int  IsBaseWidthConstant( void ) const { return TRUE; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is volatile, 
 	 * ie depend upon parameters or variables.
 	 * \return non-zero of vector or packed width of declaration 
 	 * is volatile.
 	 */
	virtual int  IsBaseWidthVolatile( void ) const { return FALSE; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration can be evaluated.
 	 * \return non-zero of packed or vector width of declaration can 
 	 * be evaluated.
 	 */
	virtual int  IsBaseWidthEvaluateable( void ) const {return TRUE; }
        /**
         * Evaluate base width (sans packed dimensions) of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetBaseWidth( void ) const;
        /**
         * Get expression for datatype's base width (sans packed dimensions)
         * \return width expression
         */
        virtual CNode* GetBaseWidthExp() const;
        /**
         * Determine if complete data structure is a vector.
         * Vectors are a multi-bit intergal type.
         * \return non-zero if complete data structure is a vector.
         */
        virtual int IsVector() const;
        /**
         * Determine if complete data structure is packed.
         * Packed data structures can be treated as a single vector.
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const; 
        /**
         * Get expression for declaration's msb. Valid only for
         * Packed and Vector declarations.
         * \return msb expression, NULL if scalar.
         */
        virtual CNode* GetPackedMsb() const;
        /**
         * Get expression for declaration's lsb. Valid only for
         * Packed and Vector declarations.
         * \return lsb expression, NULL if scalar.
         */
        virtual CNode* GetPackedLsb() const;
        /**
         * Get expression for datatype's overall packed or vector 
         * range (msb/lsb)
         * \return range expression, NULL if scalar.
         */
        virtual CNode* GetPackedRange() const;
        /**
         * Evaluate packed or vector width of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetPackedWidth( void ) const;
        /**
         * Get expression for datatype's overall packed or vetor width
         * \return width expression
         */
        virtual CNode* GetPackedWidthExp() const;
	/**
 	 * Evaluate current decl width direction. Expression must be constant.  
	 * \return 1 - big endian, -1 - little endian, 0 - cannot be determined
	 */
	virtual int   PackedWidthDirection( void ) const;
	/**
 	 * Get number of packed dimensions of declaration.
 	 * \return number of dimensions, 0 if non-array.
 	 */
	virtual INT32 GetNumberOfPackedDimensions( void ) const;
	/**
 	 * Get expression tree for upper limit of given packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for packed upper limit.
	 */
	virtual CNode* GetPackedMsi( INT32 dim ) const;
	/**
 	 * Get expression tree for lower limit of packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for packed lower limit.
	 */
	virtual CNode* GetPackedLsi( INT32 dim ) const;
	/**
 	 * Get expression for range of packed array for dimension.
 	 * \param dim dimension to get.
 	 * \return range expression for limit.
 	 */
	virtual CNode*  GetPackedRange( INT32 dim ) const;
        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const;
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CSimpleType& o );
	/**
 	 * Get width of vector. Only valid for vector types
 	 * \return width of vector in bits.
 	 */
 	int GetVectorWidth() const;
private:
 	/*
 	 * disable copy constructor
 	 */
	CSimpleType( const CSimpleType& o ); 

};

#endif // CSIMPLETYPE_H

