/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cstringtype.h
 *		- abstract class for string types 
 *
 *
 ******************************************************************************
 */

#ifndef CSTRINGTYPE_H
#define CSTRINGTYPE_H

#include "cdatatype.h"


/**
 * Base class for describing data types
 */
class CStringType : public CDataType
{
public:
        /**
         * Create class type instance
         */
        CStringType( VarDataType_t t) : CDataType(t) {}
	/**
 	 * Create a new copy with a deep copy. 
 	 * \param heap heap to use for allocation.
 	 */
	virtual CDataType* Clone( CObstack* heap ) { MASSERT(FALSE);}
	/**
 	 * Get declaration's 2 state property.
 	 * \return value of 2 state property.
 	 */
	virtual int  GetTwoState() const { return FALSE; }
	/**
 	 * Get declartion's signed property.
 	 * \return value of signed property.
 	 */
	virtual int  GetSigned() const { return FALSE; }
        /**
         * Get data type
         * \return node type
         */
        virtual NodeType_t GetNodeType( void ) const { return eB; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero if vector or packed width of declaration 
 	 * is constant.
 	 */
	virtual int  IsBaseWidthConstant( void ) const { return FALSE; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration is volatile, 
 	 * ie depend upon parameters or variables.
 	 * \return non-zero of vector or packed width of declaration 
 	 * is volatile.
 	 */
	virtual int  IsBaseWidthVolatile( void ) const { return TRUE; }
	/**
 	 * Determine if base width (sans packed dimensions)
 	 * of declaration can be evaluated.
 	 * \return non-zero of packed or vector width of declaration can 
 	 * be evaluated.
 	 */
	virtual int  IsBaseWidthEvaluateable( void ) const { return FALSE; }
        /**
         * Evaluate base width (sans packed dimensions) of declaration.
         * \return packed or vector width of declaration.
         */
        virtual INT32 GetBaseWidth( void ) const { return 0; }
        /**
         * Get expression for datatype's base width (sans packed dimensions)
         * \return width expression
         */
        virtual CNode* GetBaseWidthExp() const { return NULL; }
        /**
         * Determine if complete data structure is a vector.
         * Vectors are multi-bit intergal types.
         * \return non-zero if complete data structure is a vector.
         */
        virtual int IsVector() const { return FALSE; }
        /**
         * Determine if complete data structure is packed.
         * Packed data structures can be treated as a single vector.
         * \return non-zero if complete data structure is packed.
         */
        virtual int IsPacked() const { return FALSE; }
	/**
 	 * Get number of packed dimensions of declaration.
 	 * \return number of dimensions, 0 if non-array.
 	 */
	virtual INT32 GetNumberOfPackedDimensions( void ) const { return 0;}
	/**
 	 * Get expression tree for upper limit of given packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for packed upper limit.
	 */
	virtual CNode* GetPackedMsi( INT32 dim ) const { return NULL;}
	/**
 	 * Get expression tree for lower limit of packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for packed lower limit.
	 */
	virtual CNode* GetPackedLsi( INT32 dim ) const { return NULL;}
	/**
 	 * Get expression for range of packed array for dimension.
 	 * \param dim dimension to get.
 	 * \return range expression for limit.
 	 */
	virtual CNode*  GetPackedRange( INT32 dim ) const { return NULL;}
        /**
         * Dump data type info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) const {}
protected:
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CStringType& o );
private:
	/*
 	 * disable copy constructors
 	 */
	CStringType( const CStringType& o );

};

#endif // CSTRINGTYPE_H

