/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cstructtype.cc
 *		- abstract class for structure types 
 *
 *
 ******************************************************************************
 */

#include <list>
#include "cstructtype.h"
#include "cvar.h"
#include "cnode.h"

/**
 * Create structure type instance
 */
CStructType::CStructType( VarDataType_t t, CNode* members, list<CNode*>& rangeList ) 
  : CDataType(t, rangeList), 
    members(members), 
    packed(0), 
    signedAttr(0), 
    unsignedAttr(0) 
{
}
/**
 * Create a new copy with a deep copy. 
 * \param heap heap to use for allocation.
 */
CDataType* CStructType::Clone( CObstack* heap )
{
    list<CNode*> emptyList;
    CStructType* clone = new(heap) CStructType(GetVarDataType(), NULL, emptyList);
    clone->Copy( heap, *this ); 
    return clone;
}

/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CStructType::Copy( CObstack* heap, const CStructType& o )
{
    CDataType::Copy( heap, o );
    members = o.members ? o.members->Clone(heap) : NULL;
}

/**
 * Dump data type info to file descriptor.
 * \param f file descriptor.
 */
void CStructType::Dump( FILE* f ) const
{
    fprintf( f, "struct " );
}

/**
 * Get declaration's 2 state property.
 * \return value of 2 state property.
 */
static int _GetTwoState( CNode* n, void* data )
{
    int* twoState = static_cast<int*>(data);
    switch( n->GetOp() ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*twoState &= n->Arg<CDecl*>(0)->GetDataType()->GetTwoState();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
int CStructType::GetTwoState() const
{
    int twoState = TRUE;
    if( members ) {
	members->PreVisit1( _GetTwoState, &twoState );
    }
    return twoState;
}
/**
 * Validate data structure 
 * \param errorMessage description of validation failure
 * \return true if structure is valid
 */
static int _GetPacked( CNode* n, void* data )
{
    bool* packed = static_cast<bool*>(data);
    switch( n->GetOp() ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*packed &= (n->Arg<CDecl*>(0)->GetNumberOfUnpackedDimensions() == 0);
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
bool CStructType::Validate( string& errorMessage )
{
    if( packed ) {
	bool packedAcc = TRUE;
        if( members ) {
	    members->PreVisit1( _GetPacked, &packedAcc );
        }
	if( !packedAcc ) {
	    errorMessage = "packed structures must have packed members";
	    return false;
        }
    }
    return true;
}
/**
 * Determine if packed or vector width of declaration is constant, 
 * ie dependent upon only constants and parameters.
 * \return non-zero if vector or packed width of declaration 
 * is constant.
 */
static int _IsWidthConstant( CNode* n, void* data )
{
    int* isWidthConstant = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*isWidthConstant &= n->Arg<CDecl*>(0)->IsWidthConstant();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
int CStructType::IsBaseWidthConstant() const
{
    int isWidthConstant = TRUE;
    if( members ) {
        members->PreVisit1( _IsWidthConstant, &isWidthConstant );
    }
    return isWidthConstant;
}

/**
 * Determine if packed or vector width of declaration is volatile, 
 * ie depend upon parameters or variables.
 * \return non-zero of vector or packed width of declaration 
 * is volatile.
 */
static int _IsWidthVolatile( CNode* n, void* data )
{
    int* isWidthVolatile = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*isWidthVolatile |= n->Arg<CDecl*>(0)->IsWidthVolatile();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
int CStructType::IsBaseWidthVolatile() const
{
    int isWidthVolatile = FALSE;
    if( members ) {
        members->PreVisit1( _IsWidthVolatile, &isWidthVolatile );
    }
    return isWidthVolatile;
}
/**
 * Determine if packed or vector width of declaration can be evaluated.
 * \return non-zero of packed or vector width of declaration can 
 * be evaluated.
 */
static int _IsWidthEvaluateable( CNode* n, void* data )
{
    int* isWidthEvaluateable = static_cast<int*>(data);
    NodeOp_t t = n->GetOp();
    switch( t ) {
    case eLIST:
        break;
    case ePARAM_DECL:
    case eVAR_DECL:
        *isWidthEvaluateable &= n->Arg<CDecl*>(0)->IsWidthEvaluateable();
        return 0;
    default:
        MASSERT( FALSE );
    }
    return 1;
}
int CStructType::IsBaseWidthEvaluateable() const
{
    int isWidthEvaluateable = TRUE;
    if( members ) {
       members->PreVisit1( _IsWidthEvaluateable, &isWidthEvaluateable );
    }
    return isWidthEvaluateable;
}

/**
 * Evaluate packed or vector width of declaration.
 * \return packed or vector width of declaration.
 */
static int _GetWidthStruct( CNode* n, void* data )
{
    int* width = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*width += n->Arg<CDecl*>(0)->GetWidth();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
static int _GetWidthUnion( CNode* n, void* data )
{
    int* width = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*width = (*width > n->Arg<CDecl*>(0)->GetWidth()) ? *width : 
			n->Arg<CDecl*>(0)->GetWidth();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
INT32 CStructType::GetBaseWidth( void ) const
{
    int width = 0;
    if( members ) {
	if( GetVarDataType() == eSTRUCT ) {
            members->PreVisit1( _GetWidthStruct, &width );
	} else {
            members->PreVisit1( _GetWidthUnion, &width );
	}
    }
    return width;
}

/**
 * Get expression for datatype's overall packed or vector width
 * \return width expression
 */
static int _GetWidthExpStruct( CNode* n, void* data )
{
    CNode** width = static_cast<CNode**>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL: {
	CNode* exp = n->Arg<CDecl*>(0)->GetWidthExp();
	if( *width ) {
	    *width = cADD(*width,exp);
	} else {
	    *width = exp;
	}
	} return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
static int _GetWidthExpUnion( CNode* n, void* data )
{
    CNode** width = static_cast<CNode**>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL: {
	CNode* exp = n->Arg<CDecl*>(0)->GetWidthExp();
	if( *width ) {
	    *width = cMAX(*width,exp);
	} else {
	    *width = exp;
	}
	} return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
CNode* CStructType::GetBaseWidthExp() const
{
    CNode* width = NULL;
    if( members ) {
	if( GetVarDataType() == eSTRUCT ) {
            members->PreVisit1( _GetWidthExpStruct, &width );
	} else {
            members->PreVisit1( _GetWidthExpUnion, &width );
	}
    }
    return width;
}
/**
 * Find declaration for member
 * \param member  member to search for.
 * \return matching declaration, NULL if not found
 */

struct FindMemberInfo {
    CDecl*   decl;
    CSymbol* member;
};
static int _FindMember( CNode* n, void* data ) 
{
    FindMemberInfo* info = static_cast<FindMemberInfo*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL: {
	CDecl* decl = n->Arg<CDecl*>(0);
	if( decl->GetSymbol() == info->member ) {
	    info->decl = decl;
	}
	} return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
CDecl* CStructType::FindMember( CSymbol* member )
{
    FindMemberInfo info;
    info.decl = NULL;
    info.member = member; 
    if( members ) {
         members->PreVisit1( _FindMember, &info );
    }
    return info.decl;
}
/**
 * Walk tree invoking callback on each node before children have 
 * been visited. callback can terminate transversal early.
 * callback returns 0 to terminate transversal early.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CStructType::PreVisit1( int (*callback)(CNode*,void*), void* data )
{
    CDataType::PreVisit1( callback, data );
    if( members ) {
	members->PreVisit1( callback, data );
    }
}
/**
 * Walk tree invoking callback on each node after children have 
 * been visited.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CStructType::PostVisit1( void (*callback)(CNode*,void*), void* data )
{
    CDataType::PostVisit1( callback, data );
    if( members ) {
	members->PostVisit1( callback, data );
    }
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited. callback may modify tree and return a modified
 * subtree.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 * \return modified child subtree.
 */
void CStructType::PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data )
{
    CDataType::PostSubVisit1( callback, data );
    if( members ) {
	members->PostSubVisit1( callback, data );
    }
}
