/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   ctypedef.h
 *		- class definition of type definitions
 *
 ******************************************************************************
 */

#ifndef CTYPEDEF_H
#define CTYPEDEF_H

#include <stdio.h>
#include "glue.h"
#include "cnode.h"
#include "cdecl.h"


class CNode;


/**
 * Declaration object for type definitions.
 */
class CTypedef: public CDecl
{
public:
	/**
 	 * Create a type declaration.
 	 * \param symbol name of declaration.
 	 * \param dataType data type 
 	 * \param aLoc file coordinates of declaration.
 	 */
	CTypedef( CSymbol* symbol, Coord_t* aLoc, CDataType* dataType );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Get node type of declaration.
 	 * \return node type
 	 */
	virtual NodeType_t GetNodeType( void ) { 
            MASSERT( GetDataType() );
	    return GetDataType()->GetNodeType(); 
	}
	/**
 	 * Get width of declaration.
 	 * \return node width
 	 */
	virtual INT32 GetWidth( void ) { 
            MASSERT( GetDataType() );
	    return GetDataType()->GetPackedWidth(); 
	}
	/**
 	 * Determine if width is evaluateable.
 	 * \return always return TRUE.
 	 */
	virtual int IsWidthEvaluateable() { 
            MASSERT( GetDataType() );
	    return GetDataType()->IsPackedWidthEvaluateable(); 
	}
        /**
         * Determine if packed width of declaration is constant, 
         * ie dependent upon only constants and parameters.
         * \return non-zero of packed width of declaration is constant.
         */
        virtual int  IsWidthConstant( void ) {
            MASSERT( GetDataType() );
	    return GetDataType()->IsPackedWidthConstant(); 
	}
        /**
         * Determine if packed width of declaration is volatile, ie depend
         * upon parameters or variables.
         * \return non-zero of packed width of declaration is volatile.
         */
        virtual int  IsWidthVolatile( void ) {
            MASSERT( GetDataType() );
	    return GetDataType()->IsPackedWidthVolatile(); 
	}
	/**
 	 * Get expression representing type width.
	 * \return width expression.
	 */
	virtual CNode* GetWidthExp() {
            MASSERT( GetDataType() );
	    return GetDataType()->GetPackedWidthExp(); 
	}
	/**
 	 * Dump genvar info to file descriptor.
 	 * \param f file descriptor.
 	 */
        virtual void Dump( FILE* f );
protected:
	/**
 	 * Deep copy of declaration.
 	 * \param heap heap to use for allocation.
 	 * \param d declaration to copy.
 	 */
	void Copy( CObstack* heap, CTypedef& d ) {};
private:
	/*
 	 * disable copy constructor
 	 */
	CTypedef( const CTypedef& );
};

#endif // CTYPEDEF_H
