/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/*******************************************************************
 *
 *	cuint.cpp	
 *		- a class creating a unsigned infinite precision integer 
 *		  
 *
 *******************************************************************
 */

#define CUINT_CPP

#include "cuint.h"
#include "lex.h"

int		CountBits( unsigned int*, int );
unsigned int 	Divide64By32( unsigned int, unsigned int, unsigned int );

/******************************************************************

	operator+( CUInt & );
	operator+( unsigned int );

******************************************************************/

CUInt CUInt::operator+( const CUInt &aValue ) const
{
	CUInt	result;
	unsigned int	carry;
	int		i;


	result.grow( MAX( size, aValue.size ) );
	carry = 0;
	for( i = 0; i < result.size; i++ ) {
		if( i < size ) {
			result.data[i] = data[i] + carry;
			if( result.data[i] < carry ) {
				if( i < aValue.size ) {
					result.data[i] = aValue.data[i];
				}
				continue;
			}
		} else {
			result.data[i] = carry;
		}
			
		if( i < aValue.size ) {
			result.data[i] += aValue.data[i];
			if( result.data[i] < aValue.data[i] ){
				carry = 1;
			} else {
				carry = 0;
			}
		} else {
			carry = 0;
		}
	}	
	if( carry ) {
		result.grow( result.size+1 );
		result.data[result.size-1] = 1;
	}
	

	return( result );

}

CUInt CUInt::operator+( unsigned int aValue ) const
{
	CUInt	result;
	unsigned int	carry;
	int		i;


	result.grow( size );
	carry = 0;
	for( i = 0; i < size; i++ ) {
		result.data[i] = data[i] + carry;
		if( result.data[i] < carry ) {
			if( i == 0 ) {
				result.data[i] = aValue;
			} else {
				result.data[i] = 0;
			}
			continue;
		}
		if( i == 0 ) {
			result.data[0] += aValue;
			if( result.data[0] < aValue ){
				carry = 1;
			} else {
				carry = 0;
			}
		} else {
			carry = 0;
		}
	}
	if( carry ) {
		result.grow( size+1 );
		result.data[result.size-1] = 1;
	}

	return( result );
}


/******************************************************************

	operator-( CUInt & );
	operator-( unsigned int );

******************************************************************/

CUInt CUInt::operator-( const CUInt &aValue ) const
{
	CUInt	result;
	unsigned int	borrow;
	int		i;


	MASSERT( size >= aValue.size );
	result.grow(  size );
	borrow = 0;
	for( i = 0; i < result.size; i++ ) {
		result.data[i] = data[i] - borrow;
		if( result.data[i] > ((unsigned int)0xffffffff - borrow) ) {
			result.data[i] = (unsigned int)0xffffffff - aValue.data[i];
			continue;
		}
		if( i < aValue.size ) {
			result.data[i] -= aValue.data[i];
			if( result.data[i] > ((unsigned int)0xffffffff - aValue.data[i]) ){
				borrow = 1;
			} else {
				borrow = 0;
			}
		} else {
			borrow = 0;
		}
	}
	MASSERT( borrow == 0 );

	result.calibrate();
	return( result );
}

CUInt CUInt::operator-( unsigned int aValue ) const
{
	CUInt	result;
	unsigned int	borrow;
	int		i;


	result.grow( size );
	borrow = 0;
	for( i = 0; i < size; i++ ) {
		result.data[i] = data[i] - borrow;
		if( result.data[i] > ((unsigned int)0xffffffff - borrow) ) {
			if( i == 0 ) {
				result.data[i] = (unsigned int)0xffffffff - aValue;
			} else {
				result.data[i] = 0xffffffff;
			}
			continue;
		}
		if( i == 0 ) {
			result.data[0] -= aValue;
			if( result.data[0] > ((unsigned int)0xffffffff - aValue) ){
				borrow = 1;
			} else {
				borrow = 0;
			}
		} else {
			borrow = 0;
		}
	}
	MASSERT( borrow == 0 );

	result.calibrate();
	return( result );
}


/******************************************************************

	operator*( CUInt & );
	operator*( unsigned int );

******************************************************************/

CUInt CUInt::operator*( const CUInt &aValue ) const
{
	CUInt result;
	int		  i;
	int		  j;
	CUInt	  partial;	
	unsigned int	  high1;
	unsigned int	  high2;
	unsigned int	  low1;
	unsigned int	  low2;


	result = 0;
	for( i = 0; i < aValue.size; i++ ) {
		for( j = 0; j < size; j++ ) {
			high1 = aValue.data[i] >> 16;
			high2 = data[j] >> 16;
			low1 = aValue.data[i] & 0xffff;
			low2 = data[j] & 0xffff;
			partial = high1*high2;
			partial = partial << 16;
			partial = partial + high1*low2;
			partial = partial + high2*low1;
			partial = partial << 16;
			partial = partial + low1*low2;
			result = result + (partial << ((i+j)*32) );
		}
	}	

	return( result );
}

CUInt CUInt::operator*( unsigned int aValue ) const
{
	CUInt result;
	int		  j;
	CUInt	  partial;	
	unsigned int	  high1;
	unsigned int	  high2;
	unsigned int	  low1;
	unsigned int	  low2;


	result = 0;
	for( j = 0; j < size; j++ ) {
		high1 = aValue >> 16;
		high2 = data[j] >> 16;
		low1 = aValue & 0xffff;
		low2 = data[j] & 0xffff;
		partial = high1*high2;
		partial = partial << 16;
		partial = partial + high1*low2;
		partial = partial + high2*low1;
		partial = partial << 16;
		partial = partial + low1*low2;
		result = result + (partial << (j*32) );
	}	

	return( result );
}



/******************************************************************

	operator/( CUInt & );
	operator/( unsigned int );

******************************************************************/

CUInt CUInt::operator/( const CUInt &aValue ) const
{
	CUInt quotient;
	int		  aBitCount;
	int		  bBitCount;
	int		  aWordCount;
	int		  bWordCount;
	int		  temp1;
	CUInt	  acc1;
	CUInt	  temp;
	unsigned int	  a1;
	unsigned int	  a2;
	unsigned int	  q;
	unsigned int	  b;
	int		  i;
	int		  shift;

	aBitCount = CountBits( data, size );
	aWordCount = (aBitCount+sizeof(data[0])*8-1)/(sizeof(data[0])*8);
	bBitCount = CountBits( aValue.data, aValue.size );
	bWordCount = (bBitCount+sizeof(data[0])*8-1)/(sizeof(data[0])*8);
	MASSERT( bBitCount != 0 );
	if( bBitCount < 32 ) {
		b = aValue.data[0];
		b = b << (sizeof(b)*8 - bBitCount);
	} else {
		temp = aValue>>(bBitCount-32);
		b = temp.GetUINT32();
	}
	
	acc1 = *this;
	quotient = 0;
	for( i = aWordCount-bWordCount; i >=0; i-- ) {
		quotient = quotient << sizeof(data[0])*8;
		temp1 = 32 - bBitCount%32;
		if( temp1 == 32 ) {
			temp1 = 0;
		}
		shift = (bWordCount+i-1)*32 - temp1; 
		if( shift < 0 ) {
			temp = acc1 << -shift;
		} else {
			temp = acc1 >> shift;
		}
		if( temp.size > 1 ) {
			a1 = temp.data[1];
		} else {
			a1 = 0;
		}
		a2 = temp.data[0];
		if( b == 0xffffffff ) {
			q = a1;
		} else {
			q = Divide64By32( a1, a2, b+1 );
		}
		quotient = quotient + q;
		temp = aValue << i*sizeof(data[0])*8;
		acc1 = acc1 - (temp)  * q;
		while( acc1 >= temp ) {
			acc1 = acc1 - temp;
			quotient = quotient+1;
		}
	}
	
	
	return( quotient );
}

CUInt CUInt::operator/( unsigned int aValue ) const
{
	CUInt result;
	CUInt	  temp;

	temp = aValue;
	result = *this/temp;
	
	return( result );
}



/******************************************************************

	operator%( CUInt & );
	operator%( unsigned int );

******************************************************************/


CUInt CUInt::operator%( const CUInt &aValue ) const
{
	CUInt  	quotient;
	int		  	aBitCount;
	int		 	bBitCount;
	int		  	aWordCount;
	int		  	bWordCount;
	int			temp1;
	CUInt  	acc1;
	CUInt	temp;
	unsigned int	  a1;
	unsigned int	  a2;
	unsigned int	  q;
	unsigned int	  b;
	int		  i;
	int		  shift;

	aBitCount = CountBits( data, size );
	aWordCount = (aBitCount+sizeof(data[0])*8-1)/(sizeof(data[0])*8);
	bBitCount = CountBits( aValue.data, aValue.size );
	bWordCount = (bBitCount+sizeof(data[0])*8-1)/(sizeof(data[0])*8);
	MASSERT( bBitCount != 0 ) ;
	if( bBitCount < 32 ) {
		b = aValue.data[0];
		b = b << (sizeof(b)*8 - bBitCount);
	} else {
		temp = aValue>>(bBitCount-32);
		b = temp.GetUINT32();
	}
	
	acc1 = *this;
	quotient = 0;
	for( i = aWordCount-bWordCount; i >=0; i-- ) {
		quotient = quotient << sizeof(data[0])*8;
		temp1 = 32 - bBitCount%32;
		if( temp1 == 32 ) {
			temp1 = 0;
		}
		shift = (bWordCount+i-1)*32 - temp1; 
		if( shift < 0 ) {
			temp = acc1 << -shift;
		} else {
			temp = acc1 >> shift;
		}
		if( temp.size > 1 ) {
			a1 = temp.data[1];
		} else {
			a1 = 0;
		}
		a2 = temp.data[0];
		if( b == 0xffffffff ) {
			q = a1;
		} else {
			q = Divide64By32( a1, a2, b+1 );
		}
		quotient = quotient + q;
		temp = aValue << i*sizeof(data[0])*8;
		acc1 = acc1 - (temp)  * q;
		while( acc1 >= temp ) {
			acc1 = acc1 - temp;
			quotient = quotient+1;
		}
	}
	
	
	return( acc1 );
}

CUInt CUInt::operator%( unsigned int aValue ) const
{
	CUInt	result;
	CUInt		temp;

	temp = aValue;
	result = *this % temp;	
	return( result );
}



/******************************************************************

	operator>>( CUInt & );
	operator>>( unsigned int );

******************************************************************/

CUInt CUInt::operator>>( const CUInt &aValue ) const
{
	CUInt	result;
	int			wordCount;
	int			bitCount;
	unsigned int		acc1;
	unsigned int		acc2;
	int			i;

	if( aValue > 0xffffffff ) {
	    result = 0;
	    result.calibrate();
	    return( result );
	}
	wordCount = aValue.GetUINT32()/32;
	bitCount = aValue.GetUINT32()%32;	
	if( wordCount < size ) {
		result.grow( size-wordCount );
	} else {
		result = 0;
		return( result );
	}
	if( wordCount != 0  ) {
		for( i=0; i<(size-wordCount); i++ ) {
			result.data[i] = data[i+wordCount];		
		}
	} else {
		result = *this;
	}
	if( bitCount != 0 ) {
		acc2 = 0;
		for( i = result.size-1; i >= 0; i-- ) {
			acc1 = (result.data[i] >> bitCount) | acc2;
			acc2 = result.data[i] << (32-bitCount);
			result.data[i] = acc1;
		}
	}

	result.calibrate();
	return( result );
}

CUInt CUInt::operator>>( unsigned int aValue ) const 
{
	CUInt	result;
	int			wordCount;
	int			bitCount;
	unsigned int		acc1;
	unsigned int		acc2;
	int			i;

	wordCount = aValue/32;
	bitCount = aValue%32;	
	if( wordCount < size ) {
		result.grow( size - wordCount );
	} else {
		result = 0;
		return( result );
	}
	if( wordCount != 0  ) {
		for( i=0; i<(size-wordCount); i++ ) {
			result.data[i] = data[i+wordCount];		
		}
	} else {
		result = *this;
	}
	if( bitCount != 0 ) {
		acc2 = 0;
		for( i = result.size-1; i >= 0; i-- ) {
			acc1 = (result.data[i] >> bitCount) | acc2;
			acc2 = result.data[i] << (32-bitCount);
			result.data[i] = acc1;
		}
	}

	result.calibrate();
	return( result );
}


/******************************************************************

	operator<<( CUInt & );
	operator<<( unsigned int );

******************************************************************/

CUInt CUInt::operator<<( const CUInt &aValue ) const
{
	CUInt	result;
	int			wordCount;
	int			bitCount;
	unsigned int		acc1;
	unsigned int		acc2;
	int			i;

	if( aValue > 0xffffffff ) {
	    result = 0;
	    result.calibrate();
	    return( result );
	}
	wordCount = aValue.GetUINT32()/32;
	bitCount = aValue.GetUINT32()%32;	
	if( wordCount != 0  ) {
		result.grow( size + wordCount + 1 );
		for( i=result.size-1; i>=0; i-- ) {
			if( i < wordCount ) {
				result.data[i] = 0;
			} else if( i-wordCount >= size ) {
				result.data[i] = 0;
			} else {
				result.data[i] = data[i-wordCount];		
			}
		}
		for( i=result.size-1; i>=0; i-- ) {
			if( i < wordCount ) {
				result.data[i] = 0;
			} else if( i-wordCount > size-1 ) {
				result.data[i] = 0;
			} else {
				result.data[i] = data[i-wordCount];		
			}
		}
	} else {
		result = *this;
		result.grow( size + 1 );
	}
	if( bitCount != 0 ) {
		acc2 = 0;
		for( i = 0; i < result.size; i++ ) {
			acc1 = (result.data[i] << bitCount) | acc2;
			acc2 = result.data[i] >> (32-bitCount);
			result.data[i] = acc1;
		}
	}

	result.calibrate();
	return( result );
}

CUInt CUInt::operator<<( unsigned int aValue ) const
{
	CUInt	result;
	int			wordCount;
	int			bitCount;
	unsigned int		acc1;
	unsigned int		acc2;
	int			i;

	wordCount = aValue/32;
	bitCount = aValue%32;	
	if( wordCount != 0  ) {
		result.grow( size + wordCount + 1 );
		for( i=size-1; i>=0; i-- ) {
			if( i < wordCount ) {
				result.data[i] = 0;
			} else if( i-wordCount >= size ) {
				result.data[i] = 0;
			} else {
				result.data[i] = data[i-wordCount];		
			}
		}
		for( i=result.size-1; i>=0; i-- ) {
			if( i < wordCount ) {
				result.data[i] = 0;
			} else if( i-wordCount > size-1 ) {
				result.data[i] = 0;
			} else {
				result.data[i] = data[i-wordCount];		
			}
		}
	} else {
		result = *this;
		result.grow( size + 1 );
	}
	if( bitCount != 0 ) {
		acc2 = 0;
		for( i = 0; i < result.size; i++ ) {
			acc1 = (result.data[i] << bitCount) | acc2;
			acc2 = result.data[i] >> (32-bitCount);
			result.data[i] = acc1;
		}
	}

	result.calibrate();
	return( result );
}


/******************************************************************

	operator&( CUInt & );

******************************************************************/

CUInt CUInt::operator&( const CUInt &aValue ) const
{
	CUInt	result;
	int	i;
	int 	newSize;

	newSize = size < aValue.size ? aValue.size : size;
	

	result.grow( newSize );
	for( i=0; i < newSize; i++ ) {
		if( i < size && i < aValue.size ) {
			result.data[i] = data[i] & aValue.data[i];		
		} else {
			result.data[i] = 0;
		}
	}

	result.calibrate();
	return( result );
}

CUInt CUInt::operator&( unsigned int aValue ) const
{
	CUInt	result;
	CUInt	temp;

	temp = aValue;
	result = *this & temp;	
	return( result );
}

/******************************************************************

	operator~( );

******************************************************************/

CUInt CUInt::operator~( void ) const
{
	CUInt	result;
	int	i;

	result.grow( size );
	for( i=0; i < size; i++ ) {
		result.data[i] = ~data[i];		
	}

	result.calibrate();
	return( result );
}

/******************************************************************

	operator|( CUInt & );

******************************************************************/

CUInt CUInt::operator|( const CUInt &aValue ) const
{
	CUInt	result;
	int	i;
	int 	newSize;

	newSize = size < aValue.size ? aValue.size : size;
	
	result.grow( newSize );
	for( i=0; i < newSize; i++ ) {
		if( i < size && i < aValue.size ) {
			result.data[i] = data[i] | aValue.data[i];		
		} else if( i < size ) {
			result.data[i] = data[i];
		} else if( i < aValue.size ) {
			result.data[i] = aValue.data[i];
		} else {
			result.data[i] = 0;
		}
	}

	result.calibrate();
	return( result );
}

CUInt CUInt::operator|( unsigned int aValue ) const
{
	CUInt	result;
	CUInt	temp;

	temp = aValue;
	result = *this | temp;	
	return( result );
}

/******************************************************************

	operator^( CUInt & );

******************************************************************/

CUInt CUInt::operator^( const CUInt &aValue ) const
{
	CUInt	result;
	int	i;
	int 	newSize;

	newSize = size < aValue.size ? aValue.size : size;
	
	result.grow( newSize );
	for( i=0; i < newSize; i++ ) {
		if( i < size && i < aValue.size ) {
			result.data[i] = data[i] ^ aValue.data[i];		
		} else if( i < size ) {
			result.data[i] = data[i];
		} else if( i < aValue.size ) {
			result.data[i] = aValue.data[i];
		} else {
			result.data[i] = 0;
		}
	}

	result.calibrate();
	return( result );
}

CUInt CUInt::operator^( unsigned int aValue ) const
{
	CUInt	result;
	CUInt	temp;

	temp = aValue;
	result = *this ^ temp;	
	return( result );
}

/******************************************************************

	operator<( CUInt & );
	operator<( unsigned int );

******************************************************************/
	
int CUInt::operator<( const CUInt &aValue ) const
{
	int	i;
	
	if( size < aValue.size ) {
		return( TRUE );
	} else if( size > aValue.size ) {
		return( FALSE );
	}
	for( i = size-1; i >= 0; i-- ) {
		if( data[i] < aValue.data[i] ) {
			return( TRUE );
		} else if( data[i] > aValue.data[i] ) {
			return( FALSE );
		}
	} 
	return( FALSE );
}
	
int CUInt::operator<( unsigned int aValue ) const
{

	if( size != 1 ) {
		return( FALSE );
	}
	return( data[0] < aValue );
}

/******************************************************************

	operator<=( CUInt & );
	operator<=( int );

******************************************************************/
	
int CUInt::operator<=( const CUInt &aValue ) const
{
	int	i;
	
	if( size < aValue.size ) {
		return( TRUE );
	} else if( size > aValue.size ) {
		return( FALSE );
	}
	for( i = size-1; i >= 0; i-- ) {
		if( data[i] < aValue.data[i] ) {
			return( TRUE );
		} else if( data[i] > aValue.data[i] ) {
			return( FALSE );
		}
	} 
	return( TRUE );
}
	
int CUInt::operator<=( unsigned int aValue ) const
{

	if( size != 1 ) {
		return( FALSE );
	}
	return( data[0] <= aValue );
}

/******************************************************************

	operator>( CUInt & );
	operator>( unsigned int );

******************************************************************/
	
int CUInt::operator>( const CUInt &aValue ) const
{
	int	i;
	
	if( size > aValue.size ) {
		return( TRUE );
	} else if( size < aValue.size ) {
		return( FALSE );
	}
	for( i = size-1; i >= 0; i-- ) {
		if( data[i] > aValue.data[i] ) {
			return( TRUE );
		} else if( data[i] < aValue.data[i] ) {
			return( FALSE );
		}
	} 
	return( FALSE );
}
	
int CUInt::operator>( unsigned int aValue ) const
{

	if( size != 1 ) {
		return( TRUE );
	}
	return( data[0] > aValue );
}

/******************************************************************

	operator>=( CUInt & );
	operator>=( unsigned int );

******************************************************************/
	
int CUInt::operator>=( const CUInt &aValue ) const
{
	int	i;
	
	if( size > aValue.size ) {
		return( TRUE );
	} else if( size < aValue.size ) {
		return( FALSE );
	}
	for( i = size-1; i >= 0; i-- ) {
		if( data[i] > aValue.data[i] ) {
			return( TRUE );
		} else if( data[i] < aValue.data[i] ) {
			return( FALSE );
		}
	} 
	return( TRUE );
}
	
int CUInt::operator>=( unsigned int aValue ) const
{

	if( size != 1 ) {
		return( TRUE );
	}
	return( data[0] >= aValue );
}

/******************************************************************

	operator==( CUInt & );
	operator==( unsigned int );

******************************************************************/
	
int CUInt::operator==( const CUInt &aValue ) const
{
	int	i;
	
	if( size != aValue.size ) {	
		return( FALSE );	
	}
	for( i = size-1; i >= 0; i-- ) {
		if( data[i] != aValue.data[i] ) {
			return( FALSE );
		}
	} 
	return( TRUE );
}
	
	
int CUInt::operator==( unsigned int aValue ) const
{

	if( size != 1 ) {
		return( FALSE );
	}
	return( data[0] == aValue );
}
	

/******************************************************************

	operator!=( CUInt & );
	operator!=( unsigned int );

******************************************************************/
	
int CUInt::operator!=( const CUInt &aValue ) const
{
	int	i;
	
	if( size != aValue.size ) {
		return( TRUE );	
	}
	for( i = size-1; i >= 0; i-- ) {
		if( data[i] != aValue.data[i] ) {
			return( TRUE );
		}
	} 
	return( FALSE );
}
	
	
int CUInt::operator!=( unsigned int aValue ) const
{

	if( size != 1 ) {
		return( TRUE );
	}	
	return( data[0] != aValue );
}

/******************************************************************
  	Size
	- returns number of non-zero significant bits
 ******************************************************************/
UINT32 CUInt::Size() 
{ 
	return CountBits( data, size );
}	

/******************************************************************

	GetUINT32()

******************************************************************/
	
UINT32 CUInt::GetUINT32() const
{
	return( data[0] );
}
	
/******************************************************************

	GetUINT64()

******************************************************************/
	
UINT64 CUInt::GetUINT64() const
{
	if( size >= 2 ) {
	    return (((UINT64)data[1])<< 32) | data[0];
	} else {
	    return data[0];
	}
}
	


/******************************************************************

	CountBits	count the number of significant bits

******************************************************************/

int	CountBits( unsigned int *data, int size )
{
	int		ones;
	int		zeros;
	int		i;
	unsigned int	acc;	

	zeros = 0;
	for( i = size - 1; i >= 0; i-- ) {
		if( data[i] == 0 ) {
			zeros += 32;
			continue;
		}
		acc = data[i];
		ones  = 0;
		while( acc != 0 ) {
			acc = acc >> 1;
			ones++;
		}
		zeros += sizeof(data[0])*8 - ones;		
		break;
	}	
	return( size*sizeof(data[0])*8 - zeros );	
}

/******************************************************************

	Divide64by32
		-	divide 64 bits by 32 bits unsigned

******************************************************************/

unsigned int	Divide64By32( unsigned int b1, unsigned int b2, unsigned int c )
{
	unsigned int	accHigh;
	unsigned int	accLow;
	unsigned int 	cHigh;
	unsigned int 	cLow;
	unsigned int  	aHigh;
	unsigned int  	aLow;
	unsigned int 	u;
	unsigned int	v;

	cHigh = c >> 16;
	cLow = c & 0xffff;
	accHigh = b1;
	accLow  = b2;

	if (cHigh == 0xffff) {
		aHigh = accHigh >> 16;
	} else {
		aHigh = accHigh / (cHigh + 1);
	}

  	u = aHigh * cLow;
  	v = aHigh * cHigh;
	if( (accLow -= u<<16) > (0xffffffff - (u<<16)) ) {
    		accHigh--;
	}
  	accHigh -= u >> 16;
  	accHigh -= v;

  	while ((accHigh > cHigh) ||
         (( accHigh == cHigh) && (accLow >= (cLow<<16)))) {
    		if ((accLow -= cLow<<16)
          		> 0xffffffff - (cLow<<16)) {
      			accHigh--;
		}
    		accHigh -= cHigh;
    		aHigh++;
  	}

	
  	if (cHigh == 0xffff ) {
    		aLow = accHigh & 0xffff;
	} else {
    		aLow = ((accHigh << 16) + (accLow>>16))/(cHigh+1);
	}
  	u = aLow * cLow;
  	v = aLow * cHigh;
  	if ((accLow -= u) > (0xffffffff - u)) {
    		accHigh--;
	}
  	if ((accLow -= v<<16) > (0xffffffff - (v<<16))) {
    		accHigh--;
	}
	accHigh -= v >> 16;

  	while ((accHigh > 0) || ((accHigh == 0) && accLow >= c)) {
    		if ((accLow -= c) > (0xffffffff - c)) {
      			accHigh--;
		}
    		aLow++;
	}

	return( (aHigh<<16)+aLow );
}
