/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/*******************************************************************
 *
 *	cuint.hpp	
 *		- header file for cuint.hpp 
 *		  
 *
 *******************************************************************
 */

#ifndef CUINT_HPP
#define CUINT_HPP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "glue.h"
#include "cobstack.h"


#ifndef MAX
#define MAX(a,b)	((a)<(b)?(b):(a))
#endif // MAX

#define	CUINT_DEFAULT_SIZE	(4)	// size in 32 bit words

/**
 * Infinite precision unsigned arithmetic class
 * Storage within object will be used whenever possible.
 */
class	CUInt {
private:
    unsigned int	defaultData[CUINT_DEFAULT_SIZE];///< default data array
    unsigned int	*data;				///< pointer to data
    int		allocatedSize;			///< actual allocated size
    int		size;				///< data size in 32 bit words 
    CObstack*	heap;				///< heap to use for allocation
public:

    /**
     * Create a unsigned integer object
     */
    CUInt( void ) 
    {
    	heap = NULL;
    	allocatedSize = CUINT_DEFAULT_SIZE;
    	data = defaultData;
    	size = 1;
    	data[0] = 0;
    };		
    
    /**
     * Create a unsigned integer object initialize with value
     * \param aValue initial value.
     */
    CUInt( unsigned int aValue ) 
    {
    	heap = NULL;
    	allocatedSize = CUINT_DEFAULT_SIZE;
    	data = defaultData;
    	size = 1;
    	data[0] = aValue;
    };		
    
    
    /**
     * Create a copy of unsigned integer object.
     * \param aValue reference to object to be copied.
     */
    CUInt( const CUInt &aValue ) 
    {
    	INT32	i;
    	
    	heap = NULL;
    	allocatedSize = CUINT_DEFAULT_SIZE;
    	data = defaultData;
    	size = 1;
    	grow( aValue.size );
    	for( i = 0; i < size; i++ ) {
    		data[i] = aValue.data[i];
    	}
    };		
    
    /**
     * Destory object and free any storage.
     */
    virtual ~CUInt()	
    {
    	if( heap == NULL && data != defaultData ) {
    		shell_xfree( data, size );
    		data = NULL;
    	}
    } ;
    
    
    /**
     * Set heap to be used for storage allocation.
     * \param aHeap obstack to be used for storage.
     */
    void SetHeap( CObstack* aHeap )
    {
    	heap = aHeap;
    }
    
    /** \name Assignment operators */ 
    //@{
    const CUInt &operator=( const CUInt &aValue )	 
    {
    	INT32	i;
    	
    	grow( aValue.size );
    	for( i = 0; i < size; i++ ) {
    		data[i] = aValue.data[i];
    	}
    	return( aValue );
    };
    
    const UINT64 operator=( const UINT64 aValue )	 
    {
    	if( aValue>>32 != 0 ) {
    		grow( 2 );
    		data[0] = aValue;
    		data[1] = aValue>>32;
    	} else {
    		grow( 1 );
    		data[0] = aValue;
    	}
    	return( aValue );
    };
    //@}
    
    /**
     * Change size of unsigned integer.
     * Reduction in size truncates integer.
     * Increase in size allocate space for max value.
     * \param newSize new size if integer in bits.
     */
    void	Truncate( int newSize )
    {
    	grow( newSize );
    };

    /** \name Unary operators */
    //@{
    CUInt operator~( void ) const;
    //@}
    
    /** \name Binary operators */
    //@{
    CUInt operator+( const CUInt & ) const;
    CUInt operator+( const unsigned int ) const;
    CUInt operator-( const CUInt & ) const;
    CUInt operator-( unsigned int ) const;
    CUInt operator*( const CUInt & ) const;
    CUInt operator*( unsigned int ) const;
    CUInt operator/( const CUInt & ) const;
    CUInt operator/( unsigned int ) const;
    CUInt operator%( const CUInt & ) const;
    CUInt operator%( unsigned int ) const;
    CUInt operator>>( const CUInt & ) const;
    CUInt operator>>( unsigned int ) const;
    CUInt operator<<( const CUInt & ) const;
    CUInt operator<<( unsigned int ) const;
    CUInt operator&( const CUInt & ) const;
    CUInt operator&( unsigned int ) const;
    CUInt operator|( const CUInt & ) const;
    CUInt operator|( unsigned int ) const;
    CUInt operator^( const CUInt & ) const;
    CUInt operator^( unsigned int ) const;
    //@}
    
    /** \name Comparision operators */
    //@{
    int operator<( const CUInt & ) const;
    int operator<( unsigned int ) const;
    int operator<=( const CUInt & ) const;
    int operator<=( unsigned int ) const;
    int operator>( const CUInt & ) const;
    int operator>( unsigned int ) const;
    int operator>=( const CUInt & ) const;
    int operator>=( unsigned int ) const;
    int operator==( const CUInt & ) const;
    int operator==( unsigned int ) const;
    int operator!=( const CUInt & ) const;
    int operator!=( unsigned int ) const;
    //@}
    
    /** \name Cast operations */
    //@{
    UINT32 GetUINT32() const; 
    UINT64 GetUINT64() const;
    //@}
    
    /**
     * Calculate the number of non-zero significant bits.
     * \return number non-zero significant bits.
     */
    UINT32 Size();
    
    /**
     * \cond
     */
private:
    void calibrate( void )
    {
    	int	i;
    	for( i=size-1; i>0; i-- ) {
    	    if( data[i] != 0 ) {
    		break;
    	    }
    	}	
    	size = i+1;
    	return;
    };
    void grow( int aSize )
    {
    	int i;
    
    	if( aSize <= size ) {
    	    size = aSize;
    	    return;
    	}
    	if( allocatedSize < aSize ) {
    	    if( heap == NULL ) {
    		if( data != defaultData ) {
    		    unsigned int* newData;
    		    newData = (unsigned int*)shell_xmalloc( aSize*4);
    		    memcpy( newData, data, size*4 );
    		    shell_xfree( data, allocatedSize*4 );
    		    data = newData;
    		} else {
    		    data = (unsigned int*)shell_xmalloc( aSize*4 );
    		    memcpy( data, defaultData, size*4 );
    		}
    	    } else {
    		unsigned int*	newData;
    		newData = (unsigned int*)heap->Alloc( aSize*4 );
    		memcpy( newData, data, size*4 );
    		data = newData;
    	    }
    	    allocatedSize = aSize;
    	}
    	for( i=size; i < aSize; i++ ) {
    	    data[i] = 0;
    	}
    	size = aSize;
    };	
    /**
     * \endcond
     */
};




#endif // CUINT_HPP
