/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cvector.cpp
 *		- methods for bit vector declaration 
 *
 *
 ******************************************************************************
 */


#include "cvector.h"
#include "cnode.h"

int   CVector::bufferSize = 0;
char* CVector::buffer = NULL;
static const int SHEADER_LENGTH = 15; /* max length of n`t */


/***************************************************
	Constructor
 ***************************************************/

CVector::CVector( INT32 width )
 : width(width),
   sized(TRUE),
   based(TRUE),
   unbased(FALSE),
   _signed(FALSE),
   overflowed(FALSE)
{
	MASSERT( width >= 0 );
	preferredBase = 10;
}
/***************************************************
	Destructor	
 ***************************************************/

CVector::~CVector()
{
}
/***************************************************
	Hash
	- convert number into a unique hashid
 ***************************************************/
unsigned CVector::Hash( void )	
{
	unsigned result = 0;
	CUInt	v;
	INT32	i;

	v = aval;
	for( i = 0; i < width; i += 32 ) {
		if( v == 0 ) {
			break;
		}
		result ^= v.GetUINT32();
		v = v >> 32;
	} 

	return( result );
}
/***************************************************
	GetWidth
	- return vectors bit width
 ***************************************************/

INT32	CVector::GetWidth( void )
{
	return width;
}
/***************************************************
	SetWidth
	- change vectors bit width
 ***************************************************/

void	CVector::SetWidth( INT32 newWidth )
{
	MASSERT( newWidth >= 0 );
	/*
 	 * Extend upper bit if new width is larger
 	 * than old. 
 	 * For signed vectors: x -> x, z -> z, 0 -> 0, 1 -> 1 
	 */
	if( newWidth > width ) {
	    CUInt oldMask = ((CUInt)1 << width) - 1;
	    CUInt newMask = ((CUInt)1 << newWidth) - 1;
	    CUInt msbMask = ((CUInt)1 << (width-1));
	    CUInt extensionMask = newMask ^ oldMask;
	    if( _signed ) {
		if( (aval & msbMask) != 0 ) {
	            aval = aval | extensionMask;	
		}
	        if( (bval & msbMask) != 0 ) {
	            bval = bval | extensionMask;	
	        }
	    }
	}
	width = newWidth;
	Adjust();
}
/***************************************************
	SetToX
	- set vectors value to X
 ***************************************************/

void 	CVector::SetToX()
{
    CUInt mask = ((CUInt)1 << width) - 1;
    aval = mask;
    bval = mask;
}
/***************************************************
	operator=
	- assign value 
 ***************************************************/

const CVector& CVector::operator=( const CVector& v )
{
	CUInt	mask;

	width = width < v.width ? width : v.width;
	mask = ((CUInt)1 << width) - 1;
	aval = v.aval & mask;
	bval = v.bval & mask;
	hook = 0;
	
	return v;
}
UINT32 CVector::operator=( UINT32 v )
{
	width = 32;
	aval = v;
	bval = 0;
	hook = 0;
	return v;
}
/***************************************************
	Adjust
	- truncate aval and bval to width 
 ***************************************************/

void	CVector::Adjust( )
{
	int	words;
	CUInt	mask = (((CUInt) 1) << width) - 1;

	words = (width+31)/32;
	aval.Truncate( words );
	bval.Truncate( words );
	aval = aval & mask;
	bval = bval & mask;

}
/***************************************************
	LoadDecimal
	- convert decimal string to vector 
	  return TRUE on successful conversion
 ***************************************************/

int	CVector::LoadDecimal( const char* string )	
{
	int result = TRUE;
	UINT32	digit;
	const char*	ptr;

	/*
 	 * special case for single xXzZ?
 	 */
	if( strlen(string) == 1 ) {
	    switch( *string ) {
	    case 'x':
	    case 'X':
	    case 'z':
	    case 'Z':
	    case '?':
		result = LoadHex( string );
		goto DONE;
	    default:
		break;
	    }
	}
	aval = 0;
	bval = 0;
	hook = 0;
	for( ptr = string; *ptr; ptr++ ) {
		if( *ptr == '_' ) {
			continue;
		}
		if( *ptr < '0' || *ptr > '9' ) {
			return FALSE;
		}				
		digit = *ptr - '0';			
		aval = aval * 10;
		aval = aval + digit;
	}	
	if( aval.Size() > width ) {
	    overflowed = 1;
	}
	
DONE:
	Adjust();
	preferredBase = 10;
	return result;
}
/***************************************************
	GetVString
	- convert vector to verilog string 
 ***************************************************/
char*	CVector::GetVString( void )	
{
	CVector		v(GetWidth());
	char*   	s;
	char		prefix[14];
	unsigned	i;
	const char* sgnd = "";
	
	if( Unbased() ) {
	    prefix[0] = '\'';
	    prefix[1] = 0;
	    s = GetBinary();
	    goto CLEANUP;
	}

	if( Signed() ) {
	    sgnd = "s";
	}
	
	
	switch( preferredBase ) {
	case 0:
		goto STRING;
	case 2:
		goto BINARY;
	case 8:
		goto OCTAL;
	case 10:
		goto DECIMAL;
	case 16:
		goto HEX;
	default:
		MASSERT( FALSE );
	}

DECIMAL:
	s = GetDecimal();
	v.LoadDecimal( s );
	if( v == *this ) {
		if( sized ) {
			sprintf( prefix, "%ld'%sd", GetWidth(), sgnd );
		} else if( (based && Signed()) || HasXZ() ) {
			sprintf( prefix, "'%sd", sgnd );
		} else {
			prefix[0] = 0;
		}
		goto CLEANUP;
	}
	
HEX:
	s = GetHex();
	v.LoadHex( s );
	if( v == *this ) {
		if( sized ) {
			sprintf( prefix, "%ld'%sh", GetWidth(), sgnd );
		} else {
			sprintf( prefix, "'%sh", sgnd );
		}
		goto CLEANUP;
	}

OCTAL:
	s = GetOctal();
	v.LoadOctal( s );
	if( v == *this ) {
		if( sized ) {
			sprintf( prefix, "%ld'%so", GetWidth(), sgnd );
		} else {
			sprintf( prefix, "'%so", sgnd );
		}
		goto CLEANUP;
	}
		
BINARY:
	if( sized ) {
		sprintf( prefix, "%ld'%sb", GetWidth(), sgnd );
	} else {
		sprintf( prefix, "'%sb", sgnd );
	}
	s =GetBinary();
	goto CLEANUP;

STRING:
	s = GetString();
	return s;

CLEANUP:
	/*
	 * strip leading 0's if constant isn't sized
	 */
        if( !sized && strpbrk( s, "xXzZ" ) == NULL ) {
		while( s[0] == '0' && s[1] != 0 ) { 
			s++;
		}
	}
	/*
	 * prepend prefix
	 */
	s -= strlen(prefix);
	for( i = 0; i < strlen(prefix); i++ ) {
		s[i] = prefix[i];
	}
	return s;
	

}
/***************************************************
	GetDecimal
	- convert vector to decimal string 
 ***************************************************/
char*	CVector::GetDecimal( void )	
{
	char*	ptr;
	CUInt	acc;
	CUInt	digit;
	CUInt	mask;

	GrowBuffer( width/3+1+SHEADER_LENGTH );
	mask = (((CUInt) 1) << width) - 1;
	ptr = buffer + bufferSize - 1;
	*ptr = 0;
	if( (~bval & mask) == 0 && aval == 0 ) {
		*--ptr = 'z';
		return ptr;
	} 
	if( (~bval & mask) == 0 && (~aval & mask) == 0 ) {
		*--ptr = 'x';
		return ptr;
	}
	if( (bval & aval & mask) != 0 ) {
		*--ptr = 'X';
		return ptr;
	}
	if( (bval & mask) != 0 ) {
		*--ptr = 'Z';
		return ptr;
	}
	acc = aval;
	do {
		digit = acc % 10;
		*--ptr = digit.GetUINT32() + '0';
		acc = acc / 10;
	} while( acc != 0 ); 

	return ptr; 
}
/*********************************************
	LoadBinary
	- convert binary string to vector 
	  return TRUE on successful conversion
 ***************************************************/

int	CVector::LoadBinary( const char* string )	
{
	const char*	ptr;
	CUInt	mask = (((CUInt) 1) << width) - 1;
	const char	leading = *string;
	CUInt	se = mask;

	aval = 0;
	bval = 0;
	hook = 0;
	switch( leading ) {
	case 'x':
	case 'X': 
	case 'z':
	case 'Z': 
	case '?':
	    string++;
	    break;
	}

	for( ptr = string; *ptr; ptr++ ) {
		if( *ptr == '_' ) {
			continue;
		}
		aval = aval << 1;
		bval = bval << 1;
		hook = hook << 1;
		se   = se << 1;
		switch( *ptr ) {
		case 'x':
		case 'X':
			aval = aval | 1;
			bval = bval | 1;
			break;
		case '?':
			bval = bval | 1;
			hook = hook | 1;
			break;
		case 'z':
		case 'Z':
			bval = bval | 1;
			break;
		case '0':
			break;
		case '1':
			aval = aval | 1;
			break;
		default:
			return FALSE;
		}
	}	

	if( aval.Size() > width ||
	    bval.Size() > width ) {
	    overflowed = 1;
	}
	switch( leading ) {
	case 'x':
	case 'X': 
	    aval = aval | se;
	    bval = bval | se;
	    break;
	case 'z':
	case 'Z': 
	    bval = bval | se;
	    break;
	case '?':
	    bval = bval | se;
	    hook = hook | se;
	    break;
	}
	Adjust();
	preferredBase = 2;
	return TRUE;
}
/***************************************************
	GetBinary
	- convert vector to binary string 
 ***************************************************/
char*	CVector::GetBinary( void )	
{
	char*	ptr;
	CUInt	acc;
	CUInt	bcc;
	CUInt   hcc;
	CUInt	digit;
	CUInt	digitx;
	CUInt   digith;
	INT32	i;

	GrowBuffer( width+1+SHEADER_LENGTH );
	ptr = buffer + bufferSize - 1;
	*ptr = 0;
	acc = aval;
	bcc = bval;
	hcc = hook;
	for( i = 0; i < width; i++ ) {
		digit = acc & 1; 
		digitx = bcc & 1;
		digith = hcc & 1;
		if( digit.GetUINT32() ) {
			if( digitx.GetUINT32() ) {
				*--ptr = 'x';
			} else {
				*--ptr = '1';
			}
		} else {
			if( digitx.GetUINT32() ) {
				if( digith.GetUINT32() ) {
					*--ptr = '?';
				} else {
					*--ptr = 'z';
				}
			} else {
				*--ptr = '0';
			}
		}
				
		acc = acc >> 1;
		bcc = bcc >> 1;
		hcc = hcc >> 1;
	}; 

	return( ptr );
}
/*********************************************
	LoadOctal
	- convert octal string to vector 
	  return TRUE on successful conversion
 ***************************************************/

int	CVector::LoadOctal( const char* string )	
{
	const char*	ptr;
	CUInt	mask = (((CUInt) 1) << width) - 1;
	const char	leading = *string;
	CUInt	se = mask;

	aval = 0;
	bval = 0;
	hook = 0;
	switch( leading ) {
	case 'x':
	case 'X': 
	case 'z':
	case 'Z': 
	case '?':
	    string++;
	    break;
	}

	for( ptr = string; *ptr; ptr++ ) {
		if( *ptr == '_' ) {
			continue;
		}
		aval = aval << 3;
		bval = bval << 3;
		hook = hook << 3;
		se = se << 3;
		switch( *ptr ) {
		case 'x':
		case 'X':
			aval = aval | 7;
			bval = bval | 7;
			break;
		case '?':
			bval = bval | 7;
			hook = hook | 7;
			break;
		case 'z':
		case 'Z':
			bval = bval | 7;
			break;
		default:
			if( *ptr < '0' || *ptr > '7' ) {
				return FALSE;
			}
			aval = aval | (*ptr -'0');
			break;
		}
	}	

	if( aval.Size() > width ||
	    bval.Size() > width ) {
	    overflowed = 1;
	}
	switch( leading ) {
	case 'x':
	case 'X': 
	    aval = aval | se;
	    bval = bval | se;
	    break;
	case 'z':
	case 'Z': 
	    bval = bval | se;
	    break;
	case '?':
	    bval = bval | se;
	    hook = hook | se;
	    break;
	}
	Adjust();
	preferredBase = 8;
	return TRUE;
}
/***************************************************
	GetOctal
	- convert vector to octal string 
 ***************************************************/
char*	CVector::GetOctal( void )	
{
	char*	ptr;
	CUInt	acc;
	CUInt	bcc;
	CUInt   hcc;
	UINT32	digit;
	UINT32	xdigit;
	UINT32  hdigit;
	INT32	i;
	UINT32	mask;

	GrowBuffer( width/3+1+SHEADER_LENGTH );
	ptr = buffer + bufferSize - 1;
	*ptr = 0;
	acc = aval;
	bcc = bval;
	hcc = hook;
	for( i = 0; i < width; i += 3 ) {
		if( width - i < 3 ) {	
			mask = (1 << (width-i)) - 1;
		} else {
			mask = 0x7;
		}
		digit = acc.GetUINT32() & mask; 
		xdigit = bcc.GetUINT32() & mask;
		hdigit = hcc.GetUINT32() & mask;
		if( digit == mask && xdigit == mask ) {
			*--ptr = 'x';
		} else if( digit == 0 && xdigit == mask ) {
			if( hdigit ) {
				*--ptr = '?';
			} else {
				*--ptr = 'z';
			}
		} else if( (xdigit & digit) != 0 ) {
			*--ptr = 'X';
		} else if( (xdigit & ~digit) != 0 ) {
			*--ptr = 'Z';
		} else {
			*--ptr = digit + '0';
		}
				
		acc = acc >> 3;
		bcc = bcc >> 3;
		hcc = hcc >> 3;
	}; 

	return( ptr );
}
/*********************************************
	LoadHex
	- convert hex string to vector 
	  return TRUE on successful conversion
 ***************************************************/

int	CVector::LoadHex( const char* string )	
{
	const char*	ptr;
	CUInt	mask = (((CUInt) 1) << width) - 1;
	const char	leading = *string;
	CUInt	se = mask;

	aval = 0;
	bval = 0;
	hook = 0;
	switch( leading ) {
	case 'x':
	case 'X': 
	case 'z':
	case 'Z': 
	case '?':
	    string++;
	    break;
	}

	for( ptr = string; *ptr; ptr++ ) {
		if( *ptr == '_' ) {
			continue;
		}
		aval = aval << 4;
		bval = bval << 4;
		hook = hook << 4;
		se = se << 4;
		switch( *ptr ) {
		case 'x':
		case 'X':
			aval = aval | 0xf;
			bval = bval | 0xf;
			break;
		case '?':
			bval = bval | 0xf;
			hook = hook | 0xf;
			break;
		case 'z':
		case 'Z':
			bval = bval | 0xf;
			break;
		default:
			if( *ptr >= '0' && *ptr <= '9' ) {
				aval = aval | (*ptr -'0');
			} else if( *ptr >= 'a' && *ptr <= 'f' ) {
				aval = aval | (*ptr -'a' + 10);
			} else if( *ptr >= 'A' && *ptr <= 'F' ) {
				aval = aval | (*ptr -'A' + 10);
			} else {
				return FALSE;
			}
			break;
		}
	}	

	if( aval.Size() > width ||
	    bval.Size() > width ) {
	    overflowed = 1;
	}
	switch( leading ) {
	case 'x':
	case 'X': 
	    aval = aval | se;
	    bval = bval | se;
	    break;
	case 'z':
	case 'Z': 
	    bval = bval | se;
	    break;
	case '?':
	    bval = bval | se;
	    hook = hook | se;
	    break;
	}
	Adjust();
	preferredBase = 16;
	return TRUE;
}
/***************************************************
	GetHex
	- convert vector to hex string 
 ***************************************************/
char*	CVector::GetHex( void )	
{
	char*	ptr;
	CUInt	acc;
	CUInt	bcc;
	CUInt	hcc;
	UINT32	digit;
	UINT32	xdigit;
	UINT32	hdigit;
	INT32	i;
	UINT32	mask;

	GrowBuffer( width/4+1+SHEADER_LENGTH );
	ptr = buffer + bufferSize - 1;
	*ptr = 0;
	acc = aval;
	bcc = bval;
	hcc = hook;
	for( i = 0; i < width; i += 4 ) {
		if( width - i < 4 ) {	
			mask = (1<< (width-i)) - 1;
		} else {
			mask = 0xf;
		}
		digit = acc.GetUINT32() & mask; 
		xdigit = bcc.GetUINT32() & mask;
		hdigit = hcc.GetUINT32() & mask;
		if( digit == mask && xdigit == mask) {
			*--ptr = 'x';
		} else if( digit == 0 && xdigit == mask ) {
			if( hdigit ) {
				*--ptr = '?';
			} else {
				*--ptr = 'z';
			}
		} else if( (xdigit & digit) != 0 ) {
			*--ptr = 'X';
		} else if( (xdigit & ~digit) != 0 ) {
			*--ptr = 'Z';
		} else if( digit < 10 ) {
			*--ptr = digit + '0';
		} else {
			*--ptr = digit - 10 + 'a';
		}
				
		acc = acc >> 4;
		bcc = bcc >> 4;
		hcc = hcc >> 4;
	}; 

	return( ptr );
}
/*********************************************
	LoadString
	- convert ascii string to vector 
	  return TRUE on successful conversion
 ***************************************************/

int	CVector::LoadString( const char* string )	
{
	const char*	ptr;

	aval = 0;
	bval = 0;
	for( ptr = string; *ptr; ptr++ ) {
		aval = aval << 8;
		aval = aval | *ptr;
	}	
	Adjust();
	preferredBase = 0;
	_signed = FALSE;
	return TRUE;
}
/***************************************************
	GetString
	- convert vector to ascii string 
 ***************************************************/
char*	CVector::GetString( void )	
{
	char*	ptr;
	CUInt	acc;
	INT32	i;

	GrowBuffer( width/8+3+SHEADER_LENGTH );
	ptr = buffer + bufferSize - 1;
	*ptr = 0;
	/*
	 * leave room for a " to be inserted 
	 */
	*--ptr = 0;
	acc = aval;
	for( i = 0; i < width; i += 8 ) {
		if( acc == 0 ) {
			break;
		}
		*--ptr = acc.GetUINT32() & 0xff; 
		acc = acc >> 8;
	} 

	return( ptr );
}
/***************************************************
	Add
	- add 2 vectors storing result in third 
 ***************************************************/

void	Add( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = (((CUInt) 1) << dst->width) - 1;
		dst->bval = (((CUInt) 1) << dst->width) - 1;
	} else {
		dst->aval = a->aval + b->aval;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Neg
	- negate vector 
 ***************************************************/

void	Neg( CVector* dst, CVector* a ) 
{
	if( a->bval != 0 ) {
		dst->aval = (((CUInt) 1) << dst->width) - 1;
		dst->bval = (((CUInt) 1) << dst->width) - 1;
	} else {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval =  (a->aval ^ mask) + 1;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Plus
	- unary plus operator 
 ***************************************************/

void	Plus( CVector* dst, CVector* a ) 
{
	dst->aval = a->aval;
	dst->bval = a->bval;
	dst->Adjust();
}
/***************************************************
	Sub
	- subtrace 2 vectors storing result in third 
 ***************************************************/

void	Sub( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = (((CUInt) 1) << dst->width) - 1;
		dst->bval = (((CUInt) 1) << dst->width) - 1;
	} else {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = a->aval + (mask ^ b->aval) + 1;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Mul	
	- multiply 2 vectors storing result in third 
 ***************************************************/

void	Mul( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = (((CUInt) 1) << dst->width) - 1;
		dst->bval = (((CUInt) 1) << dst->width) - 1;
	} else {
		dst->aval = a->aval * b->aval;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Div
	- Divide 2 vectors storing result in third 
 ***************************************************/

void	Div( CVector* dst, CVector* a, CVector* b ) 
{
	if( (a->bval != 0 || b->bval != 0 ) || b->aval == 0 ) {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = mask;
		dst->bval = mask;
	} else if( a->Signed() && b->Signed() ) {
	  	CVector na( a->GetWidth() );
	  	CVector nb( b->GetWidth() );
		na.Signed(1);
		nb.Signed(1);
		int sign = 1;
		if( a->IsNegative() ) {
		    Neg( &na, a );	
		    sign *= -1;
		} else {
		    na = *a;
		}
		if( b->IsNegative() ) {
		    Neg( &nb, b );	
		    sign *= -1;
		} else {
		    nb = *b;
		}
		CVector ndst( dst->GetWidth() );
		ndst.Signed(1);
		ndst.aval = na.aval / nb.aval;
		ndst.bval = 0;
		ndst.Adjust();
		if( sign < 0 ) {
		    Neg( dst, &ndst );
		} else {
		    *dst = ndst;
		}
	} else {
		dst->aval = a->aval / b->aval;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Pow
	- raise a to the b power 
 ***************************************************/

void	Pow( CVector* dst, CVector* a, CVector* b ) 
{
	/*
 	 * check for x or z bits
 	 */
	if( a->bval != 0 || b->bval != 0 ) {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = mask;
		dst->bval = mask;
		return;
	} 

	CUInt ap = a->aval;
	CUInt bv = b->aval;
	dst->aval = 1;
	dst->bval = 0;
	while( bv != 0 ) {
	    if( bv.GetUINT32() & 1 ) {
		dst->aval = dst->aval * ap;
	    }
	    ap = ap * ap;
	    bv = bv >> 1;
	}
	dst->Adjust();
}
/***************************************************
	Lsha	
	- arithmetic shift left 
 ***************************************************/

void	Lsha( CVector* dst, CVector* a, CVector* b ) 
{
	if( b->bval != 0 ) {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = mask;
		dst->bval = mask;
		return;
	}

/*
 * xl does this but not vcs, b is suppose to be treated as unsigned
 */
//	if( b->IsNegative() ) {
//	    CVector nb(b->GetWidth());
//	    nb.Signed(1);
//	    Neg(&nb, b );
//	    Rsha( dst, a, &nb );
//	    return;
//	}

	/*
	 * attempt to prevent the allocation of hugh intermediate storage
	 */
	if( b->aval > dst->GetWidth() ) {
	    dst->aval = 0;
	    dst->bval = 0;
	    return;
	}

	dst->aval = a->aval << b->aval;
	dst->bval = a->bval << b->aval;
	dst->Adjust();
}
/***************************************************
	Rsha	
	- arithmetic shift right 
 ***************************************************/

void	Rsha( CVector* dst, CVector* a, CVector* b ) 
{
	if( b->bval != 0 ) {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = mask;
		dst->bval = mask;
		return;
	} 

/*
 * xl does this but not vcs, b is suppose to be treated as unsigned
 */
//	if( b->IsNegative() ) {
//	    CVector nb(b->GetWidth());
//	    nb.Signed(1);
//	    Neg(&nb, b );
//	    Lsha( dst, a, &nb );
//	    return;
//	}

	CUInt amask = ((CUInt)1 << a->width) - 1;
	CUInt dstmask = ((CUInt)1 << dst->width) - 1;
	CUInt msbMask = ((CUInt)1 << (a->width-1));
	CUInt fillMask;
	if( a->Signed() ) {
	    fillMask = dstmask ^ (amask >> b->aval);
	} else {
	    fillMask =  0;
	}
	if( (a->aval & msbMask) != 0 ) {
	    dst->aval = (a->aval >> b->aval) | fillMask;
	} else {
	    dst->aval = a->aval >> b->aval;
	}
	if( (a->bval & msbMask) != 0 ) {
	    dst->bval = (a->bval >> b->aval) | fillMask;
	} else {
	    dst->bval = a->bval >> b->aval;
	}
	dst->Adjust();

}
/***************************************************
	Lsh	
	- shift left 
 ***************************************************/

void	Lsh( CVector* dst, CVector* a, CVector* b ) 
{
	if( b->bval != 0 ) {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = mask;
		dst->bval = mask;
		return;
	} 

/*
 * xl does this but not vcs, b is suppose to be treated as unsigned
 */
//	if( b->IsNegative() ) {
//	    CVector nb(b->GetWidth());
//	    nb.Signed(1);
//	    Neg(&nb, b );
//	    Rsh( dst, a, &nb );
//	    return;
//	}

	/*
	 * attempt to prevent the allocation of hugh intermediate storage
	 */
	if( b->aval > dst->GetWidth() ) {
	    dst->aval = 0;
	    dst->bval = 0;
	    return;
	}

	dst->aval = a->aval << b->aval;
	dst->bval = a->bval << b->aval;
	dst->Adjust();
}
/***************************************************
	Rsh	
	- shift right 
 ***************************************************/

void	Rsh( CVector* dst, CVector* a, CVector* b ) 
{
	if( b->bval != 0 ) {
		CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
		dst->aval = mask;
		dst->bval = mask;
		return;
	} 

/*
 * xl does this but not vcs, b is suppose to be treated as unsigned
 */
//	if( b->IsNegative() ) {
//	    CVector nb(b->GetWidth());
//	    nb.Signed(1);
//	    Neg(&nb, b );
//	    Lsh( dst, a, &nb );
//	    return;
//	}

	dst->aval = a->aval >> b->aval;
	dst->bval = a->bval >> b->aval;
	dst->Adjust();
}
/***************************************************
	Cat
	- concatinate 
 ***************************************************/

void	Cat( CVector* dst, CVector* a, CVector* b ) 
{
	dst->aval = (a->aval << b->GetWidth()) | b->aval;
	dst->bval = (a->bval << b->GetWidth()) | b->bval;
	dst->Adjust();
}
/***************************************************
	Rep
	- vector replication 
 ***************************************************/

void	Rep( CVector* dst, CVector* a, CVector* b ) 
{
	int	i;
	int	count;	

	count = a->aval.GetUINT32();

	dst->aval = 0;
	dst->bval = 0;
	for( i = 0; i < count; i++ ) {
		dst->aval = (dst->aval << b->GetWidth()) | b->aval;
		dst->bval = (dst->bval << b->GetWidth()) | b->bval;
	}
	dst->Adjust();
}
/***************************************************
	Mod	
	- arithmetic mod 
 ***************************************************/

void	Mod( CVector* dst, CVector* a, CVector* b ) 
{
	if( (a->bval != 0 || b->bval != 0 ) || b->aval == 0 ) {
		dst->aval = (((CUInt) 1) << dst->width) - 1;
		dst->bval = (((CUInt) 1) << dst->width) - 1;
	} else {
		dst->aval = (((CUInt) 1) << dst->width) - 1;
		int neg = a->IsNegative();
		CVector modulus( b->GetWidth() );
		CVector dividend( a->GetWidth() );
		CVector result( dst->GetWidth() );
		if( b->IsNegative() ) {
		    Neg( &modulus, b );
		} else {
		    modulus = *b;
		}
		if( neg ) {
		    Neg( &dividend, a );
		} else {
		    dividend = *a;
		}
		result.aval = dividend.aval % modulus.aval;
		result.bval = 0;
		if( neg ) {
		    Neg( dst, &result );
		} else {
		    *dst = result;
		}
	}
	dst->Adjust();
}
/***************************************************
	And	
	- bit wise and 
 ***************************************************/

void	And( CVector* dst, CVector* a, CVector* b ) 
{
	dst->aval = (a->aval | a->bval) & (b->aval | b->bval);
	dst->bval = dst->aval & (a->bval | b->bval);
	dst->Adjust();
}
/***************************************************
	Or	
	- bit wise or 
 ***************************************************/

void	Or( CVector* dst, CVector* a, CVector* b ) 
{
	CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
	dst->aval = a->aval | a->bval | b->aval | b->bval;
	dst->bval = dst->aval ^ ( a->aval & (a->bval ^ mask) |
				  b->aval & (b->bval ^ mask) );
	dst->Adjust();
}
/***************************************************
	Xor	
	- bit wise xor 
 ***************************************************/

void	Xor( CVector* dst, CVector* a, CVector* b ) 
{
	dst->bval = a->bval | b->bval;
	dst->aval = dst->bval | (a->aval ^ b->aval);
	dst->Adjust();
}
/***************************************************
	Xnor	
	- bit wise xnor 
 ***************************************************/

void	Xnor( CVector* dst, CVector* a, CVector* b ) 
{
	CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
	dst->bval = a->bval | b->bval;
	dst->aval = dst->bval | (mask ^ (a->aval ^ b->aval));
	dst->Adjust();
}
/***************************************************
	Le	
	- less than or equal 
 ***************************************************/

void	Le( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = 1;
		dst->bval = 1;
	} else if( !a->Signed() || !b->Signed() ) {
		dst->aval = a->aval <= b->aval;
		dst->bval = 0;
	} else {
		int aneg = a->IsNegative();
		int bneg = b->IsNegative();
		if( aneg == bneg ) {
			dst->aval = a->aval <= b->aval;
		} else if( aneg ) {
			dst->aval = 1;
		} else {
			dst->aval = 0;
		}
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Lt	
	- less than  
 ***************************************************/

void	Lt( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = 1;
		dst->bval = 1;
	} else if( !a->Signed() || !b->Signed() ) {
		dst->aval = a->aval < b->aval;
		dst->bval = 0;
	} else {
		int aneg = a->IsNegative();
		int bneg = b->IsNegative();
		if( aneg == bneg ) {
			dst->aval = a->aval < b->aval;
		} else if( aneg ) {
			dst->aval = 1;
		} else {
			dst->aval = 0;
		}
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Ge	
	- greater than or equal 
 ***************************************************/

void	Ge( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = 1;
		dst->bval = 1;
	} else if( !a->Signed() || !b->Signed() ) {
		dst->aval = a->aval >= b->aval;
		dst->bval = 0;
	} else {
		int aneg = a->IsNegative();
		int bneg = b->IsNegative();
		if( aneg == bneg ) {
			dst->aval = a->aval >= b->aval;
		} else if( aneg ) {
			dst->aval = 0;
		} else {
			dst->aval = 1;
		}
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Gt	
	- greater than
 ***************************************************/

void	Gt( CVector* dst, CVector* a, CVector* b ) 
{
	if( a->bval != 0 || b->bval != 0 ) {
		dst->aval = 1;
		dst->bval = 1;
	} else if( !a->Signed() || !b->Signed() ) {
		dst->aval = a->aval > b->aval;
		dst->bval = 0;
	} else {
		int aneg = a->IsNegative();
		int bneg = b->IsNegative();
		if( aneg == bneg ) {
			dst->aval = a->aval > b->aval;
		} else if( aneg ) {
			dst->aval = 0;
		} else {
			dst->aval = 1;
		}
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Ne	
	- not equal 
 ***************************************************/

void	Ne( CVector* dst, CVector* a, CVector* b ) 
{
	CUInt ignore = a->bval | b->bval;
	CUInt diff = a->aval ^ b->aval;
	CUInt amask = (((CUInt) 1) << a->GetWidth()) - 1;
	CUInt bmask = (((CUInt) 1) << b->GetWidth()) - 1;
	if( (diff & (ignore^(amask|bmask))) != 0 ) {
	    dst->aval = 1;
	    dst->bval = 0;
	} else if( ignore != 0 ) {
		dst->aval = 1;
		dst->bval = 1;
	} else {
		dst->aval = 0;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Eq	
	- equal 
 ***************************************************/

void	Eq( CVector* dst, CVector* a, CVector* b ) 
{
	dst->Adjust();
	CUInt ignore = a->bval | b->bval;
	CUInt diff = a->aval ^ b->aval;
	CUInt amask = (((CUInt) 1) << a->GetWidth()) - 1;
	CUInt bmask = (((CUInt) 1) << b->GetWidth()) - 1;
	if( (diff & (ignore^(amask|bmask))) != 0 ) {
	    dst->aval = 0;
	    dst->bval = 0;
	} else if( ignore != 0 ) {
		dst->aval = 1;
		dst->bval = 1;
	} else {
		dst->aval = 1;
		dst->bval = 0;
	}
	dst->Adjust();
}
/***************************************************
	Cne	
	- case not equal 
 ***************************************************/

void	Cne( CVector* dst, CVector* a, CVector* b ) 
{
	dst->aval = a->aval != b->aval ||
		    a->bval != b->bval;
	dst->bval = 0;
	dst->Adjust();
}
/***************************************************
	Ceq	
	- case equal 
 ***************************************************/

void	Ceq( CVector* dst, CVector* a, CVector* b ) 
{
	dst->aval = a->aval == b->aval &&
		    a->bval == b->bval;
	dst->bval = 0;
	dst->Adjust();
}
/***************************************************
	Lor	
	- logical or
 ***************************************************/

void	Lor( CVector* dst, CVector* a, CVector* b ) 
{
	CVector la(1);
	CVector lb(1);
	LogicalValue( &la, a );
	LogicalValue( &lb, b );
	Or( dst, &la, &lb );
}
/***************************************************
	Land	
	- logical and 
 ***************************************************/

void	Land( CVector* dst, CVector* a, CVector* b ) 
{
	CVector la(1);
	CVector lb(1);
	LogicalValue( &la, a );
	LogicalValue( &lb, b );
	And( dst, &la, &lb );
}
/***************************************************
	Com	
	- bit wise complement 
 ***************************************************/

void	Com( CVector* dst, CVector* a ) 
{
	CUInt mask = (((CUInt) 1) << dst->GetWidth()) - 1;
	dst->aval = (mask ^ a->aval) | a->bval;
	dst->bval = a->bval;	              
}
/***************************************************
	Rand	
	- reduction and 
 ***************************************************/

void	Rand( CVector* dst, CVector* a ) 
{
	CUInt acca;
	CUInt accb;
	INT32 i;
	int hasXZ = 0;

	acca = a->aval;
	accb = a->bval;

	for( i = 0; i < a->width; i++ ) {
		if( (accb & 1) != 0 ) {
			hasXZ = 1;
		} else if( (acca & 1) == 0 ) { 
			dst->aval = 0;
			dst->bval = 0;
			return;
		}
		acca = acca >> 1;
		accb = accb >> 1;
	}
	if( hasXZ ) {
	    dst->aval = 1; 
	    dst->bval = 1;
	} else {
	    dst->aval = 1; 
	    dst->bval = 0;
	}
}
/***************************************************
	Ror	
	- reduction or 
 ***************************************************/

void	Ror( CVector* dst, CVector* a ) 
{
	CUInt acca;
	CUInt accb;
	INT32 i;
	int hasXZ = 0;

	acca = a->aval;
	accb = a->bval;

	for( i = 0; i < a->width; i++ ) {
		if( (accb & 1) != 0 ) {
		        hasXZ = 1;
		} else if( (acca & 1) == 1 ) { 
			dst->aval = 1;
			dst->bval = 0;
			return;
		}
		acca = acca >> 1;
		accb = accb >> 1;
	}
	if( hasXZ ) {
		dst->aval = 1; 
		dst->bval = 1;
	} else {
		dst->aval = 0; 
		dst->bval = 0;
	}
}
/***************************************************
	Rnand	
	- reduction nand 
 ***************************************************/

void	Rnand( CVector* dst, CVector* a ) 
{
	Rand( dst, a );
	if( dst->bval == 0 ) {
	    dst->aval = dst->aval ^ 1;
	}
}
/***************************************************
	Rnor	
	- reduction nor 
 ***************************************************/

void	Rnor( CVector* dst, CVector* a ) 
{
	Ror( dst, a );
	if( dst->bval == 0 ) {
	    dst->aval = dst->aval ^ 1;
	}
}
/***************************************************
	Rxor	
	- reduction xor 
 ***************************************************/

void	Rxor( CVector* dst, CVector* a ) 
{
	CUInt acca;
	CUInt accb;
	INT32 result;
	INT32 i;

	acca = a->aval;
	accb = a->bval;

	result = 0;
	for( i = 0; i < a->width; i++ ) {
		if( (accb & 1) != 0 ) {
			dst->aval = 1;
			dst->bval = 1;
			return;
		} 
		if( (acca & 1) == 1 ) { 
			result ^= 1;
		}
		acca = acca >> 1;
		accb = accb >> 1;
	}
	dst->aval = result; 
	dst->bval = 0;
}
/***************************************************
	Rxnor	
	- reduction xnor 
 ***************************************************/

void	Rxnor( CVector* dst, CVector* a ) 
{
	CUInt acca;
	CUInt accb;
	INT32 result;
	INT32 i;

	acca = a->aval;
	accb = a->bval;

	result = 1;
	for( i = 0; i < a->width; i++ ) {
		if( (accb & 1) != 0 ) {
			dst->aval = 1;
			dst->bval = 1;
			return;
		} 
		if( (acca & 1) == 1 ) { 
			result ^= 1;
		}
		acca = acca >> 1;
		accb = accb >> 1;
	}
	dst->aval = result; 
	dst->bval = 0;
}
/***************************************************
	Not	
	- logical not 
 ***************************************************/

void	Not( CVector* dst, CVector* a ) 
{
    LogicalValue( dst, a );
    if( !dst->HasXZ() ) {
	dst->aval = dst->aval ^ 1;
    }
}
/***************************************************
	Hook	
	- ? operator 
 ***************************************************/

void	Hook( CVector* dst, CVector* c, CVector* a, CVector* b ) 
{
	CVector cond(1);
	LogicalValue( &cond, c );
	if( cond.HasXZ() ) {
		dst->bval = a->bval | b->bval | (a->aval ^ b->aval);
		dst->aval = dst->bval | a->aval;
	} else if( (cond.aval & 1) != 0 ) {
		*dst = *a;
	} else {
		*dst = *b;
	}
	dst->Adjust();
}
/***************************************************
	operator == CVector&
	- compare to vectors for exact match 
 ***************************************************/

int	CVector::operator==( CVector& v ) 
{
	return( v.aval == aval && v.bval == bval );
}
/***************************************************
	operator == UINT32
	- compare to vector to UINT32
 ***************************************************/

int	CVector::operator==( UINT32 v ) 
{
	return( aval == v && bval == 0 );
}
/***************************************************
	GetWidthExp	
	- return expression for width
 ***************************************************/
CNode* CVector::GetWidthExp() 
{
    return cINT32(GetWidth());
}
/***************************************************
	GetNodeType	
	- return type
 ***************************************************/
NodeType_t CVector::GetNodeType()
{
    if( _signed ) {
	return eS;
    }
    return eB;
}

/***************************************************
	LoadReal	
	- load vector with real value
 ***************************************************/
void CVector::LoadReal( double d )
{
    int sign;
    if( d > 0 ) {
	sign = 1;
    } else {
	sign = -1;
	d = -d;
    }
    d = d + .5;
    aval = 0;
    bval = 0;
    for( int index = 0; index < GetWidth(); index += 32 ) {
	aval = aval | (unsigned) (d/ldexp(1,index));	
    }
    if( sign < 0 ) {
	CVector tmp( GetWidth() );
	tmp = *this;
	Neg( this, &tmp );
    }
}

/***************************************************
	GetReal	
	- convert to real
 ***************************************************/
double CVector::GetReal()
{
    CVector n(GetWidth());
    CUInt mask = (((CUInt) 1) << GetWidth()) - 1;
    n.aval = aval & (bval ^ mask);
    n.bval = 0;
    double sign = 1;
    if( IsNegative() ) {
        sign = -1;
	n.Signed(1);
    	Neg( &n, this );
    }
    return sign*(double)n.aval.GetUINT64();
}

