/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

/******************************************************
 
	glue.cc
	- utility routines for log/console io

 *****************************************************/

#include <stdio.h>
#include "glue.h"
#include "cdecl.h"
#include "cbackend.h"

/******************************************************
 *
 *	Message registration routines
 *
 *****************************************************/

map<string,Message*> *Message::abbrev2Message = NULL;
list<Message*> *Message::messages = NULL;

/****************************************************
 	RegisterWarning
	- register a warning message 
 ***************************************************/
Message* Message::RegisterWarning( CBackend* tool, 
		      Action_t defaultAction,
		      const char* abbrev, 
		      const char* format, ... )
{
	char buffer[1024];
	string abuffer;
	va_list ap;
	va_start( ap, format );
	vsnprintf( buffer, sizeof(buffer), format, ap );
	va_end( ap );
	if( tool ) {
	    abuffer = tool->GetToolName();
	} else {
	    abuffer = "vrq";
	}
	abuffer += "-";
	abuffer += abbrev;
	return new Message( tool, defaultAction, 0, 
				abuffer.c_str(), buffer, format );
}

/****************************************************
 	RegisterError
	- register an error message 
 ***************************************************/
Message* Message::RegisterError( CBackend* tool, 
		      const char* abbrev, 
		      const char* format, ... )
{
	char buffer[1024];
	string abuffer;
	va_list ap;
	va_start( ap, format );
	vsnprintf( buffer, sizeof(buffer), format, ap );
	va_end( ap );
	if( abbrev ) {
	    abuffer = tool->GetToolName();
	} else {
	    abuffer = "vrq";
	}
	abuffer += "-";
	abuffer += abbrev;
	return new Message( tool, eERROR, 1, abuffer.c_str(), buffer, format );
}

/*****************************************************
     	Find
	- find message from abbreviation
 ****************************************************/
Message* Message::Find( const char* abbrev )
{
	if( !abbrev2Message && abbrev2Message->find(abbrev) == abbrev2Message->end() ) {
	    return NULL;
	}	
 	return (*abbrev2Message)[abbrev];


}

/*****************************************************
     	Message
	- constructor
 ****************************************************/
Message::Message( CBackend* tool,
   	     Action_t action,
	     int    locked,
	     const char* abbrev,
	     const char* helpText,
	     const char* format )
	: tool(tool),
	  action(action),
	  locked(locked),
	  abbrev(abbrev),
	  helpText(helpText),
	  format(format)
{
	/*
 	 * make sure message abbreviations are unique
 	 */
	if( !abbrev2Message ) {
	    abbrev2Message = new map<string,Message*>();
	    messages = new list<Message*>();
	}
	MASSERT( abbrev2Message->find(abbrev) == abbrev2Message->end() );
 	(*abbrev2Message)[abbrev] = this;
	messages->push_back(this);
}

/*****************************************************
     	message
	- print a message 
 ****************************************************/
void message( struct Coord_t* loc, Message* message, ... )
{
	const char* format = message->Format();
	const char* abbreviation = message->Abbreviation();

	switch( message->Action() ) {
	case Message::eIGNORE:
	    return;
	case Message::eWARNING:
	    logprintf( "WARNING " );
	    warningCount++;
	    break;
	case Message::eERROR:
	    logprintf( "ERROR " );
	    errorCount++;
	    break;
	case Message::eINFO:
	    logprintf( "INFO " );
	    break;
	default:
	    MASSERT( FALSE );
	}
	logprintf( "[%s] ", abbreviation );
	if( loc ) {
	    logprintf( "\"%s\" line %d: ", loc->filename, loc->lineno );
	}
	logprintf( ": " );

	va_list	args;
	va_start( args, message );
	vlogprintf( format, args );
	va_end( args );
	logprintf( "\n" );
}

/*******************************************************
	info
	- info printing routine
 *******************************************************/

void	info( Coord_t* loc, const char* format, ... )
{
	va_list	args;

	va_start( args, format );
	if( loc ) {
		logprintf( "INFO \"%s\" line %d: ", loc->filename, loc->lineno );
	} else {
		logprintf( "INFO: " );
	}
	vlogprintf( format, args );
	logprintf( "\n" );
	va_end( args );
}

/*******************************************************
	trace
	- trace printing routine
 *******************************************************/

void	trace( Coord_t* loc, const char* format, ... )
{
	va_list	args;

	return;

	va_start( args, format );
	if( loc ) {
		logprintf( "TRACING \"%s\" line %d: ", loc->filename, loc->lineno );
	} else {
		logprintf( "TRACING: " );
	}
	vlogprintf( format, args );
	logprintf( "\n" );
	va_end( args );
}

/*******************************************************
	warning
	- warning printing routine
 *******************************************************/

void	warning( Coord_t* loc, const char* format, ... )
{
	va_list	args;

	va_start( args, format );
	if( loc ) {
		logprintf( "WARNING \"%s\" line %d: ", loc->filename, loc->lineno );
	} else {
		logprintf( "WARNING: " );
	}
	vlogprintf( format, args );
	logprintf( "\n" );
	va_end( args );
	warningCount++;
}

/*******************************************************
	fatal
	- fatal printing routine
 *******************************************************/

void	fatal( Coord_t* loc, const char* format, ... )
{
	va_list	args;

	va_start( args, format );
	if( loc ) {
		logprintf( "INTERNAL ERROR \"%s\" line %d: ", loc->filename, loc->lineno );
	} else {
		logprintf( "INTERNAL ERROR: " );
	}
	vlogprintf( format, args );
	logprintf( "\n" );
	va_end( args );
	abort();
}

/*******************************************************
	error
	- error printing routine
 *******************************************************/

void	error( struct Coord_t* loc, const char* format, ... )
{
	va_list	args;

	va_start( args, format );
	if( loc ) {
		logprintf( "ERROR \"%s\" line %d: ", loc->filename, loc->lineno );
	} else {
		logprintf( "ERROR: " );
	}
	vlogprintf( format, args );
	logprintf( "\n" );
	va_end( args );
	errorCount++;
}

/*******************************************************
	vlogprintf
	- print message to log file 
 *******************************************************/

void	vlogprintf( const char* format, va_list args )
{
	if( logfile != stdout ) {
	    va_list args1;
	    va_copy( args1, args );
	    vprintf( format, args1 );
	}
	vfprintf( logfile, format, args );
}

/*******************************************************
	logprintf
	- print message to log file 
 *******************************************************/

void	logprintf( const char* format, ... )
{
	va_list	args;

	va_start( args, format );
	vlogprintf( format, args );
	va_end( args );
}
