/*****************************************************************************
 * Copyright (C) 1997-2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   member.cc
 *		- member info routines
 *
 ******************************************************************************
 */

#include "member.h"
#include "array.h"
#include "cinterfacetype.h"

/**
 * Get width of structure member.
 * \param ref reference to a decl
 * \param index index list
 * \return width of expression
 */
int Member::Width( CNode* ref, CSymbol* member )
{
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    if( ResolveReference( ref, member, dataType, 
	   unpackedDimensions, packedDimensions ) != 1  ||
	unpackedDimensions.size() != 0 ) {
        return 0; 
    } 

    MASSERT( dataType );
    return dataType->GetPackedWidth();
}
/**
 * Get width of structure member as an expression.
 * \param ref reference to a decl
 * \param index index list
 * \return width expression.
 */
CNode* Member::WidthExp( CNode* ref, CSymbol* member )
{
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    if( ResolveReference( ref, member, dataType, 
	  unpackedDimensions, packedDimensions ) != 1  ||
	unpackedDimensions.size() != 0 ) {
        return NULL; 
    } 
    MASSERT( dataType );
    return dataType->GetPackedWidthExp();
}
/**
 * Determine if width of structure member is constant.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return non-zero if width is constant. ie dependent upon only
 *         on parameters and constants.
 */
int Member::WidthConstant( CNode* ref, CSymbol* member ) 
{
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    if( ResolveReference( ref, member, dataType, 
	  unpackedDimensions, packedDimensions ) != 1  ||
	unpackedDimensions.size() != 0 ) {
        return 0; 
    } 
    MASSERT( dataType );
    return dataType->IsPackedWidthConstant();
}
/**
 * Determine if width of structure member is volatile.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return non-zero if width is volatile. ie dependent upon only
 *         on parameters or variables.
 */
int Member::WidthVolatile( CNode* ref, CSymbol* member ) 
{ 
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    if( ResolveReference( ref, member, dataType, 
	  unpackedDimensions, packedDimensions ) != 1  ||
	unpackedDimensions.size() != 0 ) {
        return 0; 
    } 
    MASSERT( dataType );
    return dataType->IsPackedWidthVolatile();
}
/**
 * Determine if width of structure member can be evaluated.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return non-zero if width can be evaluated.
 */
int Member::WidthEvaluateable( CNode* ref, CSymbol* member ) 
{ 
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    if( ResolveReference( ref, member, dataType, 
	  unpackedDimensions, packedDimensions ) != 1  ||
	unpackedDimensions.size() != 0 ) {
        return 0; 
    } 
    MASSERT( dataType );
    return dataType->IsPackedWidthEvaluateable();
}
/**
 * Determine type of structure member.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return type of array return value.
 */
NodeType_t Member::Type( CNode* ref, CSymbol* member )
{ 
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    int result = ResolveReference( ref, member, dataType, 
		unpackedDimensions, packedDimensions );
    if( result != 1  ||
	unpackedDimensions.size() != 0 ) {
        return eU; 
    } 
    MASSERT( dataType );
    return dataType->GetNodeType();
}
/**
 * Get datatype info for struct reference
 * \param ref reference to a decl/subreference
 * \param index index list 
 * \param dataType resolved base data type
 * \param unpackedDimensions list of unresolved unpacked dimensions
 * \return -1 if reference is valid but unresolvable
 * 	    0 if reference is invalid
 * 	    1 if reference is valid and was resolved
 */
int Member::ResolveReference( 
	CNode*      ref, 
	CSymbol*    symbol,
	CDataType*& dataType,
	vector<CNode*>& unpackedDimensions,
	vector<CNode*>& packedDimensions )
{
    switch( ref->GetOp() ) {
    case eARRAY: {
	vector<CNode*> unpackedDimensions;
	vector<CNode*> packedDimensions;
	vector<CNode*> residualPackedIndices;
	int result = Array::ResolveReference( 
		ref->Arg<CNode*>(0),
		ref->Arg<CNode*>(1),
		dataType,
		unpackedDimensions,
		packedDimensions,
		residualPackedIndices );
	if( result != 1 ) {
	    return result;
	}
	if( unpackedDimensions.size() != 0 ) {
	    return 0;
	} 
	if( residualPackedIndices.size() != packedDimensions.size() ) {
	    return 0;
	}
	} break;
    case eMEMBER: {
	vector<CNode*> dummy1;
	vector<CNode*> dummy2;
	int result =ResolveReference( ref->Arg<CNode*>(0),
		ref->Arg<CSymbol*>(1), dataType, dummy1, dummy2 );
	if( result != 1 ) {
	    return result;
	}
     	} break;
    case eEXTERNAL_REF:
	return -1;
    default: 
	dataType = ref->Arg<CDecl*>(0)->GetDataType();
	break;
    }
    while( dataType->GetVarDataType() == ePROXY &&
	   unpackedDimensions.size() == 0 && packedDimensions.size() == 0 ) {
	CDecl* decl = dynamic_cast<CProxyType*>(dataType)->GetDecl();
        for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
            unpackedDimensions.push_back(decl->GetUnpackedRange(i));
        }
	dataType = dynamic_cast<CProxyType*>(dataType)->GetDataType();
        for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
            packedDimensions.push_back(dataType->GetPackedRange(i));
        }
    }
    CDecl* decl = NULL;
    switch( dataType->GetVarDataType() ) {
    case eSTRUCT:
    case eUNION:
    case eUNION_TAGGED:
    {
        CStructType* st = dynamic_cast<CStructType*>(dataType);
        decl = st->FindMember( symbol );
    }
    break;
    case eINTERFACE_TYPE:
    {
        CInterfaceType* it = dynamic_cast<CInterfaceType*>(dataType);
        decl = it->FindMember( symbol );
    }
    break;
    default:
	return -1;
    }

    if( !decl ) {
        return 0;
    }
    dataType = decl->GetDataType();
    for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
        unpackedDimensions.push_back(decl->GetUnpackedRange(i));
    }
    for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
        packedDimensions.push_back(dataType->GetPackedRange(i));
    }
    while( dataType->GetVarDataType() == ePROXY ) {
        CDecl* decl = dynamic_cast<CProxyType*>(dataType)->GetDecl();
        for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
            unpackedDimensions.push_back(decl->GetUnpackedRange(i));
        }
        dataType = dynamic_cast<CProxyType*>(dataType)->GetDataType();
        for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
            packedDimensions.push_back(dataType->GetPackedRange(i));
        }
    }
    return 1;
}
